<?php

if ( ! defined( 'ABSPATH' ) ) {
	exit;	// Exit if accessed directly
}

/**
 * Check if WooCommerce is active
 */
if ( in_array( 'woocommerce/woocommerce.php', apply_filters( 'active_plugins', get_option( 'active_plugins' ) ) ) ) {

	/**
	 * Persistent class for admin order creation functions
	 */
	if ( ! class_exists( 'WC_Myfreight_Admin_Order' ) ) {
		class WC_Myfreight_Admin_Order {

			public function __construct() {

				add_action( 'woocommerce_order_item_add_line_buttons', array( $this, 'myfreightshipping_add_shipping_button' ), 10, 1 );
				add_action( 'wp_ajax_woocommerce_add_order_myfreight_shipping', array( $this, 'myfreightshipping_add_shipping_ajax_service' ), 10, 0 );
			}


			/**
			 * Insert add shipping line button in admin order add item section
			 */
			public function myfreightshipping_add_shipping_button( $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				wp_register_script( 'myfreight-admin-order', plugins_url( 'js/myfreightAdminOrder.js', __FILE__ ), array( 'wc-admin-order-meta-boxes' ), false, true );
				wp_enqueue_script( 'myfreight-admin-order' );

?>
<button type="button" class="button add-order-myfreight-shipping"><?php esc_html_e( 'Add ' . $settings['account_type'] . ' shipping', 'woocommerce' ); ?></button>
<?php
			}


			/**
			 * Ajax service invoked when admin order add shipping button clicked
			 */
			public function myfreightshipping_add_shipping_ajax_service() {
				check_ajax_referer( 'order-item', 'security' );

				if ( ! current_user_can( 'edit_shop_orders' ) ) {
					wp_die( -1 );
				}

				$timezone_string = get_option( 'timezone_string' );
				date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );

				$response = array(
					'html' => ''
				);

				try {
					$order_id = isset( $_POST['order_id'] ) ? absint( $_POST['order_id'] ) : 0;
					$order    = wc_get_order( $order_id );

					if ( ! $order ) {
						throw new Exception( __( 'Invalid order', 'woocommerce' ) );
					}

					if ( empty( $order->get_date_created() ) ) {
						throw new Exception( __( 'Please create order first', 'woocommerce' ) );
					}

					if ( method_exists( $order, 'has_shipping_address' ) && $order->has_shipping_address() ) {	// If shipping address entered
						$receiver_city     = $order->get_shipping_city();
						$receiver_state    = $order->get_shipping_state();
						$receiver_postcode = $order->get_shipping_postcode();
						$receiver_company  = $order->get_shipping_company();
					} else {
						$receiver_city     = $order->get_billing_city();
						$receiver_state    = $order->get_billing_state();
						$receiver_postcode = $order->get_billing_postcode();
						$receiver_company  = $order->get_billing_company();
					}

					if ( empty( $receiver_city ) ) {
						throw new Exception( __( 'City missing', 'woocommerce' ) );
					}
					if ( empty( $receiver_state ) ) {
						throw new Exception( __( 'State missing', 'woocommerce' ) );
					}
					if ( empty( $receiver_postcode ) ) {
						throw new Exception( __( 'Postcode missing', 'woocommerce' ) );
					}

					$receiver_address_type = get_post_meta( $order_id, '_address_type', true );
					if ( empty($receiver_address_type) ) {
						$settings = get_option( 'woocommerce_myfreightshipping_settings' );
						$receiver_address_type = (string)array_pop( explode( ' ', $settings['delivery_address_type'] ) );
					}
					$unloading_capability = get_post_meta( $order_id, '_unloading_capability', true );
					if ( !empty($unloading_capability) && $unloading_capability == 'Yes' ) {
						$receiver_unloading = true;
					} else {
						$receiver_unloading = false;
					}

					$package = array(
						'contents'    => array(),
						'destination' => array(
							'city'                 => $receiver_city,
							'state'                => $receiver_state,
							'postcode'             => $receiver_postcode,
							'country'              => 'AU',
							'company'              => $receiver_company,
							'address_type'         => $receiver_address_type,
							'unloading_capability' => $receiver_unloading
						)
					);

					if ( method_exists( $order, 'get_items' ) && !empty( $order_items = $order->get_items() ) ) {
						foreach ( $order_items as $item_id => $item ) {

							$package['contents'][$item_id] = array(
								'product_id'        => $item->get_product_id(),
								'variation_id'      => $item->get_variation_id(),
								'quantity'          => $item->get_quantity(),
								'line_tax_data'     => $item->get_taxes(),
								'line_subtotal'     => $item->get_subtotal(),
								'line_subtotal_tax' => $item->get_subtotal_tax(),
								'line_total'        => $item->get_total(),
								'line_tax'          => $item->get_total_tax(),
								'data'              => $item->get_product()
							);
						}
					} else {
						throw new Exception( __( 'Product missing', 'woocommerce' ) );
					}

					myfreight_shipping_method_init();
					$myfreight_shipping = new WC_Myfreight_Shipping();
					$rates = $myfreight_shipping->calculate_shipping( $package );

					$order_taxes      = $order->get_taxes();
					$shipping_methods = WC()->shipping() ? WC()->shipping()->load_shipping_methods() : array();

					// Add new Myfreight shipping
					$instance_id = 0;
					$service_codes = array();
					foreach ( $rates as $rate ) {

						$item = new WC_Order_Item_Shipping();
						$item->set_order_id( $order_id );
						$item->set_name( $rate['label'] );
						$item->set_method_id( 'myfreightshipping' );
						$item->set_instance_id( $instance_id++ );
						$item->set_total( $rate['cost'] );
						$item->set_taxes( WC_Tax::calc_shipping_tax( $rate['cost'], WC_Tax::get_shipping_tax_rates() ) );
						$item->add_meta_data( 'Carrier_Service',   $rate['meta_data']['Carrier_Service'] );
						$item->add_meta_data( 'Service_Code',      $rate['meta_data']['Service_Code'] );
						$item->add_meta_data( 'Store_Code',        $rate['meta_data']['Store_Code'] );
						$item->add_meta_data( 'Sender_Postcode',   $rate['meta_data']['Sender_Postcode'] );
						$item->add_meta_data( 'Receiver_Postcode', $rate['meta_data']['Receiver_Postcode'] );
						$item_id = $item->save();

						ob_start();
						include WP_PLUGIN_DIR . '/woocommerce/includes/admin/meta-boxes/views/html-order-shipping.php';
						$response['html'] .= ob_get_clean();

						$service_codes[] = $rate['meta_data']['Service_Code'];
					}

					$order->add_order_note( sprintf( __( 'Added shipping: %s', 'woocommerce' ), implode( ', ', $service_codes ) ), false, true );

				} catch ( Exception $e ) {
					wp_send_json_error( array( 'error' => $e->getMessage() ) );
				}

				wp_send_json_success( $response );
			}

		}
	}
	if ( is_admin() ) {
		// Instantiate class
		new WC_Myfreight_Admin_Order();
	}

}

