<?php

if ( ! defined( 'ABSPATH' ) ) {
	exit;	// Exit if accessed directly
}

/**
 * Check if WooCommerce is active
 */
if ( in_array( 'woocommerce/woocommerce.php', apply_filters( 'active_plugins', get_option( 'active_plugins' ) ) ) ) {

	/**
	 * Persistent class for ancillary functions
	 */
	if ( ! class_exists( 'WC_Myfreight_Ancillary' ) ) {
		class WC_Myfreight_Ancillary {

			public function __construct() {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				// Add action links on plugins page
				add_filter( 'plugin_action_links_myfreightshipping/myfreightshipping.php', array( $this, 'myfreightshipping_plugin_action_links' ), 10, 1 );

				// Load plugin styling and scripts
				add_action( 'wp_enqueue_scripts', array( $this, 'myfreightshipping_enqueue_scripts' ), 10, 1 );

				// Initialise shipping address helper on Cart, Checkout and Single Product pages
				if ( $settings['address_helper_cart'] == 'yes' ) {
					if ( ! has_action( 'woocommerce_before_shipping_calculator' ) ) {
						add_action( 'woocommerce_before_shipping_calculator', array( $this, 'shipping_locality_selection_cart' ), 9, 0 );
					}
				}
				if ( $settings['address_helper_checkout'] == 'yes' ) {
					add_filter( 'woocommerce_checkout_fields', array( $this, 'myfreightshipping_placeholders_checkout' ), 9, 1 );
				}
				if ( $settings['address_helper_product'] == 'yes' ) {
					if ( ! has_action( 'woocommerce_single_product_summary' ) ) {
						add_action( 'woocommerce_single_product_summary', array( $this, 'shipping_locality_selection_product' ), 98, 0 );
					}
					add_filter( 'woocommerce_shipping_calculator_enable_city', array( $this, 'shipping_calculator_city' ), 10, 1 );
				}

				// Initialise address type (residential or commercial) check on Cart page, Checkout page and Order Detail
				if ( substr( $settings['delivery_address_type'], 0, 7 ) == 'Enquire' ) {
					add_action( 'woocommerce_calculated_shipping', array( $this, 'shipping_calculator_address_type' ), 9, 0 );
					add_filter( 'woocommerce_checkout_fields', array( $this, 'myfreightshipping_address_type_checkout' ), 9, 1 );
					add_filter( 'woocommerce_checkout_get_value', array( $this, 'myfreightshipping_address_type_value' ), 10, 2 );
					add_action( 'wp_ajax_receiver_address_type', array( $this, 'myfreightshipping_address_type_ajax' ), 10, 0 );
					add_action( 'wp_ajax_nopriv_receiver_address_type', array( $this, 'myfreightshipping_address_type_ajax' ), 10, 0 );
					add_action( 'woocommerce_checkout_create_order', array( $this, 'myfreightshipping_address_type_create_order' ), 9, 2 );
					add_action( 'woocommerce_admin_order_data_after_shipping_address', array( $this, 'myfreightshipping_address_type_order' ), 9, 1 );
					add_action( 'woocommerce_process_shop_order_meta', array( $this, 'myfreightshipping_address_type_save' ), 9, 2 );
				}

				// Initialise unloading capability check on Cart page, Checkout page and Order Detail
				if ( $settings['unloading_capability_check'] == 'yes' ) {
					add_action( 'woocommerce_calculated_shipping', array( $this, 'shipping_calculator_unloading_capability' ), 9, 0 );
					add_filter( 'woocommerce_checkout_fields', array( $this, 'myfreightshipping_unloading_capability_checkout' ), 8, 1 );
					add_filter( 'woocommerce_checkout_get_value', array( $this, 'myfreightshipping_unloading_capability_value' ), 10, 2 );
					add_action( 'wp_ajax_unloading_capability', array( $this, 'myfreightshipping_unloading_capability_ajax' ), 10, 0 );
					add_action( 'wp_ajax_nopriv_unloading_capability', array( $this, 'myfreightshipping_unloading_capability_ajax' ), 10, 0 );
					add_action( 'woocommerce_checkout_create_order', array( $this, 'myfreightshipping_unloading_capability_create_order' ), 9, 2 );
					add_action( 'woocommerce_admin_order_data_after_shipping_address', array( $this, 'myfreightshipping_unloading_capability_order' ), 9, 1 );
					add_action( 'woocommerce_process_shop_order_meta', array( $this, 'myfreightshipping_unloading_capability_save' ), 9, 2 );
				}

				// Add Shipping Phone Number field to Checkout page and Order Detail
				add_filter( 'woocommerce_checkout_fields', array( $this, 'myfreightshipping_shipping_phone_checkout' ), 11, 1 );
				add_filter( 'woocommerce_shipping_fields', array( $this, 'myfreightshipping_shipping_phone_shipping_fields' ), 11, 2 );
				add_filter( 'woocommerce_get_order_address', array( $this, 'myfreightshipping_shipping_phone_order_address' ), 11, 3 );
				if ( ! has_action( 'woocommerce_admin_shipping_fields' ) ) {
					add_action( 'woocommerce_admin_shipping_fields', array( $this, 'myfreightshipping_shipping_phone_order_admin' ), 11, 1 );
				}
//				if ( ! has_action( 'woocommerce_admin_order_data_after_shipping_address' ) ) {
//					add_action( 'woocommerce_admin_order_data_after_shipping_address', array( $this, 'myfreightshipping_phone_number_shipping' ), 8, 1 );	// DEPRECATED
//				}

				// Include transport carrier logo in Cart and Checkout shipping methods
				if ( $settings['show_carrier_logo'] == 'yes' ) {
					add_filter( 'woocommerce_cart_shipping_method_full_label', array( $this, 'myfreightshipping_carrier_logo' ), 10, 2 );
				}

				// Initialise item packing
				if ( $settings['packing_rules_active'] == 'yes' ) {
					add_filter( 'my_pack_items', array( $this, 'rule_pack_items' ), 6, 3 );		// Packing Rules are performed first
				}
				if ( $settings['packing_auto_active'] == 'yes' ) {
					add_filter( 'my_pack_items', array( $this, 'auto_pack_items' ), 9, 3 );		// Auto Packing is performed last
				}

				// Initialise item type estimator
				add_filter( 'my_est_item_type', array( $this, 'estimate_item_type' ), 9, 8 );

				// Initialise receiver residential check
				add_filter( 'my_residential', array( $this, 'receiver_residential' ), 9, 4 );

				// Initialise carrier surcharges filter
				add_filter( 'my_surcharges', array( $this, 'get_carrier_surcharges' ), 9, 4 );

				// Initialise quote sorting filter
				add_filter( 'my_sort_quotes', array( $this, 'sort_quotes_by_cost' ), 9, 2 );
			}


			/**
			 * Add plugin Settings link on the Installed Plugins page (plugins.php)
			 */
			public function myfreightshipping_plugin_action_links( $links ) {

				$links = array_merge( array(
					'<a href="' . admin_url( '/admin.php?page=wc-settings&tab=shipping&section=myfreightshipping' ) . '">' . 'Settings' . '</a>'
				), $links );

				return $links;
			}


			/**
			 * Enqueue plugin styling
			 */
			public function myfreightshipping_enqueue_scripts( $hook ) {
				if ( function_exists( 'is_woocommerce' ) ) {

					$settings = get_option( 'woocommerce_myfreightshipping_settings' );

					if ( ( $settings['address_helper_cart'] == 'yes' && is_cart() ) ||
					     ( $settings['address_helper_checkout'] == 'yes' && is_checkout() ) ||
					     ( $settings['address_helper_product'] == 'yes' && is_product() &&
					     		class_exists( 'rp_shipping_calculator' ) ) ) {		// If "WooCommerce Shipping Calculator on Product Page" plugin active
						wp_enqueue_style( 'jquery-ui-styles', plugins_url( 'css/myfreight-jquery-ui.min.css', __FILE__ ), null, '1.12.1', 'screen' );
						wp_register_script( 'myfreight-localities', plugins_url( 'js/myfreightLocalities.js', __FILE__ ), null, false, true );
						wp_register_script( 'myfreight-shipping', plugins_url( 'js/myfreightShipping.js', __FILE__ ), array( 'jquery-ui-autocomplete', 'myfreight-localities' ), false, true );
						wp_enqueue_script( 'myfreight-shipping' );
						if ( ! wp_style_is( 'myfreightshipping', 'enqueued' ) ) {
							wp_enqueue_style( 'myfreightshipping', plugins_url( 'css/myfreightshipping.css', __FILE__ ), null, false, 'all' );
						}
					}
				}
			}


			/**
			 * Destination Locality selection for each placement
			 */
			public function shipping_locality_selection_cart() {
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$placeholder = !empty($settings['address_helper_placeholder_cart']) ? $settings['address_helper_placeholder_cart'] : 'Enter suburb or postcode ...';
				$address_type_query = null;
				if ( substr( $settings['delivery_address_type'], 0, 7 ) == 'Enquire' ) {
					$address_type_purpose = ( substr( $settings['delivery_address_type'], -10 ) == 'Commercial' ) ? 'Residential' : 'Commercial';
					$address_type_query = 'Is delivery address ' . $address_type_purpose . '?';
				}
				$unloading_query = null;
				if ( $settings['unloading_capability_check'] == 'yes' ) {
					$unloading_query = !empty($settings['unloading_capability_text']) ? $settings['unloading_capability_text'] : 'Forklift or unloading facility at delivery address?';
				}
				$this->shipping_locality_selection( 'cart', null, $placeholder, $address_type_query, $unloading_query );
			}
			public function shipping_locality_selection_product() {
				if ( !class_exists( 'rp_shipping_calculator' ) ) {		// If "WooCommerce Shipping Calculator on Product Page" plugin not active
					return;
				}
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$product_heading = '<h5 class="shipping_estimate_title">Shipping Calculator</h5><p class="shipping_estimate_note">Includes this product qty and cart items</p>';
				$placeholder = !empty($settings['address_helper_placeholder_product']) ? $settings['address_helper_placeholder_product'] : 'Enter suburb or postcode ...';
				$address_type_query = null;
				if ( substr( $settings['delivery_address_type'], 0, 7 ) == 'Enquire' ) {
					$address_type_purpose = ( substr( $settings['delivery_address_type'], -10 ) == 'Commercial' ) ? 'Residential' : 'Commercial';
					$address_type_query = 'Is delivery address ' . $address_type_purpose . '?';
				}
				$unloading_query = null;
				if ( $settings['unloading_capability_check'] == 'yes' ) {
					$unloading_query = !empty($settings['unloading_capability_text']) ? $settings['unloading_capability_text'] : 'Forklift or unloading facility at delivery address?';
				}
//				$this->shipping_locality_selection( 'product', $product_heading, $placeholder, $address_type_query, $unloading_query );
				$this->shipping_locality_selection( 'product', $product_heading, $placeholder, null, null );
			}


			/**
			 * Destination Locality selection field and scripts
			 */
			function shipping_locality_selection( $placement, $heading, $placeholder, $address_type_query, $unloading_query ) {
				$locality_html = '';
				if ( $placeholder ) {
					$locality_html .= '<div id="' . $placement . '_locality_field">';
					$locality_html .= $heading;
					$locality_html .= '<input type="text" class="input-text locality_autocomplete" value="" placeholder="' . $placeholder . '" name="' . $placement . '_locality" id="' . $placement . '_locality">';
					$locality_html .= '</div>';
				}
				if ( $address_type_query ) {
					$checked = ( WC()->session->get( 'address_type', 0 ) == 1 ) ? ' checked="checked"' : '';
					$locality_html .= '<div id="' . $placement . '_address_type_tickbox">';
					$locality_html .= '<input type="checkbox" class="input-checkbox" value="1" name="' . $placement . '_address_type" id="' . $placement . '_address_type"' . $checked . '>';
					$locality_html .= '<span class="' . $placement . '_address_type_query">' . $address_type_query . '</span>';
					$locality_html .= '</div>';
				}
				if ( $unloading_query && WC()->session->get( 'unloading_heavy_item', false ) ) {		// Inhibit tickbox display if no heavy items
					$checked = ( WC()->session->get( 'unloading_capability', 0 ) == 1 ) ? ' checked="checked"' : '';
					$locality_html .= '<div id="' . $placement . '_unloading_capability_tickbox">';
					$locality_html .= '<input type="checkbox" class="input-checkbox" value="1" name="' . $placement . '_unloading_capability" id="' . $placement . '_unloading_capability"' . $checked . '>';
					$locality_html .= '<span class="' . $placement . '_unloading_capability_query">' . $unloading_query . '</span>';
					$locality_html .= '</div>';
				}

				/*
				 * Filter hook for customisation of address helper field.
				 */
				$locality_html = apply_filters( 'my_locality_helper', $locality_html, $placement );

				echo $locality_html;
			}


			/**
			 * Add custom placeholder to Suburb fields on checkout page
			 */
			public function myfreightshipping_placeholders_checkout( $fields ) {
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				if ( !empty($settings['address_helper_placeholder_checkout']) ) {
					$placeholder = $settings['address_helper_placeholder_checkout'];

					$fields['billing']['billing_city']['placeholder']   = _x($placeholder, 'placeholder', 'woocommerce');
					$fields['shipping']['shipping_city']['placeholder'] = _x($placeholder, 'placeholder', 'woocommerce');
				}

				$fields['billing']['billing_city']['autocomplete']       = false;
				$fields['billing']['billing_postcode']['autocomplete']   = false;
				$fields['shipping']['shipping_city']['autocomplete']     = false;
				$fields['shipping']['shipping_postcode']['autocomplete'] = false;

				return $fields;
			}


			/**
			 * Enable city field in Single Product Shipping Calculator plugin
			 */
			public function shipping_calculator_city( $enable ) {
				return true;
			}


			/**
			 * Handle input from custom field in Cart shipping calculator -> receiver address type
			 */
			public function shipping_calculator_address_type() {
				$address_type = isset( $_POST['calc_shipping_address_type'] ) ? wc_clean( wp_unslash( $_POST['calc_shipping_address_type'] ) ) : null;

				if ( $address_type !== null ) {
					WC()->session->set( 'address_type', $address_type );

					// Clear shipping rates cache to force shipping re-calculation
					$packages = WC()->cart->get_shipping_packages();
					foreach ( $packages as $key => $value ) {
						$shipping_session = 'shipping_for_package_' . $key;
						unset( WC()->session->$shipping_session );
					}
				}
			}


			/**
			 * Add address type tickbox to Checkout after Shipping Address
			 */
			public function myfreightshipping_address_type_checkout( $fields ) {
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$address_type_purpose = ( substr( $settings['delivery_address_type'], -10 ) == 'Commercial' ) ? 'Residential' : 'Commercial';
				$address_type_query = 'Is delivery address ' . $address_type_purpose . '?';
				$address_type_field = array(
					'address_type' => array(
						'type'         => 'checkbox',
						'label'        => '<span class="checkout_address_type_query">' . __($address_type_query, 'woocommerce') . '</span>',
						'autocomplete' => 'address_type',
						'required'     => false,
						'class'        => array( 'checkout_address_type_tickbox' ),
						'input-class'  => array( 'my-custom-class', 'checkout_address_type' )
					)
				);
				if ( !isset($fields['order']) ) {
					$fields['order'] = array();
				}
				$fields['order'] = $address_type_field + $fields['order'];
				return $fields;
			}


			/**
			 * Return address type value from session cache
			 */
			public function myfreightshipping_address_type_value( $value, $input ) {
				if ( $input == 'address_type' ) {
					return WC()->session->get( 'address_type', 0 );
				}
				return $value;
			}


			/**
			 * Handle address type value changes via incoming AJAX call
			 */
			public function myfreightshipping_address_type_ajax() {
				if ( isset($_POST['address_type']) ) {
					WC()->session->set( 'address_type', esc_attr( $_POST['address_type'] ) );
					echo $_POST['address_type'];

					// Clear shipping rates cache to force shipping re-calculation
					$packages = WC()->cart->get_shipping_packages();
					foreach ( $packages as $key => $value ) {
						$shipping_session = 'shipping_for_package_' . $key;
						unset( WC()->session->$shipping_session );
					}
				}
				die();
			}


			/**
			 * Save address type upon order
			 */
			public function myfreightshipping_address_type_create_order( $order, $data ) {
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				if ( substr( $settings['delivery_address_type'], 0, 7 ) == 'Enquire' ) {
					if ( isset($_POST['address_type']) && $_POST['address_type'] == 1 ) {
						$address_type_value = ( substr( $settings['delivery_address_type'], -10 ) == 'Commercial' ) ? 'Residential' : 'Commercial';
					} else {
						$address_type_value = ( substr( $settings['delivery_address_type'], -10 ) == 'Commercial' ) ? 'Commercial' : 'Residential';
					}
				} else {
					$address_type_value = 'n/a';
				}
				$order->update_meta_data( '_address_type', $address_type_value );
			}


			/**
			 * Display address type in order details shipping column
			 */
			public function myfreightshipping_address_type_order( $order ) {
				$address_type = wp_kses_post( get_post_meta( $order->get_id(), '_address_type', true ) );
				echo '<div class="address"><p><strong style="display:inline-block">Address Type:</strong> ' . $address_type . '</p></div>';
				$address_type_select_field = array(
					'id'            => '_address_type',
					'label'         => 'Address Type',
					'class'         => 'address_type',
//					'wrapper_class' => 'address_type_container',
					'options'       => array( 'Residential' => 'Residential', 'Commercial' => 'Commercial', 'n/a' => 'n/a' ),
					'value'         => $address_type
				);
				echo '<div class="edit_address">';
				woocommerce_wp_select( $address_type_select_field );
				echo '</div>';
			}


			/**
			 * Update address type following admin order edit
			 */
			public function myfreightshipping_address_type_save( $post_id, $post ) {
				if ( !isset($_POST['_address_type']) ) {
					return;
				}
				$address_type_value = wc_clean( wp_unslash( $_POST['_address_type'] ) );

				if ( in_array( $address_type_value, array( 'Residential', 'Commercial', 'n/a' ) ) ) {
					$order = wc_get_order( $post_id );
					if ( $order && is_callable( array( $order, 'update_meta_data' ) ) ) {
						$order->update_meta_data( '_address_type', $address_type_value );
						$order->save();
					}
				}
			}


			/**
			 * Handle input from custom field in Cart shipping calculator -> receiver unloading capability
			 */
			public function shipping_calculator_unloading_capability() {
				$unloading_capability = isset( $_POST['calc_shipping_unloading'] ) ? wc_clean( wp_unslash( $_POST['calc_shipping_unloading'] ) ) : null;

				if ( $unloading_capability !== null ) {
					WC()->session->set( 'unloading_capability', $unloading_capability );

					// Clear shipping rates cache to force shipping re-calculation
					$packages = WC()->cart->get_shipping_packages();
					foreach ( $packages as $key => $value ) {
						$shipping_session = 'shipping_for_package_' . $key;
						unset( WC()->session->$shipping_session );
					}

					// Alternate method to clear shipping rates cache to force shipping re-calculation
//					$shipping_city = trim( WC()->customer->get_shipping_city( 'view' ) );
//					$shipping_city = ( $unloading_capability == 1 ) ? $shipping_city . ',' : trim( $shipping_city, ',' );
//					WC()->customer->set_shipping_city( $shipping_city );
				}
			}


			/**
			 * Add unloading capability tickbox to Checkout after Shipping Address
			 */
			public function myfreightshipping_unloading_capability_checkout( $fields ) {
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$unloading_query = !empty($settings['unloading_capability_text']) ? $settings['unloading_capability_text'] : 'Forklift or unloading facility at delivery address?';
				$unloading_capability_field = array(
					'unloading_capability' => array(
						'type'         => 'checkbox',
						'label'        => '<span class="checkout_unloading_capability_query">' . __($unloading_query, 'woocommerce') . '</span>',
						'autocomplete' => 'unloading_capability',
						'required'     => false,
						'class'        => array( 'checkout_unloading_capability_tickbox' ),
						'input-class'  => array( 'my-custom-class', 'checkout_unloading_capability' )
					)
				);
				if ( !isset($fields['order']) ) {
					$fields['order'] = array();
				}
				$fields['order'] = $unloading_capability_field + $fields['order'];
				return $fields;
			}


			/**
			 * Return unloading capability value from session cache
			 */
			public function myfreightshipping_unloading_capability_value( $value, $input ) {
				if ( $input == 'unloading_capability' ) {
					return WC()->session->get( 'unloading_capability', 0 );
				}
				return $value;
			}


			/**
			 * Handle unloading capability value changes via incoming AJAX call
			 */
			public function myfreightshipping_unloading_capability_ajax() {
				if ( isset($_POST['unloading_capability']) ) {
					WC()->session->set( 'unloading_capability', esc_attr( $_POST['unloading_capability'] ) );
					echo $_POST['unloading_capability'];

					// Clear shipping rates cache to force shipping re-calculation
					$packages = WC()->cart->get_shipping_packages();
					foreach ( $packages as $key => $value ) {
						$shipping_session = 'shipping_for_package_' . $key;
						unset( WC()->session->$shipping_session );
					}
				}
				die();
			}


			/**
			 * Save unloading capability upon order
			 */
			public function myfreightshipping_unloading_capability_create_order( $order, $data ) {
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				if ( $settings['unloading_capability_check'] == 'yes' ) {
					$unloading_value = ( isset($_POST['unloading_capability']) && $_POST['unloading_capability'] == 1 ) ? 'Yes' : 'No';
				} else {
					$unloading_value = 'n/a';
				}
				$order->update_meta_data( '_unloading_capability', $unloading_value );
			}


			/**
			 * Display unloading capability in order details shipping column
			 */
			public function myfreightshipping_unloading_capability_order( $order ) {
				$unloading_capability = wp_kses_post( get_post_meta( $order->get_id(), '_unloading_capability', true ) );
				echo '<div class="address"><p><strong style="display:inline-block">Unloading Capability:</strong> ' . $unloading_capability . '</p></div>';
//				echo '<div class="edit_address"><p class="form-field _unloading_capability_field "><label for="_unloading_capability">Unloading Capability</label><input type="text" class="short" name="_unloading_capability" id="_unloading_capability" value="' . wp_kses_post( $unloading_capability ) . '" placeholder="Yes or No"></p></div>';
				$unloading_select_field = array(
					'id'            => '_unloading_capability',
					'label'         => 'Unloading Capability',
					'class'         => 'unloading_capability',
//					'wrapper_class' => 'unloading_capability_container',
					'options'       => array( 'Yes' => 'Yes', 'No' => 'No', 'n/a' => 'n/a' ),
					'value'         => $unloading_capability
				);
				echo '<div class="edit_address">';
				woocommerce_wp_select( $unloading_select_field );
				echo '</div>';
			}


			/**
			 * Update unloading capability following admin order edit
			 */
			public function myfreightshipping_unloading_capability_save( $post_id, $post ) {
				if ( !isset($_POST['_unloading_capability']) ) {
					return;
				}
				$unloading_value = wc_clean( wp_unslash( $_POST['_unloading_capability'] ) );

				if ( in_array( $unloading_value, array( 'Yes', 'No', 'n/a' ) ) ) {
					$order = wc_get_order( $post_id );
					if ( $order && is_callable( array( $order, 'update_meta_data' ) ) ) {
						$order->update_meta_data( '_unloading_capability', $unloading_value );
						$order->save();
					}
				}
			}


			/**
			 * Add Phone Number field to Checkout Shipping Address
			 */
			public function myfreightshipping_shipping_phone_checkout( $fields ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$placeholder = !empty($settings['phone_placeholder_checkout']) ? $settings['phone_placeholder_checkout'] : '10 digits with area code';

//				if ( !isset($fields['shipping']['shipping_phone']) ) {			// If WooCommerce version does not support shipping phone
//					$fields['shipping']['shipping_phone'] = array(
//						'type'         => 'tel',
//						'label'        => __('Phone', 'woocommerce'),
//						'placeholder'  => _x($placeholder, 'placeholder', 'woocommerce'),
//						'autocomplete' => 'shipping tel-national',
//						'required'     => false,
//						'class'        => array( 'form-row-wide' ),
//						'clear'        => true,
//						'validate'     => array( 'phone' ),
//						'priority'     => 100
//					);
//				} else {
				if ( isset($fields['billing']['billing_phone']) ) {
					$fields['billing']['billing_phone']['placeholder'] = _x($placeholder, 'placeholder', 'woocommerce');
					$fields['billing']['billing_phone']['autocomplete'] = 'billing tel-national';
				}
				if ( isset($fields['shipping']['shipping_phone']) ) {
					$fields['shipping']['shipping_phone']['placeholder'] = _x($placeholder, 'placeholder', 'woocommerce');
					$fields['shipping']['shipping_phone']['autocomplete'] = 'shipping tel-national';
				}
				return $fields;
			}


			/**
			 * Add Shipping Phone Number field to my account Shipping Address Edit
			 */
			public function myfreightshipping_shipping_phone_shipping_fields( $address_fields, $country ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$placeholder = !empty($settings['phone_placeholder_checkout']) ? $settings['phone_placeholder_checkout'] : '10 digits with area code';

				if ( !isset($address_fields['shipping_phone']) ) {
					$address_fields['shipping_phone'] = array(
						'type'         => 'tel',
						'label'        => __('Phone', 'woocommerce'),
						'placeholder'  => _x($placeholder, 'placeholder', 'woocommerce'),
						'autocomplete' => 'shipping tel-national',
						'required'     => false,
						'class'        => array( 'form-row-wide' ),
						'clear'        => true,
						'validate'     => array( 'phone' ),
						'priority'     => 100
					);
				} else {
					$address_fields['shipping_phone']['placeholder'] = _x($placeholder, 'placeholder', 'woocommerce');
					$address_fields['shipping_phone']['autocomplete'] = 'shipping tel-national';
				}
				return $address_fields;
			}


			/**
			 * Retrieve Shipping Phone Number from order
			 * Handle both pre- and post-WooCommerce 5.6 when shipping phone field was introduced into core
			 */
			public function myfreightshipping_shipping_phone_order_address( $fields, $address_type, WC_Order $order ) {
				$order_id = $order->get_id();
				if ( $address_type == 'shipping' ) {
					$shipping_phone = get_post_meta( $order_id, '_shipping_phone', true );
					if ( empty($shipping_phone) ) {
						$billing_address = $order->get_address( 'billing' );
						$shipping_phone = $billing_address['phone'];
					}
					$fields['phone'] = $shipping_phone;
				}
				return $fields;
			}


			/**
			 * Add Phone Number field to Order Details Shipping Column
			 */
			public function myfreightshipping_shipping_phone_order_admin( $fields ) {
				if ( !isset($fields['phone']) ) {
					$fields['phone'] = array(
						'label' => __('Phone', 'woocommerce')
					);
				}
				return $fields;
			}
			/* DEPRECATED */
//			public function myfreightshipping_phone_number_shipping( $order ) {
//				if ( !is_callable( array( $order, 'get_shipping_phone' ) ) ) {		// If WooCommerce version does not support shipping phone
//					$shipping_phone = get_post_meta( $order->get_id(), '_shipping_phone', true );
//					echo '<div class="address"><p><strong>Shipping Phone:</strong> ' . wc_make_phone_clickable( wp_kses_post( $shipping_phone ) ) . '</p></div>';
//					echo '<div class="edit_address"><p class="form-field _shipping_phone_field "><label for="_shipping_phone">Shipping Phone</label><input type="text" class="short" name="_shipping_phone" id="_shipping_phone" value="' . wp_kses_post( $shipping_phone ) . '" placeholder="10 digits with area code"></p></div>';
//				}
//			}


			/**
			 * Insert transport carrier logo image into shipping method label
			 */
			public function myfreightshipping_carrier_logo( $label, $method ) {
				$method_meta_data = $method->get_meta_data();
				$service_code = !empty($method_meta_data['Service_Code']) ? $method_meta_data['Service_Code'] : null;
				if ( $service_code ) {
					$logo_url = plugins_url( 'img/carriers/' . $service_code . '.png', __FILE__ );
					$label = '<img src="' . $logo_url . '" alt="' . $service_code . '" class="carrier-logo carrier-logo-' . $service_code . '" />' . $label;
				}
				return $label;
			}


			/**
			 * clean_myfreight_locality function.
			 *
			 * @access class
			 * @param string $city
			 * @param string $state
			 * @param string $postcode
			 * @return array
			 */
			static function clean_myfreight_locality( $city, $state, $postcode, int $min_match_percent = 80 ) {

				global $myfreight_localities;

				$best_match = -1;	// Index of best matching postcode and city
				$best_count = -1;	// Number of matching characters for best match
				$best_percent = -1;	// Percentage of matching characters for best match
				$max_localities = count( $myfreight_localities );

				if ( $min_match_percent < 100 ) {
					for ( $i = 0; $i < $max_localities; $i++ ) {
						if ( $myfreight_localities[$i][0] == $postcode ) {						// If postcodes match
							$match_count = similar_text( $myfreight_localities[$i][1], $city, $match_percent );	// Compare cities
							if ( $match_count > $best_count ) {							// Save best match so far
								$best_match = $i;
								$best_count = $match_count;
								$best_percent = $match_percent;
							}
						} else if ( $best_match >= 0 ) {								// If no more matching postcodes
							break;
						}
					}
					if ( $best_match >= 0 && $myfreight_localities[$best_match][2] != $state ) {	// If states don't match, reduce match percentage
						$best_percent -= 25;
					}

					if ( $best_percent >= $min_match_percent || ( empty($city) && $best_match >= 0 && $myfreight_localities[$best_match][2] == $state ) ) {
									// If reasonably good match or city not specified (but postcode and state ok), use best matching address
						$clean_locality = array(
							'locality' => $myfreight_localities[$best_match][1],
							'region'   => $myfreight_localities[$best_match][2],
							'postcode' => $myfreight_localities[$best_match][0]
						);
					} else {				// If poor match, use entered address and probably fail shipping price lookup (unless new locality)
						$clean_locality = array(
							'locality' => $city,
							'region'   => $state,
							'postcode' => $postcode
						);
					}
				} else {					// If perfect match required, use entered address
					$clean_locality = array(
						'locality' => $city,
						'region'   => $state,
						'postcode' => $postcode
					);
				}
				return $clean_locality;
			}


			/**
			 * next_working_date function.
			 *
			 * @access class
			 * @param string $base_date ('yyyy-mm-dd' or 'today')
			 * @param int $days_ahead
			 * @param WC_ORDER $order
			 * @return string (date 'yyyy-mm-dd')
			 *
			 * Generate default pickup date
			 */
			static function next_working_date( $base_date = 'today', $days_ahead = 0, $order = null ) {

				$australian_holidays = '01/01 26/01 25/04 25/12 26/12';			// Australian national holidays except Easter

				$timezone_string = get_option( 'timezone_string' );
				date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );

				if ( is_callable( 'easter_date' ) ) {					// Check if easter_date() in PHP library
					$easter_date = easter_date() + 43200;				// Add 12 hours for system timezone anomolies
				} else if ( is_callable( 'easter_days' ) ) {				// Check if easter_days() in PHP library
					$easter_date = strtotime( date( 'Y-03-21' ) ) + ( easter_days() * 86400 ) + 43200;
				} else {
					$easter_date = false;
				}
				if ( $easter_date ) {
					$easter_friday = date( 'd/m', $easter_date - 172800 );
					$easter_sunday = date( 'd/m', $easter_date );
					$easter_monday = date( 'd/m', $easter_date + 86400 );
					$australian_holidays .= ' ' . $easter_friday . ' ' . $easter_sunday . ' ' . $easter_monday;
				}

				$time_stamp = strtotime( $base_date . ' 12:00:00' );
				$base_date = date( 'Y-m-d', $time_stamp );
				if ( (int)$days_ahead < 0 ) {
					$time_stamp += (int)$days_ahead * 86400;
				}
				$working_days = 0;
				for ( $days = 0; $days < 90; $days++ ) {
					$day_of_week = date( 'D', $time_stamp );
					$day_month = date( 'd/m', $time_stamp );
					if ( $day_of_week != 'Sat' && $day_of_week != 'Sun' && strpos( $australian_holidays, $day_month ) === false ) {	// If neither weekend nor national holiday
						if ( $working_days >= (int)$days_ahead ) {
							break;
						}
						++$working_days;
					}
					$time_stamp += 86400;		// Increment 1 day
				}

				/*
				 * Filter hook to allow custom next working day to be specified.
				 */
				return apply_filters( 'my_next_working_date', date( 'Y-m-d', $time_stamp ), $base_date, $days_ahead, $order );
			}


			/**
			 * Sort function allowing variable key to be specified
			 */
			static function packing_sorter( $key1, $key2, $key3 ) {
				return function ( $a, $b ) use ( $key1, $key2, $key3 ) {
					if ( $a[$key1] == $b[$key1] ) {				// Key 1 sorted lowest to highest
						if ( $a[$key2] == $b[$key2] ) {			// Key 2 sorted lowest to highest
							if ( $b[$key3] == $a[$key3] ) {		// Key 3 sorted lowest to highest
								return 0;
							}
							return ( $a[$key3] < $b[$key3] ) ? -1 : 1;
						}
						return ( $a[$key2] < $b[$key2] ) ? -1 : 1;
					}
					return ( $a[$key1] < $b[$key1] ) ? -1 : 1;
				};
			}


			private static $logger;		// Handle for debug logging

			/**
			 * rule_pack_items function.
			 *
			 * @access class
			 * @param array $shipping_items
			 * @param boolean $pricing
			 * @param boolean $packing_item
			 * @return array
			 *
			 * Combine items with packing unit types into containers.
			 * If pricing true then include reference and dangerous goods fields.
			 */
			static function rule_pack_items( $shipping_items, $pricing = false, $packing_item = false ) {

				if ( !$packing_item ) {
					return $shipping_items;
				}

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				$packing_rules = ( !empty($settings['packing_rules']) && $settings['packing_rules'][0] == '{' ) ? json_decode( $settings['packing_rules'], true ) : null;
				if ( empty($packing_rules) ) {					// If no packing rules found
					return $shipping_items;
				}

				/* Initialise debug */
				if ( $settings['debug'] == 'yes' ) {
					$timezone_string = get_option( 'timezone_string' );
					date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );
					self::$logger = wc_get_logger();
					self::$logger->debug( '==== Begin Rule Packing ====', array( 'source' => 'Myfreight Shipping' ) );
				} else {
					self::$logger = null;
				}
				if ( !empty( self::$logger ) ) {
					self::$logger->debug( 'SHIPPING ITEMS UNPACKED: ' . print_r( $shipping_items, true ), array( 'source' => 'Myfreight Shipping' ) );
				}

				$packing_units = array();
				foreach ( $shipping_items as &$item ) {				// Find all packing unit types
					$packing_unit = $item['packing_unit'];
					if ( !empty($packing_unit) && !in_array( $packing_unit, $packing_units ) ) {
						$packing_units[] = $packing_unit;
					}
				}
				unset( $item );

				if ( !empty($packing_units) ) {
					sort( $packing_units );					// Sort to ensure rules execute in consistent order
				}

				foreach ( $packing_units as &$packing_unit ) {			// For each packing unit type

					$unit_rules = array();
					foreach ( $packing_rules as $packing_rule ) {		// Get the rules for packing unit type
						if ( !empty($packing_rule[$packing_unit]) && (int)$packing_rule[$packing_unit] > 0 ) {
							if ( !empty($packing_rule['content_qty_min']) ) {
								if ( substr( $packing_rule['content_qty_min'], -1 ) == '%' ) {
									$packing_rule['content_qty_min'] = (int)ceil( (int)$packing_rule[$packing_unit] * (int)substr( $packing_rule['content_qty_min'], 0, -1 ) / 100 );
								} else if ( (int)$packing_rule['content_qty_min'] < 0 ) {
									$packing_rule['content_qty_min'] = (int)$packing_rule[$packing_unit] + (int)$packing_rule['content_qty_min'];
								}
							} else if ( !empty($packing_rule['container_max_empty_slots']) && (int)$packing_rule['container_max_empty_slots'] > 0 ) {
								/*** DEPRECATED ***/
								$packing_rule['content_qty_min'] = (int)$packing_rule[$packing_unit] - (int)$packing_rule['container_max_empty_slots'];
							} else {
								$packing_rule['content_qty_min'] = 1;
							}
							$unit_rules[] = $packing_rule;
						}
					}

					usort( $unit_rules, WC_Myfreight_Ancillary::packing_sorter( $packing_unit, 'content_qty_min', 'content_weight_max' ) );	// Sort rules by container capacity

					$total_quantity = 0;
					$total_weight = 0.0;
					$max_item_length = 0;
					$max_item_width = 0;
					$max_item_height = 0;
					$item_fragile = false;
					$dangerous_goods_items = array();
					$reference = null;

					/* Aggregate items for this packing unit type */
					foreach ( $shipping_items as $item_id => &$item ) {
						if ( !empty($item['packing_unit']) && $item['packing_unit'] == $packing_unit ) {
							$total_quantity += (int)$item['quantity'];
							$total_weight   += (float)$item['weight_in_kilograms'];
							$max_item_length = max( $max_item_length, max( (int)$item['length'], (int)$item['width'] ) );
							$max_item_width  = max( $max_item_width,  min( (int)$item['length'], (int)$item['width'] ) );
							$max_item_height = max( $max_item_height, (int)$item['height'] );
							$item_fragile    = ( $item_fragile || !empty($item['fragile']) );

							if ( !empty($item['dangerous_goods_items']) ) {		// Merge DG items of same classification
								if ( empty($dangerous_goods_items) || empty($dangerous_goods_items[0]) ) {
									$dangerous_goods_items = $item['dangerous_goods_items'];
								} else {
									for ( $i = 1; $i < 10; $i++ ) {
										if ( empty($dangerous_goods_items[$i]) ) {
											$dangerous_goods_items[$i] = $item['dangerous_goods_items'][0];
											break;
										} else if ( $dangerous_goods_items[$i]['un_number'] == $item['dangerous_goods_items'][0]['un_number'] &&
											    $dangerous_goods_items[$i]['packing_group'] == $item['dangerous_goods_items'][0]['packing_group'] ) {
											$dangerous_goods_items[$i]['aggregate_quantity'] += $item['dangerous_goods_items'][0]['aggregate_quantity'];
											break;
										}
									}
								}
							}
							if ( empty($reference) && !empty($item['reference']) ) {
								$reference = $item['reference'];
							}
						}
					}
					unset( $item );
					if ( empty($dangerous_goods_items) ) {
						$dangerous_goods_items = null;
					}

					$remaining_quantity = $total_quantity;
//					$remaining_weight = $total_weight;
					$unit_weight = ( $total_weight > 0.0 && $total_quantity > 0 ) ? ( $total_weight / $total_quantity ) : 1.0;

					/* Look for capacity in previously packed containers */
					foreach ( $shipping_items as $item_id => &$item ) {	// Check all items including previously packed containers
						if ( empty($dangerous_goods_items) ) {
							if ( !empty($item['container_code']) ) {	// If item is a packed container
								$packing_rule_key = array_search( $item['container_code'], array_column( $unit_rules, 'container_code' ) );
								if ( $packing_rule_key !== false ) {
									$container_progress_units = 0;				// Initialise items in container
									$max_units = isset($unit_rules[$packing_rule_key][$packing_unit]) ? (int)($unit_rules[$packing_rule_key][$packing_unit]) : 0;
									$max_weight = !empty($unit_rules[$packing_rule_key]['content_weight_max']) ? (float)($unit_rules[$packing_rule_key]['content_weight_max']) : 1000.0;
									$available_weight = ( (int)$item['quantity'] * $max_weight ) - $item['weight_in_kilograms'];
									$max_units_for_weight = min( (int)$item['quantity'] * $max_units, (int)floor( $available_weight / $unit_weight ) );

									$packing_rule_item_type = $unit_rules[$packing_rule_key]['item_type'];
//									$item['length'] = max( $item['length'], $max_item_length );
//									$item['width']  = max( $item['width'], $max_item_width );
									if ( $packing_rule_item_type == 'Skid' || $packing_rule_item_type == 'Pallet' ) {
										$item['height'] = max( (int)$item['height'], (int)$unit_rules[$packing_rule_key]['container_height'] + $max_item_height );
									} else {
										$item['height'] = max( (int)$item['height'], $max_item_height );
									}
									$item['fragile'] = ( !empty($item['fragile']) || $item_fragile );

									if ( $remaining_quantity > $max_units_for_weight ) {	// Check applicability of container
										$item['weight_in_kilograms'] += $max_units_for_weight * $unit_weight;
										$container_progress_units += $max_units_for_weight;
										$remaining_quantity -= $max_units_for_weight;
									} else {
										$item['weight_in_kilograms'] += $remaining_quantity * $unit_weight;
										$container_progress_units += $remaining_quantity;
										$remaining_quantity = 0;
										break;
									}
								}
							}
						}
					}
					unset( $item );

					/* Add new containers of reducing size and pack them */
					for ( $i = count($unit_rules)-1; $i >= 0 && $remaining_quantity > 0; $i-- ) {		// For each packing unit container capacity
						$unit_rule = $unit_rules[$i];
						$container_progress_units = 0;				// Initialise items in container
//						$container_progress_weight = 0.0;			// Initialise weight in container
						$container_count = 0;					// Initialise number of containers
						$max_units = isset($unit_rule[$packing_unit]) ? (int)($unit_rule[$packing_unit]) : 0;
						$min_units = !empty($unit_rule['content_qty_min']) ? (int)($unit_rule['content_qty_min']) : 1;
						$max_weight = !empty($unit_rule['content_weight_max']) ? (float)($unit_rule['content_weight_max']) : 1000.0;
						$min_weight = !empty($unit_rule['content_weight_min']) ? (float)($unit_rule['content_weight_min']) : 0.0;
						$max_units_for_weight = min( $max_units, (int)floor( $max_weight / $unit_weight ) );
						$min_units_for_weight = min( $min_units, $max_units_for_weight );
						$max_length = !empty($unit_rule['container_length']) ? (int)($unit_rule['container_length']) : 0;

						while ( $remaining_quantity >= $min_units_for_weight ) {
							if ( $remaining_quantity > $max_units_for_weight ) {	// Check applicability of container
								$container_progress_units += $max_units_for_weight;
								$remaining_quantity -= $max_units_for_weight;
//								$container_progress_weight += $max_units_for_weight * $unit_weight;
//								$remaining_weight -= $max_units_for_weight * $unit_weight;
							} else {
								if ( $i > 0 ) {					// If another (better) rule ahead
									$next_unit_rule = $unit_rules[$i-1];
									$next_max_units = isset($next_unit_rule[$packing_unit]) ? (int)($next_unit_rule[$packing_unit]) : 0;
									$next_min_units = !empty($next_unit_rule['content_qty_min']) ? (int)($next_unit_rule['content_qty_min']) : 1;
									$next_max_weight = !empty($next_unit_rule['content_weight_max']) ? (float)($next_unit_rule['content_weight_max']) : 1000.0;
									$next_min_weight = !empty($next_unit_rule['content_weight_min']) ? (float)($next_unit_rule['content_weight_min']) : 0.0;
									$next_max_units_for_weight = min( $next_max_units, (int)floor( $max_weight / $unit_weight ) );
									$next_min_units_for_weight = min( $next_min_units, $next_max_units_for_weight );
									$next_max_length = !empty($next_unit_rule['container_length']) ? (int)($next_unit_rule['container_length']) : EMPTY_PALLET_LENGTH;

									if ( $remaining_quantity >= $next_min_units_for_weight && $remaining_quantity <= $next_max_units_for_weight ) {	// If next rule is a fuller fit
										if ( $max_item_length <= $next_max_length || $max_item_length > $max_length ) {		// If item not too big for next rule's container
											continue 2;		// Jump to next rule
										}
									}
								}
								$container_progress_units += $remaining_quantity;
								$remaining_quantity = 0;
//								$container_progress_weight += $remaining_weight;
//								$remaining_weight = 0;
							}
							++$container_count;				// Count number of containers required
						}
						if ( $container_count > 0 ) {				// Create new container item
							$item_type      = $unit_rule['item_type'];
							$weight         = (float)($unit_rule['container_weight_empty']);
							$length         = (int)($unit_rule['container_length']);
							$width          = (int)($unit_rule['container_width']);
							$height         = (int)($unit_rule['container_height']);
							$height_decimal = (float)($unit_rule['container_height']);
							$container_code = $unit_rule['container_code'];

							$weight_in_kilograms = $total_weight * $container_progress_units / $total_quantity;	// Calculate content weight for container
							if ( !empty($weight) ) {
								$weight_in_kilograms += $weight * $container_count;	// Add container weight
							}
							$weight_in_kilograms = round( $weight_in_kilograms, 1 );	// Round weight to 1 decimal place

							if ( $item_type == 'Skid' || $item_type == 'Pallet' ) {		// Calculate height of skid or pallet
								$height += $max_item_height;
								$height_decimal += $max_item_height;
							}

							if ( $pricing ) {
								$container_item = array(			// Create new container item
									'item_type_description' => $item_type,
									'quantity'              => $container_count,
									'weight_in_kilograms'   => $weight_in_kilograms,
									'length'                => $length ? $length : $max_item_length,
									'width'                 => $width  ? $width  : $max_item_width,
									'height'                => $height ? $height : $max_item_height,
									'height_decimal'        => (bool)$height_decimal ? $height_decimal : (float)$max_item_height,
									'fragile'               => $item_fragile,
									'container_code'        => $container_code
								);
							} else {
								$container_item = array(			// Create new container item
									'reference'             => $reference,
									'item_type_description' => $item_type,
									'quantity'              => $container_count,
									'weight_in_kilograms'   => $weight_in_kilograms,
									'length'                => $length ? $length : $max_item_length,
									'width'                 => $width  ? $width  : $max_item_width,
									'height'                => $height ? $height : $max_item_height,
									'height_decimal'        => (bool)$height_decimal ? $height_decimal : (float)$max_item_height,
									'fragile'               => $item_fragile,
									'dangerous_goods_items' => $dangerous_goods_items,
									'container_code'        => $container_code
								);
							}
							array_unshift( $shipping_items, $container_item );	// Add container item to shipping items
						}
					}
					foreach ( $shipping_items as $item_id => &$item ) {	// Delete packed items (and leave unpacked items)
						if ( !empty($item['packing_unit']) && $item['packing_unit'] == $packing_unit ) {
							if ( $remaining_quantity < $total_quantity ) {
								$remaining_quantity += (int)$item['quantity'];
								unset( $shipping_items[$item_id] );
							} else {
								break;				// Exit if finished
							}
						}
					}
					unset( $item );
				}
				unset( $packing_unit );

				$shipping_items = array_values( $shipping_items );		// Re-key array for correct JSON encoding

				if ( !empty( self::$logger ) ) {
					self::$logger->debug( 'SHIPPING ITEMS RULE PACKED: ' . print_r( $shipping_items, true ), array( 'source' => 'Myfreight Shipping' ) );
				}

				return $shipping_items;
			}


			/**
			 * auto_pack_items function.
			 *
			 * @access class
			 * @param array $shipping_items
			 * @param boolean $pricing
			 * @param boolean $packing_item
			 * @return array
			 *
			 * Combine multiple items into minuimum number of containers.
			 * If pricing true then include reference and dangerous goods fields.
			 */
			static function auto_pack_items( $shipping_items, $pricing = false, $packing_item = false ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				/* Initialise debug */
				if ( $settings['debug'] == 'yes' ) {
					$timezone_string = get_option( 'timezone_string' );
					date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );
					self::$logger = wc_get_logger();
					self::$logger->debug( '==== Begin Auto Packing ====', array( 'source' => 'Myfreight Shipping' ) );
				} else {
					self::$logger = null;
				}
				if ( !empty( self::$logger ) ) {
					self::$logger->debug( 'SHIPPING ITEMS UNPACKED: ' . print_r( $shipping_items, true ), array( 'source' => 'Myfreight Shipping' ) );
				}

				$packing_auto_include_rule_packed = ( $settings['packing_auto_include_rule_packed'] == 'yes' );
				$packing_auto_min_qty = !empty($settings['packing_auto_min_qty']) ? (int)$settings['packing_auto_min_qty'] : 2;
				$packing_auto_max_container_weight = !empty($settings['packing_auto_max_container_weight']) ? (float)$settings['packing_auto_max_container_weight'] : 1000.0;

				$total_quantity = 0;
				$total_weight = 0.0;
				$total_volume = 0;
				$max_item_length = 0;
				$max_item_width = 0;
				$max_item_height = 0;
				$item_fragile = false;
				$dangerous_goods_items = array();
				$reference = null;

				foreach ( $shipping_items as $item_id => &$item ) {		// Delete items with no weight or dimensions
					if ( empty($item['quantity']) || empty($item['weight_in_kilograms']) || empty($item['length']) || empty($item['width']) || empty($item['height']) ) {
						unset( $shipping_items[$item_id] );
					}
				}
				unset( $item );

				foreach ( $shipping_items as $item_id => &$item ) {		// Calculate nett item weight and volume
					if ( !$packing_auto_include_rule_packed && !empty($item['container_code']) ) {		// If item has already been rule packed
						continue;									// Do not include in auto packing
					}
					if ( ( (float)$item['weight_in_kilograms'] / (int)$item['quantity'] ) >= $packing_auto_max_container_weight ) {	// If item exceeds container maximum weight threshold
						$shipping_items[$item_id]['container_code'] = 'Autopack';						// Do not include in auto packing
						continue;
					}
					$total_quantity += (int)$item['quantity'];
					$total_weight   += (float)$item['weight_in_kilograms'];
					$total_volume   += (int)((int)$item['quantity'] * (int)$item['length'] * (int)$item['width'] * ( isset($item['height_decimal']) ? (float)$item['height_decimal'] : (int)$item['height'] ));
					$max_item_length = max( $max_item_length, max( (int)$item['length'], (int)$item['width'] ) );
					$max_item_width  = max( $max_item_width,  min( (int)$item['length'], (int)$item['width'] ) );
					$max_item_height = max( $max_item_height, (int)$item['height'] );
					$item_fragile    = ( $item_fragile || !empty($item['fragile']) );

					if ( empty($reference) && !empty($item['reference']) ) {			// Choose one reference
						$reference = $item['reference'];
					}
					if ( !empty($item['dangerous_goods_items']) ) {					// Aggregate dangerous goods data
						$dangerous_goods_items = array_merge( $dangerous_goods_items, $item['dangerous_goods_items'] );
					}
				}
				unset( $item );

				if ( $total_quantity < $packing_auto_min_qty || $total_weight <= 0.0 || $total_volume <= 0 ) {
					return $shipping_items;
				}

				$packing_auto_gap_percentage = isset($settings['packing_auto_gap_percentage']) ? (int)$settings['packing_auto_gap_percentage'] : PACKING_GAP_PERCENT;
				$packing_gap_variance = ( 100.0 + $packing_auto_gap_percentage ) / 100;

				/* Calculate number of containers required based on max weight per container or excess volume */
				$container_count = (int)ceil( max( $total_weight / $packing_auto_max_container_weight,
								   (float)$total_volume * $packing_gap_variance / ($packing_auto_max_container_weight * 100000),
								   (float)$total_volume * $packing_gap_variance / (EMPTY_PALLET_LENGTH * EMPTY_PALLET_WIDTH * (MAX_PALLET_HEIGHT - EMPTY_PALLET_HEIGHT)) ) );

				$item_weight = $total_weight / $container_count;
				$item_volume = $total_volume / $container_count;
				$item_length = max( $max_item_length, (int)ceil( pow( $item_volume, 1/3 ) ) );			// Even dimensions but long enough to contain items
				$item_width = min( max( $max_item_width, (int)ceil( sqrt( $item_volume / $item_length ) ) ), (int)EMPTY_PALLET_WIDTH );
																// Wide enough to contain items but no wider than a pallet
				$item_height = (int)ceil( $item_volume / ( $item_length * $item_width ) );			// Height determined by length and width

				/*
				 * Filter hook for custom estimation of shipping item type (Satchel, Carton, Skid or Pallet).
				 */
				$item_type = apply_filters( 'my_est_item_type', 'Carton', $total_quantity, $item_weight, $item_length, $item_width, $item_height, null, $settings['account_type'] );

				if ( $item_type == 'Carton' ) {
					$item_height = (int)ceil( $item_height * $packing_gap_variance );		// Calculate height assuming gap
					$item_height = max( $item_height, $max_item_height );				// Adjust if tall item
				} else if ( $item_type == 'Skid' ) {
					$total_weight += $container_count * (float)EMPTY_SKID_WEIGHT;			// Add skid weight
					$item_length = max( $item_length, $max_item_length, EMPTY_SKID_LENGTH );	// Estimate skid length
					$item_width = max( $item_width, $max_item_width, EMPTY_SKID_WIDTH );		// Estimate skid width
					$item_height = (int)ceil( ($item_volume * $packing_gap_variance) / ($item_length * $item_width) );	// Calculate height assuming skid gap
					$item_height = max( $item_height, $max_item_height ) + EMPTY_SKID_HEIGHT;	// Add skid height
				} else if ( $item_type == 'Pallet' ) {
					$total_weight += $container_count * (float)EMPTY_PALLET_WEIGHT;			// Add pallet weight
					$item_length = max( $max_item_length, EMPTY_PALLET_LENGTH );			// Estimate pallet length
					$item_width = EMPTY_PALLET_WIDTH;						// Estimate pallet width
					$item_height = (int)ceil( ($item_volume * $packing_gap_variance) / ($item_length * $item_width) );	// Calculate height assuming pallet gap
					$item_height = max( $item_height, $max_item_height ) + EMPTY_PALLET_HEIGHT;	// Add pallet height
				}
				$item_height_decimal = (float)$item_height;

				if ( count( $dangerous_goods_items ) == 0 ) {
					$dangerous_goods_items = null;
				}

				if ( $pricing ) {
					$container_item = array(		// Create new container item
						'item_type_description' => $item_type,
						'quantity'              => $container_count,
						'weight_in_kilograms'   => $total_weight,
						'length'                => $item_length,
						'width'                 => $item_width,
						'height'                => $item_height,
						'height_decimal'        => $item_height_decimal,
						'fragile'               => $item_fragile,
						'container_code'        => 'Autopack'
					);
				} else {
					$container_item = array(		// Create new container item
						'reference'             => $reference,
						'item_type_description' => $item_type,
						'quantity'              => $container_count,
						'weight_in_kilograms'   => $total_weight,
						'length'                => $item_length,
						'width'                 => $item_width,
						'height'                => $item_height,
						'height_decimal'        => $item_height_decimal,
						'fragile'               => $item_fragile,
						'dangerous_goods_items' => $dangerous_goods_items,
						'container_code'        => 'Autopack'
					);
				}
				array_unshift( $shipping_items, $container_item );		// Add container item to shipping items

				foreach ( $shipping_items as $item_id => &$item ) {		// Delete packed items (and leave unpacked items)
					if ( empty($item['container_code']) || ( $packing_auto_include_rule_packed && $item['container_code'] != 'Autopack' ) ) {
						unset( $shipping_items[$item_id] );
					}
				}
				unset( $item );

				$shipping_items = array_values( $shipping_items );		// Re-key array for correct JSON encoding

				if ( !empty( self::$logger ) ) {
					self::$logger->debug( 'SHIPPING ITEMS AUTO PACKED: ' . print_r( $shipping_items, true ), array( 'source' => 'Myfreight Shipping' ) );
				}

				return $shipping_items;
			}


			/**
			 * estimate_item_type function.
			 *
			 * @access class
			 * @param string $my_item_type
			 * @param int $quantity
			 * @param float $weight
			 * @param int $length
			 * @param int $width
			 * @param int $height
			 * @param string $account = Myfreight | Myparcel
			 * @return string
			 */
			static function estimate_item_type( $my_item_type, $quantity, $weight, $length, $width, $height, $shipping_class, $account ) {

				if ( substr( $shipping_class, 0, 5 ) == 'item-' ) {			// If product shipping class slug starts with 'item-'
					$shipping_classname = get_term_by( 'slug', $shipping_class, 'product_shipping_class' );
					if ( $shipping_classname != false && !empty($shipping_classname->name) ) {
						return $shipping_classname->name;			// Use the name of the shipping class as item type
					}
				}
					
				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				if ( $settings['item_type_auto_active'] == 'yes' ) {
					$item_types = array (						// Item Types maximum weight and dimensions
						array( 'account' => 'Myfreight', 'type' => '500g Satchel', 'weight' => 0.5,    'length' => 35,  'width' => 22,  'height' => 6 ),
						array( 'account' => 'Myfreight', 'type' => '1kg Satchel',  'weight' => 1.0,    'length' => 39,  'width' => 27,  'height' => 7 ),
						array( 'account' => 'Myfreight', 'type' => '3kg Satchel',  'weight' => 3.0,    'length' => 42,  'width' => 31,  'height' => 9 ),
						array( 'account' => 'Myfreight', 'type' => '5kg Satchel',  'weight' => 5.0,    'length' => 59,  'width' => 44,  'height' => 12 ),
						array( 'account' => 'Myfreight', 'type' => 'Satchel',      'weight' => 3.0,    'length' => 42,  'width' => 30,  'height' => 8 ),
						array( 'account' => 'Myfreight', 'type' => 'Carton',       'weight' => 80.0,   'length' => 100, 'width' => 60,  'height' => 120 ),
						array( 'account' => 'Myfreight', 'type' => 'Skid',         'weight' => 400.0,  'length' => 120, 'width' => 80,  'height' => 160 ),
						array( 'account' => 'Myfreight', 'type' => 'Pallet',       'weight' => 2000.0, 'length' => 300, 'width' => 240, 'height' => 300 ),
						array( 'account' => 'Myparcel',  'type' => 'Satchel',      'weight' => 3.0,    'length' => 42,  'width' => 30,  'height' => 8 ),
						array( 'account' => 'Myparcel',  'type' => 'Carton',       'weight' => 80.0,   'length' => 100, 'width' => 60,  'height' => 120 ),
						array( 'account' => 'Myparcel',  'type' => 'Skid',         'weight' => 400.0,  'length' => 120, 'width' => 80,  'height' => 160 ),
						array( 'account' => 'Myparcel',  'type' => 'Pallet',       'weight' => 2000.0, 'length' => 300, 'width' => 240, 'height' => 300 )
					);

					$est_weight = (float)$weight;
					$est_length = max($length, $width);
					$est_width  = min($length, $width);
					$est_height = $height;

					foreach ( $item_types as $item_type ) {				// Find smallest matching item type
						if ( ( empty($item_type['account']) ||
							   ( $item_type['account'] == $account && $settings[strtolower($account) . '_' . str_replace(' ', '', $item_type['type'])] == 'yes' ) ) &&
						     $est_weight <= $item_type['weight'] &&
						     $est_length <= $item_type['length'] &&
						     $est_width  <= $item_type['width']  &&
						     $est_height <= $item_type['height'] ) {
							return $item_type['type'];
						}
					}
				}

				return !empty($my_item_type) ? $my_item_type : 'Carton';		// If nothing matches and no default assume Carton
			}


			/**
			 * receiver_residential function.
			 *
			 * @access class
			 * @param bool $residential
			 * @param string $address_type_setting
			 * @param string|bool $company
			 * @param string|bool $address_type
			 * @return bool
			 *
			 * Decide whether the receiver address is residential based on available information.
			 * Method of deciding is determined by "Delivery Address Type" plugin admin setting.
			 * If presence/absence of company is to be used, method is different for pricing and ordering.
			 * For pricing, check if user is logged in and (shipping or billing) address completed and whether Company is provided.
			 * For order, use order address (shipping or billing) Company provided in $company parameter.
			 */
			static function receiver_residential( $residential, $address_type_setting, $company = false, $address_type = false ) {

				if ( $address_type_setting == 'Residential' ) {
					return true;
				} else if ( $address_type_setting == 'Commercial' ) {
					return false;
				} else if ( substr( $address_type_setting, 0, 7 ) == 'Company' ) {
					if ( $company !== false ) {
						return empty( $company );
					} else {
						$user_id = get_current_user_id();
						if ( $user_id > 0 ) {
							if ( !empty( get_user_meta( $user_id, 'shipping_city', true ) ) ) {
								return empty( get_user_meta( $user_id, 'shipping_company', true ) );
							} else if ( !empty( get_user_meta( $user_id, 'billing_city', true ) ) ) {
								return empty( get_user_meta( $user_id, 'billing_company', true ) );
							}
						}
						if ( $address_type_setting == 'Company Residential' ) {
							return true;
						} else if ( $address_type_setting == 'Company Commercial' ) {
							return false;
						}
					}
				} else if ( substr( $address_type_setting, 0, 7 ) == 'Enquire' ) {
					if ( !empty($address_type) ) {
						return ( $address_type == 'Residential' );
					} else {
						return ( substr( $address_type_setting, 8 ) == 'Residential' );
					}
				}
				return $residential;
			}


			/**
			 * get_carrier_surcharges function.
			 *
			 * @access class
			 * @param array $carrier_surcharges
			 * @param string $carrier
			 * @param string $service_code
			 * @param array $all_surcharges
			 * @return array
			 *
			 * Extract the service availability and surcharges data for the specified carrier and optional service code from the $all_surcharges list.
			 * If the value is false, the service option is not available for that carrier service.
			 * If no values are available for a carrier or service code, the defaults will be returned.
			 */
			static function get_carrier_surcharges( $carrier_surcharges, $carrier, $service_code, $all_surcharges ) {

				$surcharge_types = array( 'constraint',
							  'residential',
							  'manual_handling',
							  'tailgate_lift',
							  'dangerous_goods',
							  'minimum_shipping',
							  'minimum_total'
							);

				foreach ( $surcharge_types as $surcharge_type ) {
					$carrier_surcharges[$surcharge_type] = 0;							// Default to service available and surcharge = $0
					if ( isset($all_surcharges['default'][$surcharge_type]) ) {
						$carrier_surcharges[$surcharge_type] = $all_surcharges['default'][$surcharge_type];			// Get overall default data
					}
					if ( !empty($carrier) && isset($all_surcharges[$carrier][$surcharge_type]) ) {
						$carrier_surcharges[$surcharge_type] = $all_surcharges[$carrier][$surcharge_type];			// Get carrier default data
					}
					if ( !empty($carrier) && !empty($service_code) && isset($all_surcharges[$carrier][$service_code][$surcharge_type]) ) {
						$carrier_surcharges[$surcharge_type] = $all_surcharges[$carrier][$service_code][$surcharge_type];	// Get carrier service data
					}
				}
				return $carrier_surcharges;
			}


			/**
			 * sort_quotes_by_cost function.
			 *
			 * @access class
			 * @param array $prices
			 * @param string $sender_code
			 * @return array
			 *
			 * Sort quotes by cost to customer (then delivery date).
			 * Sender code (warehouse id) is available to assist in sorting or to modify cost by warehouse if required.
			 */
			static function sort_quotes_by_cost( $prices, $sender_code ) {
				usort( $prices, function ($a, $b) {
					if ( !isset($a['cost']) || !isset($b['cost']) ) {
						return 0;
					}
					if ( $a['cost'] == $b['cost'] ) {							// Cost sorted lowest to highest
						if ( !isset($a['estimated_delivery_date']) || !isset($b['estimated_delivery_date']) ) {
							return 0;
						}
						if ( $a['estimated_delivery_date'] == $b['estimated_delivery_date'] ) {		// ETA sorted earliest to lastest
							return 0;
						}
						return ( $a['estimated_delivery_date'] < $b['estimated_delivery_date'] ) ? -1 : 1;
					}
					return ( (float)$a['cost'] < (float)$b['cost'] ) ? -1 : 1;
				});
				return $prices;
			}

		}
	}
	// Instantiate class
	new WC_Myfreight_Ancillary();

}

