<?php

if ( ! defined( 'ABSPATH' ) ) {
	exit;	// Exit if accessed directly
}

/**
 * Check if WooCommerce is active
 */
if ( in_array( 'woocommerce/woocommerce.php', apply_filters( 'active_plugins', get_option( 'active_plugins' ) ) ) ) {

	/**
	 * Persistent class for order consignment creation functions
	 */
	if ( ! class_exists( 'WC_Myfreight_Order' ) ) {
		class WC_Myfreight_Order {

			public function __construct() {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				// Add dangerous goods fields to product shipping tab and variations
				if ( $settings['account_type'] == 'Myfreight' ) {
					// Load dangerous goods scripts
					add_action( 'admin_enqueue_scripts', array( $this, 'myfreightshipping_enqueue_admin_dg_scripts' ), 10, 1 );

					add_action( 'woocommerce_product_options_shipping', array( $this, 'myfreightshipping_dangerous_goods_fields' ), 10, 1 );
					add_action( 'woocommerce_process_product_meta', array( $this, 'myfreightshipping_dangerous_goods_fields_save' ), 10, 1 );
					add_action( 'woocommerce_variation_options_dimensions', array( $this, 'myfreightshipping_dangerous_goods_variation_field' ), 10, 3 );
					add_action( 'woocommerce_save_product_variation', array( $this, 'myfreightshipping_dangerous_goods_variation_field_save' ), 10, 2 );
					add_filter( 'woocommerce_available_variation', array( $this, 'myfreightshipping_dangerous_goods_variation_data' ), 10, 3 );
					add_filter( 'woocommerce_get_cart_contents', array( $this, 'myfreightshipping_dangerous_goods_cart_contents' ), 10, 1 );
				}
				// Add packing unit field to product shipping tab and variations
				if ( $settings['packing_rules_active'] == 'yes' ) {
					add_action( 'woocommerce_product_options_shipping', array( $this, 'myfreightshipping_packing_unit_field' ), 10, 1 );
					add_action( 'woocommerce_process_product_meta', array( $this, 'myfreightshipping_packing_unit_field_save' ), 10, 1 );
					add_action( 'woocommerce_variation_options_dimensions', array( $this, 'myfreightshipping_packing_unit_variation_field' ), 10, 3 );
					add_action( 'woocommerce_save_product_variation', array( $this, 'myfreightshipping_packing_unit_variation_field_save' ), 10, 2 );
					add_filter( 'woocommerce_available_variation', array( $this, 'myfreightshipping_packing_unit_variation_data' ), 10, 3 );
				}
				// Add sender address code field to product shipping tab
				if ( $settings['multiple_senders'] == 'yes' ) {
					add_action( 'woocommerce_product_options_shipping', array( $this, 'myfreightshipping_sender_code_field' ), 10, 1 );
					add_action( 'woocommerce_process_product_meta', array( $this, 'myfreightshipping_sender_code_field_save' ), 10, 1 );
				}
				// Create and track consignment on successful order
				if ( $settings['auto_consigning'] == 'yes' ) {
					// Auto generate consignment when order status changes to processing
					add_action( 'woocommerce_order_status_changed', array( $this, 'myfreightshipping_order_status_change' ), 10, 4 );

					// Create admin order tracking box
					add_action( 'woocommerce_admin_order_items_after_shipping', array( $this, 'myfreightshipping_order_tracking_box' ), 10, 1 );

					// Add action to the orders list page (ship or view)
					add_action( 'woocommerce_admin_order_actions_end', array( $this, 'myfreightshipping_order_list_action_button' ), 10, 1 );
					add_action( 'admin_head', array( $this, 'myfreightshipping_order_list_action_button_css' ), 10, 0 );

					if ( $settings['email_completed_shipping'] == 'yes' ) {
						// Add shipping info to Completed order email
						add_action( 'woocommerce_email_before_order_table', array( $this, 'myfreightshipping_email_completed_shipping' ), 10, 4 );
					}
				}
				// Allow admin manual consignment creation
				if ( is_admin() ) {
					// Add selections to the order actions menu
					add_action( 'woocommerce_order_actions', array( $this, 'add_order_meta_box_actions' ) );
					// Process the custom order actions
					add_action( 'woocommerce_order_action_regenerate_shipping_consignment', array( $this, 'myfreightshipping_regenerate_consignment' ), 10, 1 );
					add_action( 'woocommerce_order_action_remove_shipping_consignment',     array( $this, 'myfreightshipping_remove_consignment' ), 10, 1 );
					add_action( 'woocommerce_order_action_tracking_status_booked',          array( $this, 'myfreightshipping_tracking_status_booked' ), 10, 1 );
					add_action( 'woocommerce_order_action_tracking_status_in_transit',      array( $this, 'myfreightshipping_tracking_status_in_transit' ), 10, 1 );
					add_action( 'woocommerce_order_action_tracking_status_delivering',      array( $this, 'myfreightshipping_tracking_status_delivering' ), 10, 1 );
					add_action( 'woocommerce_order_action_tracking_status_failed_delivery', array( $this, 'myfreightshipping_tracking_status_failed_delivery' ), 10, 1 );
					add_action( 'woocommerce_order_action_tracking_status_delivered',       array( $this, 'myfreightshipping_tracking_status_delivered' ), 10, 1 );
					add_action( 'before_delete_post', array( $this, 'myfreightshipping_deleted' ), 10, 1 );
				}
			}


			/**
			 * Enqueue dangerous goods scripts
			 */
			public function myfreightshipping_enqueue_admin_dg_scripts( $hook ) {
				if ( $hook == 'post.php' && function_exists( 'is_woocommerce' ) ) {

//					wp_enqueue_style( 'jquery-ui-styles', plugins_url( 'css/myfreight-jquery-ui.min.css', __FILE__ ), null, '1.12.1', 'screen' );
					wp_register_script( 'myfreight-dg-codes', plugins_url( 'js/myfreightDGCodes.js', __FILE__ ), null, false, true );
					wp_register_script( 'myfreight-dg-product', plugins_url( 'js/myfreightDGProduct.js', __FILE__ ), array( 'jquery-ui-autocomplete', 'myfreight-dg-codes' ), false, true );
					wp_enqueue_script( 'myfreight-dg-product' );
				}
			}


			/**
			 * Add field to product shipping tab for entering dangerous goods data for that product
			 */
			public function myfreightshipping_dangerous_goods_fields( $links ) {

				echo '</div><div class="options_group">';	// New option group

 				$field = array(
					'id' => 'product_dangerous_goods_data',
					'label' => __( 'Dangerous goods spec', 'myfreightshipping' ),
					'desc_tip' => true,
					'description' => __( 'Start entering the 4-digit United Nations hazardous material identification number or Proper Shipping Name then select from the dropdown list.<br /><br />If Technical Name required (when Proper Shipping Name contains N.O.S.), add after comma and space at the end.<br /><br />For information on transporting Dangerous Goods refer to the National Transport Commissions "Australian Dangerous Goods Code".', 'myfreightshipping' ),
					'style' => 'width: 90%;',
					'placeholder' => __( 'Enter UN # or description ...', 'myfreightshipping' ),
					'type' => 'text',
					'default' => ''
				);
  
				woocommerce_wp_text_input( $field );

 				$field = array(
					'id' => 'product_dangerous_goods_qty',
					'class' => 'short',
					'label' => __( 'DG aggregate qty', 'myfreightshipping' ),
					'desc_tip' => true,
					'description' => __( 'Enter the aggregate quantity of hazardous material per item in kilograms (for solids, liquids and aerosols) or litres (for liquids and gases).', 'myfreightshipping' ),
					'placeholder' => __( 'kgs / litres', 'myfreightshipping' ),
					'type' => 'number',
					'custom_attributes' => array(
						'step' => '0.1',
						'min'  => '0'
					),
					'default' => ''
				);
  
				woocommerce_wp_text_input( $field );

				echo '</div>';
				echo '<div>';
			}

			/**
			 * Save dangerous goods field data to product meta data
			 */
			public function myfreightshipping_dangerous_goods_fields_save( $post_id ) {
				$product = wc_get_product( $post_id );

				$product_dangerous_goods_data = isset( $_POST['product_dangerous_goods_data'] ) ? $_POST['product_dangerous_goods_data'] : '';
				$product->update_meta_data( 'product_dangerous_goods_data', sanitize_text_field( $product_dangerous_goods_data ) );

				$product_dangerous_goods_qty = isset( $_POST['product_dangerous_goods_qty'] ) ? $_POST['product_dangerous_goods_qty'] : '';
				$product->update_meta_data( 'product_dangerous_goods_qty', sanitize_text_field( $product_dangerous_goods_qty ) );

				$product->save();
			}

			/**
			 * Add field to product variation for entering dangerous goods aggregate quantity for that variation
			 */
			public function myfreightshipping_dangerous_goods_variation_field( $loop, $variation_data, $variation ) {
				$product = wc_get_product( $variation->ID );

				$field = array(
					'id' => 'dangerous_goods_qty[' . $loop . ']',
					'class' => 'short',
					'label' => __( 'Dangerous Goods aggregate qty', 'myfreightshipping' ),
					'desc_tip' => true,
					'description' => __( 'Enter the aggregate quantity of hazardous material per variation item in kilograms (for solids, liquids and aerosols) or litres (for liquids and gases).<br /><br />Leave empty to use the parent product dangerous goods aggregate quantity.', 'myfreightshipping' ),
					'style' => 'width: 100%; margin: 2px 0 0 0; padding: 5px;',
					'placeholder' => __( 'Same as parent', 'myfreightshipping' ),
					'type' => 'number',
					'custom_attributes' => array(
						'step' => '0.1',
						'min'  => '0'
					),
					'default' => '',
					'value' => $product->get_meta( 'dangerous_goods_qty', true )
				);

				woocommerce_wp_text_input( $field );
			}

			/**
			 * Save dangerous goods field data to product variation meta data
			 */
			public function myfreightshipping_dangerous_goods_variation_field_save( $variation_id, $i ) {
				$product = wc_get_product( $variation_id );

				$dangerous_goods_qty = isset( $_POST['dangerous_goods_qty'][$i] ) ? $_POST['dangerous_goods_qty'][$i] : '';
				$product->update_meta_data( 'dangerous_goods_qty', sanitize_text_field( $dangerous_goods_qty ) );

				$product->save();
			}

			/**
			 * Include dangerous goods data in variation products
			 */
			public function myfreightshipping_dangerous_goods_variation_data( $variation_data, $parent_product, $variation ) {
				$product = wc_get_product( $variation_data['variation_id'] );

				$variation_data['dangerous_goods_qty'] = $product->get_meta( 'dangerous_goods_qty', true );

				return $variation_data;
			}

			public function myfreightshipping_dangerous_goods_cart_contents( $cart_contents ) {

				return $cart_contents;
			}


			/**
			 * Add field to product shipping tab for entering packing unit type for that product
			 */
			public function myfreightshipping_packing_unit_field( $links ) {

				echo '</div><div class="options_group">';	// New option group

 				$field = array(
					'id' => 'product_packing_unit',
					'class' => 'short',
					'label' => __( 'Packing unit', 'myfreightshipping' ),
					'desc_tip' => true,
					'description' => __( 'Enter the alphanumeric packing unit type for this product. Used for consolidating multiple items into packing containers for shipping.<br /><br />Packing rules are specified in the Shipping Settings.<br /><br />Leave empty if product shipped singly.', 'myfreightshipping' ),
					'placeholder' => __( '- optional -', 'myfreightshipping' ),
					'type' => 'text',
					'wrapper_class' => 'form-field',
					'default' => ''
				);
  
				woocommerce_wp_text_input( $field );
			}

			/**
			 * Save packing unit field data to product meta data
			 */
			public function myfreightshipping_packing_unit_field_save( $post_id ) {
				$product = wc_get_product( $post_id );

				$product_packing_unit = isset( $_POST['product_packing_unit'] ) ? $_POST['product_packing_unit'] : '';
				$product->update_meta_data( 'product_packing_unit', sanitize_text_field( $product_packing_unit ) );

				$product->save();
			}

			/**
			 * Add field to product variation for entering packing unit type for that variation
			 */
			public function myfreightshipping_packing_unit_variation_field( $loop, $variation_data, $variation ) {
				$product = wc_get_product( $variation->ID );

				$field = array(
					'id' => 'packing_unit[' . $loop . ']',
					'class' => 'short',
					'label' => __( 'Packing unit', 'myfreightshipping' ),
					'desc_tip' => true,
					'description' => __( 'Enter the alphanumeric packing unit type for this product variation. Used for consolidating multiple items into packing containers for shipping.<br /><br />Packing rules are specified in the Shipping Settings.<br /><br />Leave empty to use the parent product packing unit.', 'myfreightshipping' ),
					'style' => 'width: 100%; margin: 2px 0 0 0; padding: 5px;',
					'placeholder' => __( 'Same as parent', 'myfreightshipping' ),
					'type' => 'text',
					'wrapper_class' => 'form-row form-row-full hide_if_variation_virtual',
					'default' => '',
					'value' => $product->get_meta( 'packing_unit', true )
				);

				woocommerce_wp_text_input( $field );
			}

			/**
			 * Save packing unit field data to product variation meta data
			 */
			public function myfreightshipping_packing_unit_variation_field_save( $variation_id, $i ) {
				$product = wc_get_product( $variation_id );

				$packing_unit = isset( $_POST['packing_unit'][$i] ) ? $_POST['packing_unit'][$i] : '';
				$product->update_meta_data( 'packing_unit', sanitize_text_field( $packing_unit ) );

				$product->save();
			}

			/**
			 * Include packing unit data in variation products
			 */
			public function myfreightshipping_packing_unit_variation_data( $variation_data, $parent_product, $variation ) {
				$product = wc_get_product( $variation_data['variation_id'] );

				$variation_data['packing_unit'] = $product->get_meta( 'packing_unit', true );

				return $variation_data;
			}


			/**
			 * Add field to product shipping tab for entering store/warehouse sender address codes for that product
			 */
			public function myfreightshipping_sender_code_field( $links ) {

				echo '</div><div class="options_group">';	// New option group

 				$field = array(
					'id' => 'product_sender_codes',
					'label' => __( 'Store code(s)', 'myfreightshipping' ),
					'desc_tip' => true,
					'description' => __( 'Enter the code(s) of the store(s) from where this product is shipped. Separate multiple store codes with commas.<br /><br />Leave empty if product shipped from default stores/warehouses.', 'myfreightshipping' ),
					'placeholder' => __( 'Same as default(s)', 'myfreightshipping' ),
					'type' => 'text',
					'default' => ''
				);
  
				woocommerce_wp_text_input( $field );
			}

			/**
			 * Save sender address codes field data to product meta data
			 */
			public function myfreightshipping_sender_code_field_save( $post_id ) {
				$product = wc_get_product( $post_id );

				$product_sender_codes = isset( $_POST['product_sender_codes'] ) ? $_POST['product_sender_codes'] : '';
				$product->update_meta_data( 'product_sender_codes', sanitize_text_field( $product_sender_codes ) );

				$product->save();
			}


			/**
			 * Add selections to order action menu
			 */
			public function add_order_meta_box_actions( $actions ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				$actions['regenerate_shipping_consignment'] = __( 'Regenerate shipping consignment', 'myfreightshipping' );
				if ( $settings['account_type'] == 'Myparcel' ) {
					$actions['remove_shipping_consignment'] = __( 'Remove shipping consignment', 'myfreightshipping' );
				}
				$actions['tracking_status_booked']          = __( 'Tracking Status to Booked', 'myfreightshipping' );
				$actions['tracking_status_in_transit']      = __( 'Tracking Status to In Transit', 'myfreightshipping' );
				$actions['tracking_status_delivering']      = __( 'Tracking Status to Delivering Today', 'myfreightshipping' );
				$actions['tracking_status_failed_delivery'] = __( 'Tracking Status to Delivery Failed', 'myfreightshipping' );
				$actions['tracking_status_delivered']       = __( 'Tracking Status to Delivered', 'myfreightshipping' );

				return $actions;
			}


			/**
			 * Add ship action button to admin orders list
			 */
			public function myfreightshipping_order_list_action_button( $order ) {

				if ( method_exists( $order, 'has_status' ) && $order->has_status( array( 'processing', 'completed', 'dispatched' ) ) ) {

					$settings = get_option( 'woocommerce_myfreightshipping_settings' );

					/* Set default links */
					if ( $settings['account_type'] == 'Myfreight' ) {					// If account type Myfreight
						if ( $settings['test_mode'] == 'yes' ) {
							$url = 'https://uat.teamwilberforce.com/#/consignments';
						} else {
							$url = 'https://go.myfreight.com.au/#/consignments';			// Go to Myfreight consignments
						}
					} else {
						$url = MYPARCEL_URL . '/member-dashboard';					// Go to Myparcel dashboard latest shipments
					}

					$con_info = null;
					$delivery_status = 'View Shipment';

					if ( method_exists( $order, 'get_meta' ) ) {
						$consignments = $order->get_meta( 'consignment', false );
						if ( !empty($consignments) ) {
							$consignment = end($consignments);					// Get most recent consignment
							$con_meta = $consignment->get_data();
							$con_info = $con_meta['value'];
							$delivery_status = self::get_myfreight_state_short_description( $con_info['status'] );
							if ( $con_info['account_type'] == 'Myfreight' ) {			// If consignment created in Myfreight
								$url = $con_info['con_link'];
							} else {
								if ( $order->has_status( 'processing' ) && empty($con_info['id']) ) {	// If order status is processing and not booked yet
									if ( method_exists( $order, 'get_items' ) && count( $order->get_items() ) <= MYPARCEL_LINES_PER_BOOKING ) {	// If order generated single booking
										$url = $con_info['con_link'];
									} else {						// If order generated multiple bookings
										$order_query = '?order=' . $order->get_id();
										$url = MYPARCEL_URL . '/member-dashboard/' . $order_query . '#_shipmentspending';	// Go to Myparcel dashboard shipments pending
									}
								} else {							// If status not processing
									$booking_query = ( !empty($con_info['id']) && !empty($con_info['booking']) ) ? '?record=' . $con_info['booking'] : '';
									$url = MYPARCEL_URL . '/member-dashboard/' . $booking_query . '#_latestbookings';	// Go to Myparcel dashboard latest shipments
								}
							}
						}
					}

					if ( $delivery_status == 'In Transit' ) {
						$action_slug = 'order-in-transit';
						$name = __( 'View/Track Shipment', 'myfreightshipping' );			// In Transit tooltip
					} else if ( substr( $delivery_status, 0, 9 ) == 'Delivered' ) {
						$action_slug = 'order-delivered';
						$name = __( 'Delivered', 'myfreightshipping' );					// Delivered tooltip
					} else if ( $order->has_status( 'processing' ) && !empty($con_info) && $con_info['account_type'] == 'Myfreight' && ( $con_info['status'] == 'open' || $con_info['status'] == 'despatchable' ) ) {
						$action_slug = 'ship-order';
						$name = __( 'Ship Order', 'myfreightshipping' );				// Despatch tooltip
					} else if ( $order->has_status( 'processing' ) && !empty($con_info) && $con_info['account_type'] != 'Myfreight' && $con_info['status'] == 'unbooked' ) {
						$action_slug = 'ship-order';
						$name = __( 'Ship Order', 'myfreightshipping' );				// Book tooltip
					} else {
						$action_slug = 'ship-order';
						$name = __( $delivery_status, 'myfreightshipping' );				// Other Status tooltip
					}

					// Custom action button (with a target='_blank' opening a new browser window)
					printf( '<a class="button wc-action-button wc-action-button-%s %s" href="%s" aria-label="%s" title="%s" target="_blank">%s</a>', $action_slug, $action_slug, $url, $name, $name, $name );
				}
			}

			/**
			 * Style admin orders list action button
			 */
			public function myfreightshipping_order_list_action_button_css() {

				// Show coloured truck icon on action button
				$action_slug = 'ship-order';
				echo '<style>.wc-action-button-' . $action_slug . '::after { font-family: woocommerce !important; content: "\e01a" !important; }</style>';
				$action_slug = 'order-in-transit';
				echo '<style>.wc-action-button-' . $action_slug . '::after { font-family: woocommerce !important; content: "\e01a" !important; color: #40a040; }</style>';
				$action_slug = 'order-delivered';
				echo '<style>.wc-action-button-' . $action_slug . '::after { font-family: woocommerce !important; content: "\e01a" !important; color: #a04040; }</style>';
			}


			/**
			 * Initiate or cancel shipping on order status changes
			 */
			public function myfreightshipping_order_status_change( $order_id, $status_from, $status_to, $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				if ( $status_to == 'processing' && ( ( $status_from != 'completed' && $status_from != 'dispatched' ) || ( $settings['account_type'] == 'Myparcel' && $settings['myparcel_remove_completed'] == 'yes' ) ) ) {
					$this->myfreightshipping_processing( $order );		// If processing, generate consignment
				} else if ( ( $status_to == 'completed' || $status_to == 'dispatched' ) && ( $settings['account_type'] == 'Myparcel' && $settings['myparcel_remove_completed'] == 'yes' ) ) {
					$this->myfreightshipping_completed( $order );		// If completed, remove consignment (unless settings indicate keep)
				} else if ( ( $status_to != 'processing' && $status_to != 'completed' && $status_to != 'dispatched' ) && ( $status_from == 'processing' || $status_from == 'completed' || $status_from == 'dispatched' ) ) {
					$this->myfreightshipping_completed( $order );		// If any other status, remove consignment
				}
			}


			/**
			 * Initiate shipping on custom order action to regenerate consignment
			 */
			public function myfreightshipping_regenerate_consignment( $order ) {

				$con_id = $this->myfreightshipping_processing( $order );	// Generate new consignment for order

				if ( $con_id !== false ) {					// If consignment successfully generated
					$order->add_order_note( 'Consignment ' . ( $con_id > 0 ? $con_id . ' ' : '' ) . 'regenerated.', false, false );
				} else {
					$order->add_order_note( 'Consignment regeneration failed.', false, false );
				}
			}


			/**
			 * Cancel shipping on custom order action to remove consignment
			 */
			public function myfreightshipping_remove_consignment( $order ) {

				$removed = $this->myfreightshipping_completed( $order );	// Cancel consignment for order

				if ( $removed !== false ) {					// If consignment successfully cancelled
					$order->add_order_note( 'Consignment removed.', false, false );
				} else {
					$order->add_order_note( 'Consignment removal unsuccessful.', false, false );
				}
			}


			/**
			 * Update shipment tracking status to booked
			 */
			public function myfreightshipping_tracking_status_booked( $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$account_type = $settings['account_type'];

				(new WC_Myfreight_Tracking)->mf_shipping_manual_status_update( $order, ( $account_type == 'Myfreight' ) ? 'despatched' : 'booked' );
			}

			/**
			 * Update shipment tracking status to in transit
			 */
			public function myfreightshipping_tracking_status_in_transit( $order ) {

				(new WC_Myfreight_Tracking)->mf_shipping_manual_status_update( $order, 'in_transit' );
			}

			/**
			 * Update shipment tracking status to on board for delivery
			 */
			public function myfreightshipping_tracking_status_delivering( $order ) {

				(new WC_Myfreight_Tracking)->mf_shipping_manual_status_update( $order, 'on_board_for_delivery' );
			}

			/**
			 * Update shipment tracking status to failed delivery
			 */
			public function myfreightshipping_tracking_status_failed_delivery( $order ) {

				(new WC_Myfreight_Tracking)->mf_shipping_manual_status_update( $order, 'failed_delivery' );
			}

			/**
			 * Update shipment tracking status to delivered
			 */
			public function myfreightshipping_tracking_status_delivered( $order ) {

				(new WC_Myfreight_Tracking)->mf_shipping_manual_status_update( $order, 'delivered' );
			}


			private static $logger;		// Handle for debug logging

			/**
			 * myfreightshipping_processing function.
			 *
			 * @access public
			 * @param array $order
			 * @return void
			 *
			 * Initiate shipping on successful order
			 */
			public function myfreightshipping_processing( $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				/* Initialise debug */
				if ( $settings['debug'] == 'yes' ) {
					$timezone_string = get_option( 'timezone_string' );
					date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );
					self::$logger = wc_get_logger();
					self::$logger->debug( '==== Begin Creating ' . $settings['account_type'] . ' Consignment ====', array( 'source' => 'Myfreight Shipping' ) );
				} else {
					self::$logger = null;
				}

				/* Check whether Myfreight or Myparcel */
				if ( $settings['account_type'] == 'Myfreight' ) {
					return $this->consign_myfreight_order( $order );
				} else {
					return $this->consign_myparcel_order( $order );
				}
			}


			/**
			 * myfreightshipping_completed function.
			 *
			 * @access public
			 * @param array $order
			 * @return void
			 *
			 * Remove shipment when status set to complete
			 */
			public function myfreightshipping_completed( $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				/* Check whether Myfreight or Myparcel */
				if ( $settings['account_type'] == 'Myparcel' ) {
					return $this->remove_myparcel_order( $order );
				} else {
					return false;
				}
			}


			/**
			 * myfreightshipping_deleted function.
			 *
			 * @access public
			 * @param int $post_id
			 * @return void
			 *
			 * Remove shipment when deleted from trash
			 */
			public function myfreightshipping_deleted( $post_id ) {

				if ( get_post_type( $post_id ) != 'shop_order' ) {
					return;
				}

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				/* Check whether Myfreight or Myparcel */
				if ( $settings['account_type'] == 'Myparcel' ) {

					$order = new WC_Order( $post_id );

					return $this->remove_myparcel_order( $order );
				}
			}


			/**
			 * Consign order via Myfreight
			 */
			function consign_myfreight_order( $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				$shipping_methods = $order->get_shipping_methods();
				if ( !empty($shipping_methods) ) {
					$shipping_method = reset( $shipping_methods );
				} else {
					$shipping_method = null;
				}

				/* Prepare Sender */

				$sender_code = null;
				if ( $settings['multiple_senders'] == 'yes' ) {
					$sender_code = method_exists( $shipping_method, 'get_meta' ) ? $shipping_method->get_meta( 'Store_Code' ) : null;
				}
				if ( empty($sender_code) ) {
					if ( !empty($settings['myfreight_sender_codes']) ) {
						$sender_codes = explode( ',', str_replace( ' ', '', $settings['myfreight_sender_codes'] ) );
						$sender_code = $sender_codes[0];
					}
				}

				if ( $settings['multiple_senders'] == 'yes' && !empty($sender_code) && $settings['myfreight_sender_addresses'][0] == '{' ) {
					$sender_locations = json_decode( $settings['myfreight_sender_addresses'], true );
					$sender_location = $sender_locations[$sender_code];

					$sender_address = array(
						'code'     => $sender_code,
						'locality' => $sender_location['sender_city'],
						'postcode' => $sender_location['sender_postcode'],
						'region'   => $sender_location['sender_state'],
						'country'  => 'Australia'
					);
				} else {
					if ( !empty($settings['store_name']) ) {
						$sender_address = array(
							'name'           => $settings['store_company'],
							'code'           => $sender_code,
							'email'          => $settings['store_email'],
							'phone'          => $settings['store_phone'],
							'address_line_1' => 'ATTN ' . $settings['store_name'],
							'address_line_2' => WC()->countries->get_base_address(),
							'address_line_3' => WC()->countries->get_base_address_2(),
							'address_line_4' => null,
							'sublocality'    => null,
							'locality'       => WC()->countries->get_base_city(),
							'postcode'       => WC()->countries->get_base_postcode(),
							'region'         => WC()->countries->get_base_state(),
							'country'        => 'Australia'
						);
					} else {
						$sender_address = array(
							'name'           => $settings['store_company'],
							'code'           => $sender_code,
							'email'          => $settings['store_email'],
							'phone'          => $settings['store_phone'],
							'address_line_1' => WC()->countries->get_base_address(),
							'address_line_2' => WC()->countries->get_base_address_2(),
							'address_line_3' => null,
							'address_line_4' => null,
							'sublocality'    => null,
							'locality'       => WC()->countries->get_base_city(),
							'postcode'       => WC()->countries->get_base_postcode(),
							'region'         => WC()->countries->get_base_state(),
							'country'        => 'Australia'
						);
					}
				}

				/* Prepare Receiver */

				$order_id = method_exists( $order, 'get_id' ) ? $order->get_id() : $order->id;
				if ( get_post_meta( $order_id, '_codisto_orderid', true ) ) {		// If order generated by plugin "WooCommerce Amazon & eBay Integration" (by Codisto LINQ)
					$codisto_plugin = true;

					if ( !empty(get_post_meta( $order_id, '_shipping_postcode', true )) ) {				// If shipping address provided
						$receiver_name      = get_post_meta( $order_id, '_shipping_first_name', true ) . ' ' . get_post_meta( $order_id, '_shipping_last_name', true );
						$receiver_company   = get_post_meta( $order_id, '_shipping_company', true );
						$receiver_address_1 = get_post_meta( $order_id, '_shipping_address_1', true );
						$receiver_address_2 = get_post_meta( $order_id, '_shipping_address_2', true );
						$receiver_city      = get_post_meta( $order_id, '_shipping_city', true );
						$receiver_state     = get_post_meta( $order_id, '_shipping_state', true );
						$receiver_postcode  = get_post_meta( $order_id, '_shipping_postcode', true );
						$receiver_phone     = get_post_meta( $order_id, '_shipping_phone', true );
						$receiver_email     = get_post_meta( $order_id, '_shipping_email', true );
					} else {
						$receiver_name      = get_post_meta( $order_id, '_billing_first_name', true ) . ' ' . get_post_meta( $order_id, '_billing_last_name', true );
						$receiver_company   = get_post_meta( $order_id, '_billing_company', true );
						$receiver_address_1 = get_post_meta( $order_id, '_billing_address_1', true );
						$receiver_address_2 = get_post_meta( $order_id, '_billing_address_2', true );
						$receiver_city      = get_post_meta( $order_id, '_billing_city', true );
						$receiver_state     = get_post_meta( $order_id, '_billing_state', true );
						$receiver_postcode  = get_post_meta( $order_id, '_billing_postcode', true );
						$receiver_phone     = get_post_meta( $order_id, '_billing_phone', true );
						$receiver_email     = get_post_meta( $order_id, '_billing_email', true );
					}
				} else {								// If order generated by WooCommerce checkout
					$codisto_plugin = false;

					if ( method_exists( $order, 'has_shipping_address' ) && $order->has_shipping_address() ) {	// If customer has entered shipping address in addition to billing address
						$receiver_name      = $order->get_formatted_shipping_full_name();
						$receiver_company   = $order->get_shipping_company();
						$receiver_address_1 = $order->get_shipping_address_1();
						$receiver_address_2 = $order->get_shipping_address_2();
						$receiver_city      = $order->get_shipping_city();
						$receiver_state     = $order->get_shipping_state();
						$receiver_postcode  = $order->get_shipping_postcode();
						if ( is_callable( array( $order, 'get_shipping_phone' ) ) ) {
							$receiver_phone = $order->get_shipping_phone();
						} else {
							$receiver_phone = get_post_meta( $order_id, '_shipping_phone', true );
						}
						if ( empty( $receiver_phone ) ) {
							$receiver_phone = $order->get_billing_phone();
						}
					} else {
						$receiver_name      = $order->get_formatted_billing_full_name();
						$receiver_company   = $order->get_billing_company();
						$receiver_address_1 = $order->get_billing_address_1();
						$receiver_address_2 = $order->get_billing_address_2();
						$receiver_city      = $order->get_billing_city();
						$receiver_state     = $order->get_billing_state();
						$receiver_postcode  = $order->get_billing_postcode();
						$receiver_phone     = $order->get_billing_phone();
					}
					$receiver_email = $order->get_billing_email();
				}

				if ( empty($receiver_address_1) && empty($receiver_city) && $settings['inhibit_empty_receiver'] == 'yes' ) {	// Check receiver address not empty
					return false;
				}

				$receiver_locality = WC_Myfreight_Ancillary::clean_myfreight_locality(	$receiver_city,
													$receiver_state,
													$receiver_postcode,
													(int)($settings['address_match_percentage']) );

				/* Modify receiver email as specified to prevent Advanced Shipping Notice (ASN) going to purchaser */
				$consign_receiver_email = $settings['consign_receiver_email'];
				if ( $consign_receiver_email == 'Sender' ) {
					$receiver_email = $settings['store_email'];
				} else if ( $consign_receiver_email == 'Empty' ) {
					$receiver_email = '';
				}							// Else if 'Purchaser', leave as purchaser email

				if ( !empty($receiver_company) ) {
					$receiver_address = array(
						'name'           => $receiver_company,
						'code'           => null,
						'email'          => $receiver_email,
						'phone'          => $receiver_phone,
						'address_line_1' => 'ATTN ' . $receiver_name,
						'address_line_2' => $receiver_address_1,
						'address_line_3' => $receiver_address_2,
						'address_line_4' => null,
						'sublocality'    => null,
						'locality'       => $receiver_locality['locality'],
						'postcode'       => $receiver_locality['postcode'],
						'region'         => $receiver_locality['region'],
						'country'        => 'Australia'
					);
				} else {
					$receiver_address = array(
						'name'           => $receiver_name,
						'code'           => null,
						'email'          => $receiver_email,
						'phone'          => $receiver_phone,
						'address_line_1' => $receiver_address_1,
						'address_line_2' => $receiver_address_2,
						'address_line_3' => null,
						'address_line_4' => null,
						'sublocality'    => null,
						'locality'       => $receiver_locality['locality'],
						'postcode'       => $receiver_locality['postcode'],
						'region'         => $receiver_locality['region'],
						'country'        => 'Australia'
					);
				}

				/* Prepare Items */

				$fragile_item = false;
				$packing_item = false;

				$shipping_items = array();
				$items_line = 0;

				$order_number = $order->get_order_number();
				$items = $order->get_items();

				foreach ( $items as $item_id => &$item ) {

					++$items_line;

					$product = null;
					$variation_id = method_exists( $item, 'get_variation_id' ) ? (int)($item->get_variation_id()) : ( isset($item['variation_id']) ? $item['variation_id'] : 0 );
					if ( $variation_id > 0 ) {									// If variation
						$variable_product_id = method_exists( $item, 'get_product_id' ) ? (int)($item->get_product_id()) : (int)($item['product_id']);
						$variable_product = wc_get_product( $variable_product_id );				// Get parent (variable) product
						if ( empty($variable_product) ) {
							$variable_product = new WC_Product_Variable( $variable_product_id );
						}

						$product = method_exists( $item, 'get_product' ) ? $item->get_product() : null;		// Get child (variation) product
						if ( empty($product) ) {
							$product = wc_get_product( $variation_id );
							if ( empty($product) ) {
								$product = new WC_Product_Variation( $variation_id );
							}
						}
					} else {											// If not variation
						$product = method_exists( $item, 'get_product' ) ? $item->get_product() : null;
						if ( empty($product) ) {
							$product_id = method_exists( $item, 'get_product_id' ) ? (int)($item->get_product_id()) : (int)($item['product_id']);
							if ( $product_id > 0 ) {
								$product = wc_get_product( $product_id );
								if ( empty($product) ) {
									$product = new WC_Product( $product_id );
								}
							}
						}
					}
					if ( !is_object( $product ) ) {
						if ( $codisto_plugin ) {
							$product = null;
						} else {
							continue;
						}
					}
					if ( method_exists( $product, 'is_virtual' ) && $product->is_virtual() ) {		// If virtual product no shipping is required
						continue;
					}

					$item_quantity = method_exists( $item, 'get_quantity' ) ? (int)($item->get_quantity()) : (int)($item['quantity']);
					if ( empty($item_quantity) ) {
						continue;
					}

					/* Retrieve product weight and dimensions, or defaults if missing */
					$item_weight = method_exists( $product, 'get_weight' ) ? (float)($product->get_weight()) : 0.0;
					if ( empty($item_weight) && $variation_id > 0 ) {			// If variation
						$item_weight = method_exists( $variable_product, 'get_weight' ) ? (float)($variable_product->get_weight()) : 0.0;
					}
					if ( empty($item_weight) ) {
						$item_weight = (float)($settings['default_weight']);
					}

					$item_length = method_exists( $product, 'get_length' ) ? (int)($product->get_length()) : 0;
					if ( empty($item_length) && $variation_id > 0 ) {			// If variation
						$item_length = method_exists( $variable_product, 'get_length' ) ? (int)($variable_product->get_length()) : 0;
					}
					if ( empty($item_length) ) {
						$item_length = (int)($settings['default_length']);
					}

					$item_width  = method_exists( $product, 'get_width' ) ? (int)($product->get_width()) : 0;
					if ( empty($item_width) && $variation_id > 0 ) {			// If variation
						$item_width = method_exists( $variable_product, 'get_width' ) ? (int)($variable_product->get_width()) : 0;
					}
					if ( empty($item_width) ) {
						$item_width = (int)($settings['default_width']);
					}

					$height_string = method_exists( $product, 'get_height' ) ? $product->get_height() : null;
					if ( empty($height_string) && $variation_id > 0 ) {			// If variation
						$height_string = method_exists( $variable_product, 'get_height' ) ? $variable_product->get_height() : null;
					}
					if ( empty($height_string) ) {
						$height_string = $settings['default_height'];
					}
					$item_height = (int)$height_string;
					$item_height_decimal = (float)$height_string;
					if ( $item_height == 0 && $item_height_decimal > 0.0 ) {
						$item_height = 1;
					}

//					$shipping_class = method_exists( $product, 'get_shipping_class' ) ? $product->get_shipping_class() : get_term_by( 'id', (int)($product->shipping_class_id), 'product_shipping_class' )->slug;
					$shipping_class = method_exists( $product, 'get_shipping_class' ) ? $product->get_shipping_class() : null;
					if ( empty($shipping_class) && $variation_id > 0 ) {			// If variation
						$shipping_class = method_exists( $variable_product, 'get_shipping_class' ) ? $variable_product->get_shipping_class() : null;
					}
					$shipping_class_prefix = strtolower( substr( $shipping_class, 0, 6 ) );
					$item_fragile = false;
					if ( $shipping_class_prefix == 'fragil' || $shipping_class_prefix == 'handle' || $shipping_class_prefix == 'manual' ) {		// If product shipping class is fragile
						$item_fragile = true;
						$fragile_item = true;
					}

					$dangerous_goods_items = null;
					if ( $variation_id > 0 ) {						// If variation
						$dg_data = method_exists( $variable_product, 'get_meta' ) ? $variable_product->get_meta('product_dangerous_goods_data') : null;
					} else {
						$dg_data = method_exists( $product, 'get_meta' ) ? $product->get_meta('product_dangerous_goods_data') : null;
					}
					if ( !empty($dg_data) ) {
						if ( $variation_id > 0 ) {					// If variation
							$dg_qty = (float)($product->get_meta('dangerous_goods_qty'));
							if ( empty($dg_qty) ) {
								$dg_qty = (float)($variable_product->get_meta('product_dangerous_goods_qty'));
							}
						} else {
							$dg_qty = (float)($product->get_meta('product_dangerous_goods_qty'));
						}
						$dg_qty *= $item_quantity;
					} else {
						$dg_qty = 0.0;
					}
					if ( !empty($dg_data) && !empty($dg_qty) && $dg_qty > 0.0 ) {		// If product dangerous goods data available
						$dg_fields = array();
						if ( preg_match( '/^\s*UN:\s*([0-9]+),\s*(.+),\s*Class:\s*([0-9\.]+[a-zA-Z]?),\s*SubRisk:\s*([0-9\.]+[a-zA-Z]?|NIL),\s*PkgGrp:\s*(I+|NIL)(,\s*.*)?\s*$/', $dg_data, $dg_fields ) ) {
							$dangerous_goods_items = array(
								array(
									'un_number'            => $dg_fields[1],
									'proper_shipping_name' => $dg_fields[2],
									'technical_name'       => !empty($dg_fields[6]) ? trim( $dg_fields[6], ', ' ) : null,
									'hazard_class'         => $dg_fields[3],
									'sub_risk'             => $dg_fields[4] != 'NIL' ? $dg_fields[4] : null,
									'packing_group'        => $dg_fields[5] != 'NIL' ? $dg_fields[5] : null,
									'aggregate_quantity'   => $dg_qty
								)
							);
						}
					}

					if ( $variation_id > 0 ) {						// If variation
						$packing_unit = method_exists( $product, 'get_meta' ) ? $product->get_meta('packing_unit') : null;
						if ( empty($packing_unit) ) {
							$packing_unit = method_exists( $variable_product, 'get_meta' ) ? $variable_product->get_meta('product_packing_unit') : null;
						}
					} else {
						$packing_unit = method_exists( $product, 'get_meta' ) ? $product->get_meta('product_packing_unit') : null;
					}
					if ( !empty($packing_unit) ) {						// If packing unit field completed
						$packing_item = true;
					} else {
						$packing_unit = null;
					}

					$shipping_items[] = array(
						'reference'             => 'Order #' . $order_number . ' line ' . (string)$items_line,
						'item_type_description' => apply_filters( 'my_est_item_type', 'Carton', $item_quantity, $item_weight, $item_length, $item_width, $item_height, $shipping_class, $settings['account_type'] ),
						'quantity'              => $item_quantity,
						'weight_in_kilograms'   => $item_weight * $item_quantity,
						'length'                => $item_length,
						'width'	                => $item_width,
						'height'                => $item_height,
						'height_decimal'        => $item_height_decimal,
						'fragile'               => $item_fragile,
						'dangerous_goods_items' => $dangerous_goods_items,
						'packing_unit'          => $packing_unit
					);

					// If plugin "Add More Shipping Fields (For Multi Part Product)" (by PluginHive) active to support multiple shipping items per product
					if ( is_plugin_active( 'Addon_Plugin_dividing_single_product/wf-add-more-shipping-fileds-addon.php' ) && method_exists( $product, 'get_meta' ) ) {
						for ( $i = 2; $i < 10 && !empty($product->get_meta( '_weight' . $i )); $i++ ) {		// For each additional shipping item
							$item_weight = (float)($product->get_meta( '_weight' . $i ));
							$item_length = !empty($product->get_meta( '_length' . $i )) ? (int)($product->get_meta( '_length' . $i )) : (int)($settings['default_length']);
							$item_width  = !empty($product->get_meta( '_width' . $i ))  ? (int)($product->get_meta( '_width' . $i ))  : (int)($settings['default_width']);
							$height_string = !empty($product->get_meta( '_height' . $i )) ? $product->get_meta( '_height' . $i ) : $settings['default_height'];
							$item_height = (int)$height_string;
							$item_height_decimal = (float)$height_string;
							if ( $item_height == 0 && $item_height_decimal > 0.0 ) {
								$item_height = 1;
							}

							/* Load additional items into items array */
							$shipping_items[] = array(
								'reference'             => 'Order #' . $order_number . ' line ' . (string)$items_line . '.' . $i,
								'item_type_description' => apply_filters( 'my_est_item_type', 'Carton', $item_quantity, $item_weight, $item_length, $item_width, $item_height, $shipping_class, $settings['account_type'] ),
								'quantity'              => $item_quantity,
								'weight_in_kilograms'   => $item_weight * $item_quantity,
								'length'                => $item_length,
								'width'                 => $item_width,
								'height'                => $item_height,
								'height_decimal'        => $item_height_decimal,
								'fragile'               => $item_fragile,
								'dangerous_goods_items' => $dangerous_goods_items,
								'packing_unit'          => $packing_unit
							);
						}
					}
				}
				unset( $item );

				/*
				 * Filter hook to allow custom packing of items into containers.
				 * Packing rules are applied first, then auto packing.
				 * Custom packing rules can also be hooked.
				 */
				$shipping_items = apply_filters( 'my_pack_items', $shipping_items, false, $packing_item );	// Pack shipping items together as defined by settings

				if ( empty($shipping_items) ) {
					return false;
				}

				/* Check for container over weight threshold */
				$unloading_threshold = (int)$settings['unloading_threshold'];
				$unloading_capability_active = ( $settings['unloading_capability'] == 'yes' );
				$heavy_container = false;
				if ( !empty($unloading_threshold) ) {
					$max_item_weight = 0.0;
					foreach ( $shipping_items as &$shipping_item ) {
						$item_weight = (float)round( $shipping_item['weight_in_kilograms'] / $shipping_item['quantity'], 1 );
						if ( $item_weight > $max_item_weight ) {
							$max_item_weight = $item_weight;
						}
					}
					unset( $shipping_item );

					if ( $max_item_weight >= $unloading_threshold ) {
						$heavy_container = true;
					}
				}

				/* Prepare Consignment */

				$credentials = $settings['myfreight_username'] . ':' . $settings['myfreight_password'];
				$site_code = $settings['myfreight_site_code'];

				if ( $settings['test_mode'] == 'yes' ) {
					$url = 'https://uat.teamwilberforce.com/public/api/v1/consignments';
				} else {
					$url = 'https://go.myfreight.com.au/public/api/v1/consignments';
					if ( strpos( strtolower( $credentials . $site_code ), 'myparcel' ) !== false ) {
						return false;
					}
				}

				$header = array(
					'Content-type: application/json',
					'Accept: application/json',
					'Authorization: Basic ' . base64_encode( $credentials )
				);

				$delivery_date = null;
				$delivery_datime = null;
				if ( is_callable( array( 'Orddd_Lite_Common', 'orddd_lite_get_order_delivery_date' ) ) ) {	// If plugin "Order Delivery Date" (by Tyche Softwares) active
					$order_delivery_date = Orddd_Lite_Common::orddd_lite_get_order_delivery_date( $order->get_id() );
					if ( !empty($order_delivery_date) ) {
				$timezone_string = get_option( 'timezone_string' );
				date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );

						$delivery_timestamp = Orddd_Lite_Common::orddd_lite_get_timestamp( $order_delivery_date, get_option( 'orddd_lite_delivery_date_format' ) );
						$delivery_date = date( 'D j M', $delivery_timestamp );
						$delivery_datime = date( 'Y-m-d\TH:i:s', $delivery_timestamp );
					}
				}

				$special_instructions = !empty($delivery_date) ? 'DELIVER: ' . $delivery_date . '. ' : '';
				$special_instructions .= $fragile_item ? 'FRAGILE MANUAL HANDLING. ' : '';
				$special_instructions .= ( $heavy_container && ( !$unloading_capability_active || get_post_meta( $order_id, '_unloading_capability', true ) != 'Yes' ) ) ? 'TAILGATE LIFT REQD. ' : '';
				$special_instructions .= method_exists( $order, 'get_customer_note' ) ? $order->get_customer_note() : '';
				$special_instructions = substr( $special_instructions, 0, 80 );

				$carrier_service = method_exists( $shipping_method, 'get_meta' ) ? $shipping_method->get_meta( 'Service_Code' ) : null;
				if ( empty($carrier_service) ) {
					$carrier_service = null;
				}

				foreach ( $shipping_items as &$shipping_item ) {
					if ( isset($shipping_item['height_decimal']) ) {
						unset( $shipping_item['height_decimal'] );	// Remove decimal place (float) variants of item height
					}
				}
				unset( $shipping_item );

				$consignment_create = array(
						'site_code'            => $site_code,
						'despatch_date'        => WC_Myfreight_Ancillary::next_working_date( 'today', MY_DESPATCH_DELAY_DAYS, $order ),
						'deliver_at'           => $delivery_datime,
						'reference'            => 'Order #' . $order_number,
						'custom_reference'     => parse_url( get_home_url(), PHP_URL_HOST ) . '#' . $order_number,
						'account_payable'      => 'sender',
						'charge_account'       => null,
						'special_instructions' => $special_instructions,
						'carrier_service'      => $carrier_service,
						'carrier_route'        => null,
						'custom_label_printer' => null,
						'third_party'          => false,
						'sender_address'       => $sender_address,
						'receiver_address'     => $receiver_address,
						'items'                => $shipping_items
				);

				if ( !empty( self::$logger ) ) {
					self::$logger->debug( 'CONSIGNMENT CREATE: ' . print_r( $consignment_create, true ), array( 'source' => 'Myfreight Shipping' ) );
				}

				$body = json_encode( $consignment_create );

				$ch = curl_init();
				curl_setopt($ch, CURLOPT_URL, $url);
				curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
				curl_setopt($ch, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
				curl_setopt($ch, CURLOPT_HTTPHEADER, $header);
				curl_setopt($ch, CURLOPT_POST, true);
				curl_setopt($ch, CURLOPT_POSTFIELDS, $body);
				curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
				curl_setopt($ch, CURLOPT_ENCODING, '');
				curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 10);
				curl_setopt($ch, CURLOPT_TIMEOUT, 20);
				curl_setopt($ch, CURLOPT_VERBOSE, false);

				if ( !empty( self::$logger ) ) {
					self::$logger->debug( 'CONSIGNMENT JSON SENT: ' . print_r( $body, true ), array( 'source' => 'Myfreight Shipping' ) );
				}

				$session_status = session_status();
				if ( $session_status == PHP_SESSION_ACTIVE ) {
					session_write_close();		// Close session to prevent AJAX blocking
				}

				$data = curl_exec($ch);			// Send consignment data

				if ( $session_status == PHP_SESSION_ACTIVE ) {
					session_start();		// Reopen session
				}

				if ( !empty( self::$logger ) ) {
					self::$logger->debug( 'CONSIGNMENT JSON RECEIVED: ' . print_r( $data, true ), array( 'source' => 'Myfreight Shipping' ) );
				}

				if ( !curl_errno($ch) && !empty($data) ) {
					if ( $consignment = (array)json_decode( $data, true ) ) {

						if ( empty( $consignment['errors'] ) ) {

							if ( !empty( $consignment['id'] ) ) {
								curl_close($ch);

								/* Save order shipping details */
								if ( $settings['test_mode'] == 'yes' ) {
									$con_link = 'https://uat.teamwilberforce.com/#/consignment/' . $consignment['id'];
								} else {
									$con_link = 'https://go.myfreight.com.au/#/consignment/' . $consignment['id'];
								}

								$con_info = array(
									'account_type'  => 'Myfreight',
									'test_mode'     => $settings['test_mode'],
									'id'            => $consignment['id'],
									'status'        => 'open',
									'con_link'      => $con_link,
									'tracking_link' => null
								);

								if ( method_exists( $order, 'add_meta_data' ) ) {
									$order->add_meta_data( 'consignment', $con_info );
									$order->save_meta_data();
								}

								$consign_count = strtolower( $settings['account_type'] ) . ( $settings['test_mode'] == 'no' ? '_live' : '_test' ) . '_consign_count';
								$counters = get_option( 'myfreightshipping_counters', array() );
								if ( empty($counters[$consign_count]) ) {
									$counters[$consign_count] = 1;
									$counters[$consign_count . '_start'] = date( 'Y-m-d' );
								} else {
									$counters[$consign_count]++;
								}
								update_option( 'myfreightshipping_counters', $counters, false );

								if ( !empty( self::$logger ) ) {
									self::$logger->debug( 'CONSIGNMENT RESPONSE: ' . print_r( $consignment, true ), array( 'source' => 'Myfreight Shipping' ) );
								}

								return $consignment['id'];
							} else {
								if ( !empty( self::$logger ) ) {
									self::$logger->debug( 'CONSIGNMENT ID MISSING: ' . print_r( $consignment, true ), array( 'source' => 'Myfreight Shipping' ) );
								}
							}
						} else {
							if ( !empty( self::$logger ) ) {
								self::$logger->debug( 'CONSIGNMENT ERROR: ' . print_r( $consignment['errors'][0], true ), array( 'source' => 'Myfreight Shipping' ) );
							}
						}
					} else {
						if ( !empty( self::$logger ) ) {
							self::$logger->debug( 'CONSIGNMENT DECODE ERROR: ' . print_r( $data, true ), array( 'source' => 'Myfreight Shipping' ) );
						}
					}
				} else if ( curl_errno($ch) ) {
					if ( !empty( self::$logger ) ) {
						self::$logger->debug( 'API ERROR: ' . print_r( curl_errno($ch), true ), array( 'source' => 'Myfreight Shipping' ) );
					}
				}
				curl_close($ch);

				return false;
			}


			/**
			 * Consign order via Myparcel by adding to pending shipments list
			 */
			function consign_myparcel_order( $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				$shipping_methods = $order->get_shipping_methods();
				if ( !empty($shipping_methods) ) {
					$shipping_method = reset( $shipping_methods );
				} else {
					$shipping_method = null;
				}

				/* Prepare Sender */

				$sender_code = null;
				if ( $settings['multiple_senders'] == 'yes' ) {
					$sender_code = ( !empty($shipping_method) && method_exists( $shipping_method, 'get_meta' ) ) ? $shipping_method->get_meta( 'Store_Code' ) : null;
					if ( empty($sender_code) ) {
						if ( !empty($settings['myparcel_sender_codes']) ) {
							$sender_codes = explode( ',', str_replace( ' ', '', $settings['myparcel_sender_codes'] ) );
							$sender_code = $sender_codes[0];
						}
					}
				}

				if ( !empty($sender_code) && $settings['myparcel_sender_addresses'][0] == '{' ) {
					$sender_locations = json_decode( $settings['myparcel_sender_addresses'], true );
					$sender_location = $sender_locations[$sender_code];

					$sender_details = array(
						'sender_name_first'     => $sender_location['sender_name_first'],
						'sender_name_last'      => $sender_location['sender_name_last'],
						'sender_company'        => $sender_location['sender_company'],
						'sender_street'         => $sender_location['sender_street'],
						'sender_street2'        => $sender_location['sender_street2'],
						'sender_locality'       => $sender_location['sender_city'] . ' ' . $sender_location['sender_state'] . ' ' . $sender_location['sender_postcode'],
//						'sender_city'           => $sender_location['sender_city'],
//						'sender_state'          => $sender_location['sender_state'],
//						'sender_postcode'       => $sender_location['sender_postcode'],
						'sender_residential'    => $sender_location['sender_residential'],
						'sender_phone'          => $sender_location['sender_phone'],
						'sender_email'          => $sender_location['sender_email'],
						'sender_loading_assist' => $sender_location['sender_loading_assist']
					);
				} else {
					$name_words = explode( ' ', $settings['store_name'] );
					$sender_name_first = array_shift( $name_words );
					$sender_name_last  = implode( $name_words );

					$sender_details = array(
						'sender_name_first'     => $sender_name_first,
						'sender_name_last'      => $sender_name_last,
						'sender_company'        => $settings['store_company'],
						'sender_street'         => WC()->countries->get_base_address(),
						'sender_street2'        => WC()->countries->get_base_address_2(),
						'sender_locality'       => WC()->countries->get_base_city() . ' ' . WC()->countries->get_base_state() . ' ' . WC()->countries->get_base_postcode(),
//						'sender_city'           => WC()->countries->get_base_city(),
//						'sender_state'          => WC()->countries->get_base_state(),
//						'sender_postcode'       => WC()->countries->get_base_postcode(),
						'sender_residential'    => empty($settings['store_company']) ? 'Yes' : 'No',
						'sender_phone'          => $settings['store_phone'],
						'sender_email'          => $settings['store_email'],
						'sender_loading_assist' => ''
					);
				}

				/* Prepare Receiver */

				if ( $order->has_shipping_address() ) {		// If customer has entered shipping address in addition to billing address
					$name_words = explode( ' ', $order->get_formatted_shipping_full_name() );
					$receiver_name_first = array_shift( $name_words );
					$receiver_name_last  = implode( $name_words );

					$locality = WC_Myfreight_Ancillary::clean_myfreight_locality(	$order->get_shipping_city(),
													$order->get_shipping_state(),
													$order->get_shipping_postcode(),
													(int)($settings['address_match_percentage']) );

					/*
					 * Filter hook to determine whether delivery address is residential (as opposed to commercial).
					 * The "Delivery Address Type" setting specifies how this is determined.
					 * If the setting is "Check for Company" and the purchaser is logged in, the absence of a Shipping (or Billing) company implies residential.
					 * Web developers can use this hook to determine whether a delivery address is residential via other means such as a cart input or customer database.
					 * The result determines whether residential delivery surcharges are applied to certain services.
					 */
					$receiver_residential = apply_filters( 'my_residential', true, $settings['delivery_address_type'], $order->get_shipping_company(), get_post_meta( $order->get_id(), '_address_type', true ) );	// Is delivery address residential?

					if ( is_callable( array( $order, 'get_shipping_phone' ) ) ) {
						$receiver_phone = $order->get_shipping_phone();
					} else {
						$receiver_phone = get_post_meta( $order->get_id(), '_shipping_phone', true );
					}
					if ( empty( $receiver_phone ) ) {
						$receiver_phone = $order->get_billing_phone();
					}
					
					$receiver_details = array(
						'receiver_name_first'  => $receiver_name_first,
						'receiver_name_last'   => $receiver_name_last,
						'receiver_company'     => $order->get_shipping_company(),
						'receiver_street'      => $order->get_shipping_address_1(),
						'receiver_street2'     => $order->get_shipping_address_2(),
						'receiver_locality'    => $locality['locality'] . ' ' . $locality['region'] . ' ' . $locality['postcode'],
//						'receiver_city'        => $order->get_shipping_city(),
//						'receiver_state'       => $order->get_shipping_state(),
//						'receiver_postcode'    => $order->get_shipping_postcode(),
						'receiver_country'     => 'Australia',
						'receiver_residential' => $receiver_residential ? 'Yes' : 'No',
						'receiver_phone'       => $receiver_phone,
						'receiver_email'       => $order->get_billing_email()
					);
				} else {
					$name_words = explode( ' ', $order->get_formatted_billing_full_name() );
					$receiver_name_first = array_shift( $name_words );
					$receiver_name_last  = implode( $name_words );

					$locality = WC_Myfreight_Ancillary::clean_myfreight_locality(	$order->get_billing_city(),
													$order->get_billing_state(),
													$order->get_billing_postcode(),
													(int)($settings['address_match_percentage']) );

					/*
					 * Filter hook to determine whether delivery address is residential (as opposed to commercial).
					 * The "Delivery Address Type" setting specifies how this is determined.
					 * If the setting is "Check for Company" and the purchaser is logged in, the absence of a Shipping (or Billing) company implies residential.
					 * Web developers can use this hook to determine whether a delivery address is residential via other means such as a cart input or customer database.
					 * The result determines whether residential delivery surcharges are applied to certain services.
					 */
					$receiver_residential = apply_filters( 'my_residential', true, $settings['delivery_address_type'], $order->get_billing_company(), get_post_meta( $order->get_id(), '_address_type', true ) );	// Is delivery address residential?

					$receiver_details = array(
						'receiver_name_first'  => $receiver_name_first,
						'receiver_name_last'   => $receiver_name_last,
						'receiver_company'     => $order->get_billing_company(),
						'receiver_street'      => $order->get_billing_address_1(),
						'receiver_street2'     => $order->get_billing_address_2(),
						'receiver_locality'    => $locality['locality'] . ' ' . $locality['region'] . ' ' . $locality['postcode'],
//						'receiver_city'        => $order->get_billing_city(),
//						'receiver_state'       => $order->get_billing_state(),
//						'receiver_postcode'    => $order->get_billing_postcode(),
						'receiver_country'     => 'Australia',
						'receiver_residential' => $receiver_residential ? 'Yes' : 'No',
						'receiver_phone'       => $order->get_billing_phone(),
						'receiver_email'       => $order->get_billing_email()
					);
				}

				if ( empty($receiver_details['receiver_street']) && $settings['inhibit_empty_receiver'] == 'yes' ) {	// Check receiver address not empty
					return false;
				}

				/* Modify receiver email as specified to prevent Advanced Shipping Notice (ASN) going to purchaser */
				$consign_receiver_email = $settings['consign_receiver_email'];
				if ( $consign_receiver_email == 'Sender' ) {
					$receiver_details['receiver_email'] = $settings['store_email'];
				} else if ( $consign_receiver_email == 'Empty' ) {
					$receiver_details['receiver_email'] = '';
				}							// Else if 'Purchaser', leave as purchaser email

				/* Prepare Items */

				$fragile_item = false;
				$dangerous_goods = false;
				$packing_item = false;

				$shipping_items = array();

				$items = $order->get_items();

				foreach ( $items as $item_id => &$item ) {

					$variation_id = isset($item['variation_id']) ? $item['variation_id'] : 0;
					if ( $variation_id > 0 ) {									// If variation
						$variable_product_id = method_exists( $item, 'get_product_id' ) ? (int)($item->get_product_id()) : (int)($item['product_id']);
						$variable_product = wc_get_product( $variable_product_id );				// Get parent (variable) product
						if ( empty($variable_product) ) {
							$variable_product = new WC_Product_Variable( $variable_product_id );
						}

						$product = method_exists( $item, 'get_product' ) ? $item->get_product() : null;		// Get child (variation) product
						if ( empty($product) ) {
							$product = wc_get_product( $variation_id );
							if ( empty($product) ) {
								$product = new WC_Product_Variation( $variation_id );
							}
						}
					} else {											// If not variation
						$product = method_exists( $item, 'get_product' ) ? $item->get_product() : null;
						if ( empty($product) ) {
							$product_id = method_exists( $item, 'get_product_id' ) ? (int)($item->get_product_id()) : (int)($item['product_id']);
							if ( $product_id > 0 ) {
								$product = wc_get_product( $product_id );
								if ( empty($product) ) {
									$product = new WC_Product( $product_id );
								}
							}
						}
					}
					if ( !is_object( $product ) ) {
						continue;
					}
					if ( method_exists( $product, 'is_virtual' ) && $product->is_virtual() ) {		// If virtual product no shipping is required
						continue;
					}

					$item_quantity = method_exists( $item, 'get_quantity' ) ? (int)($item->get_quantity()) : (int)($item['quantity']);
					if ( empty($item_quantity) ) {
						continue;
					}

					/* Retrieve product weight and dimensions, or defaults if missing */
					$item_weight = method_exists( $product, 'get_weight' ) ? (float)($product->get_weight()) : 0.0;
					if ( empty($item_weight) && $variation_id > 0 ) {			// If variation
						$item_weight = method_exists( $variable_product, 'get_weight' ) ? (float)($variable_product->get_weight()) : 0.0;
					}
					if ( empty($item_weight) ) {
						$item_weight = (float)($settings['default_weight']);
					}

					$item_length = method_exists( $product, 'get_length' ) ? (int)($product->get_length()) : 0;
					if ( empty($item_length) && $variation_id > 0 ) {			// If variation
						$item_length = method_exists( $variable_product, 'get_length' ) ? (int)($variable_product->get_length()) : 0;
					}
					if ( empty($item_length) ) {
						$item_length = (int)($settings['default_length']);
					}

					$item_width  = method_exists( $product, 'get_width' ) ? (int)($product->get_width()) : 0;
					if ( empty($item_width) && $variation_id > 0 ) {			// If variation
						$item_width = method_exists( $variable_product, 'get_width' ) ? (int)($variable_product->get_width()) : 0;
					}
					if ( empty($item_width) ) {
						$item_width = (int)($settings['default_width']);
					}

					$height_string = method_exists( $product, 'get_height' ) ? $product->get_height() : null;
					if ( empty($height_string) && $variation_id > 0 ) {			// If variation
						$height_string = method_exists( $variable_product, 'get_height' ) ? $variable_product->get_height() : null;
					}
					if ( empty($height_string) ) {
						$height_string = $settings['default_height'];
					}
					$item_height = (int)$height_string;
					$item_height_decimal = (float)$height_string;
					if ( $item_height == 0 && $item_height_decimal > 0.0 ) {
						$item_height = 1;
					}

//					$shipping_class = method_exists( $product, 'get_shipping_class' ) ? $product->get_shipping_class() : get_term_by( 'id', (int)($product->shipping_class_id), 'product_shipping_class' )->slug;
					$shipping_class = method_exists( $product, 'get_shipping_class' ) ? $product->get_shipping_class() : null;
					if ( empty($shipping_class) && $variation_id > 0 ) {			// If variation
						$shipping_class = method_exists( $variable_product, 'get_shipping_class' ) ? $variable_product->get_shipping_class() : null;
					}
					$shipping_class_prefix = strtolower( substr( $shipping_class, 0, 6 ) );
					$item_fragile = false;
					if ( $shipping_class_prefix == 'fragil' || $shipping_class_prefix == 'handle' || $shipping_class_prefix == 'manual' ) {		// If product shipping class is fragile
						$item_fragile = true;
						$fragile_item = true;
					}
					if ( $shipping_class_prefix == 'danger' || $shipping_class_prefix == 'hazard' ) {
						$dangerous_goods = true;
					}

					if ( $variation_id > 0 ) {						// If variation
						$packing_unit = $product->get_meta('packing_unit');
						if ( empty($packing_unit) ) {
							$packing_unit = $variable_product->get_meta('product_packing_unit');
						}
					} else {
						$packing_unit = $product->get_meta('product_packing_unit');
					}
					if ( !empty($packing_unit) ) {						// If packing unit field completed
						$packing_item = true;
					} else {
						$packing_unit = null;
					}

					$shipping_items[] = array(
						'item_type_description' => apply_filters( 'my_est_item_type', 'Carton', $item_quantity, $item_weight, $item_length, $item_width, $item_height, $shipping_class, $settings['account_type'] ),
						'quantity'              => $item_quantity,
						'weight_in_kilograms'   => $item_weight * $item_quantity,
						'length'                => $item_length,
						'width'	                => $item_width,
						'height'                => $item_height,
						'height_decimal'        => $item_height_decimal,
						'fragile'               => $item_fragile,
						'packing_unit'          => $packing_unit
					);

					// If plugin "Add More Shipping Fields (For Multi Part Product)" (by PluginHive) active to support multiple shipping items per product
					for ( $i = 2; $i < 10 && !empty($product->get_meta( '_weight' . $i )); $i++ ) {		// For each additional shipping item
						$item_weight = (float)($product->get_meta( '_weight' . $i ));
						$item_length = !empty($product->get_meta( '_length' . $i )) ? (int)($product->get_meta( '_length' . $i )) : (int)($settings['default_length']);
						$item_width  = !empty($product->get_meta( '_width' . $i ))  ? (int)($product->get_meta( '_width' . $i ))  : (int)($settings['default_width']);
						$height_string = !empty($product->get_meta( '_height' . $i )) ? $product->get_meta( '_height' . $i ) : $settings['default_height'];
						$item_height = (int)$height_string;
						$item_height_decimal = (float)$height_string;
						if ( $item_height == 0 && $item_height_decimal > 0.0 ) {
							$item_height = 1;
						}

						/* Load additional items into items array */
						$shipping_items[] = array(
							'item_type_description' => apply_filters( 'my_est_item_type', 'Carton', $item_quantity, $item_weight, $item_length, $item_width, $item_height, $shipping_class, $settings['account_type'] ),
							'quantity'              => $item_quantity,
							'weight_in_kilograms'   => $item_weight * $item_quantity,
							'length'                => $item_length,
							'width'                 => $item_width,
							'height'                => $item_height,
							'height_decimal'        => $item_height_decimal,
							'fragile'               => $item_fragile,
							'packing_unit'          => $packing_unit
						);
					}
				}
				unset( $item );

				/*
				 * Filter hook to allow custom packing of items into containers.
				 * Packing rules are applied first, then auto packing.
				 * Custom packing rules can also be hooked.
				 */
				$shipping_items = apply_filters( 'my_pack_items', $shipping_items, false, $packing_item );	// Pack shipping items together as defined by settings

				if ( empty($shipping_items) ) {
					return false;
				}

				$unloading_threshold = (int)$settings['unloading_threshold'];
				$unloading_capability_active = ( $settings['unloading_capability'] == 'yes' );
				$max_item_weight = 0.0;

				$booking_items = array();
				$items_line = 0;

				foreach ( $shipping_items as &$shipping_item ) {

					++$items_line;

					$item_weight = (float)round( $shipping_item['weight_in_kilograms'] / $shipping_item['quantity'], 1 );
					if ( $item_weight > $max_item_weight ) {
						$max_item_weight = $item_weight;
					}

					$booking_items[ $items_line . '_new_item_type' ] = 1;
					$booking_items[ $items_line . '_item_type' ]     = $shipping_item['item_type_description'];
					$booking_items[ $items_line . '_item_quantity' ] = $shipping_item['quantity'];
					$booking_items[ $items_line . '_item_weight' ]   = $item_weight;
					$booking_items[ $items_line . '_item_length' ]   = $shipping_item['length'];
					$booking_items[ $items_line . '_item_width' ]    = $shipping_item['width'];
					$booking_items[ $items_line . '_item_height' ]   = $shipping_item['height'];
				}
				unset( $shipping_item );

				/* Prepare Booking Pre-fill */

				$lift_assist = ( !empty($unloading_threshold) && $max_item_weight >= $unloading_threshold && ( !$unloading_capability_active || get_post_meta( $order->get_id(), '_unloading_capability', true ) != 'Yes' ) );

				$timezone_string = get_option( 'timezone_string' );
				date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );

				$delivery_date = null;
				if ( is_callable( array( 'Orddd_Lite_Common', 'orddd_lite_get_order_delivery_date' ) ) ) {	// If plugin "Order Delivery Date" (by Tyche Softwares) active
					$order_delivery_date = Orddd_Lite_Common::orddd_lite_get_order_delivery_date( $order->get_id() );
					if ( !empty($order_delivery_date) ) {
						$delivery_timestamp = Orddd_Lite_Common::orddd_lite_get_timestamp( $order_delivery_date, get_option( 'orddd_lite_delivery_date_format' ) );
						$delivery_date = date( 'D j M', $delivery_timestamp );
					}
				}

				$order_number = $order->get_order_number();
				$special_instructions = !empty($delivery_date) ? 'DELIVER: ' . $delivery_date . '. ' : '';
				$special_instructions .= method_exists( $order, 'get_customer_note' ) ? $order->get_customer_note() : '';

				$miscellaneous = array(
					'manual_handling'           => $fragile_item ? 'Yes' : 'No',
					'hazardous_goods'           => $dangerous_goods ? 'true' : 'void',
					'selected_service_code'     => ( !empty($shipping_method) && method_exists( $shipping_method, 'get_meta' ) ) ? $shipping_method->get_meta( 'Service_Code' ) : null,
					'tailgate_lift'             => $lift_assist ? 'Dropoff' : 'None',
					'sender_loading_assist'     => 'No',					// Can be overridden in Myparcel
					'receiver_unloading_assist' => $lift_assist ? 'Yes' : 'No',		// Can be overridden in Myparcel
					'customer_reference'        => 'Order #' . $order_number,
					'special_instructions'      => $special_instructions,
					'parcel_warranty'           => '0.00'
				);

				$success = false;
				$booking_parts = (int)ceil( $items_line / MYPARCEL_LINES_PER_BOOKING );		// Determine number of bookings required for item lines

				for ( $booking_part = 1; $booking_part <= $booking_parts; $booking_part++ ) {

					if ( $booking_parts <= 1 ) {					// If single booking
						$booking_number = $order_number;
						$myparcel_fields = array_merge( $booking_items, $sender_details, $receiver_details, $miscellaneous );
					} else {							// If multiple bookings
						$booking_number = $order_number . '-' . (string)$booking_part;
//						$miscellaneous['customer_reference'] = 'Order #' . $order_number . ' part ' . (string)$booking_part . ' of ' . (string)$booking_parts;
						$miscellaneous['customer_reference'] = 'Order #' . $order_number . ' part ' . (string)$booking_part;

						$line_offset = ( $booking_part - 1 ) * MYPARCEL_LINES_PER_BOOKING;
						$part_items = array();
						for ( $i = 1; $i <= MYPARCEL_LINES_PER_BOOKING; $i++ ) {		// Copy part booking item lines
							$j = $i + $line_offset;
							if ( isset($booking_items[ $j . '_new_item_type' ]) ) {
								$part_items[ $i . '_new_item_type' ] = $booking_items[ $j . '_new_item_type' ];
								$part_items[ $i . '_item_type' ]     = $booking_items[ $j . '_item_type' ];
								$part_items[ $i . '_item_quantity' ] = $booking_items[ $j . '_item_quantity' ];
								$part_items[ $i . '_item_weight' ]   = $booking_items[ $j . '_item_weight' ];
								$part_items[ $i . '_item_length' ]   = $booking_items[ $j . '_item_length' ];
								$part_items[ $i . '_item_width' ]    = $booking_items[ $j . '_item_width' ];
								$part_items[ $i . '_item_height' ]   = $booking_items[ $j . '_item_height' ];
							}
						}

						$myparcel_fields = array_merge( $part_items, $sender_details, $receiver_details, $miscellaneous );
					}
					$order_fields = array(
						$booking_number => $myparcel_fields
					);

					if ( !empty( self::$logger ) ) {
						self::$logger->debug( 'BOOKING CREATE: ' . print_r( $order_fields, true ), array( 'source' => 'Myfreight Shipping' ) );
					}

					$order_json = json_encode( $order_fields );

					$datime = date( 'Y-m-d\TH:i:s' );
					$index_entry = array(
						$booking_number => array(
							$datime,
							$sender_details['sender_locality'],
							$receiver_details['receiver_locality'] . ( $miscellaneous['selected_service_code'] == 'COLLECT' ? ' -> CC' : '' )
						)
					);
					$index_json = trim( json_encode( $index_entry ), '{}' ) . ",\n";

					$sync_directory = ABSPATH . ltrim( parse_url( $settings['myparcel_sync_url'], PHP_URL_PATH ), '/' );
					$index_file_name = preg_replace( '/[^a-z0-9]+/', '-', strtolower( $settings['myparcel_member_username'] ) ) . '.txt';
					$index_file_path = $sync_directory . $index_file_name;
					$order_file_name = $booking_number . '.txt';
					$order_file_path = $sync_directory . $order_file_name;
					$empty_file_path = $sync_directory . 'index.html';

					$order_file = fopen( $order_file_path, 'w' );
					chmod( $order_file_path, 0644 );
					if ( !$order_file ) {
						if ( !is_dir( $sync_directory ) ) {
							if ( mkdir( $sync_directory, 0755, true ) ) {
								touch( $empty_file_path );
								chmod( $empty_file_path, 0644 );

								if ( !empty( self::$logger ) ) {
									self::$logger->debug( 'SYNC DIRECTORY CREATED: ' . print_r( $sync_directory, true ), array( 'source' => 'Myfreight Shipping' ) );
								}
							} else {
								if ( !empty( self::$logger ) ) {
									self::$logger->debug( 'SYNC DIRECTORY CREATION ERROR: ' . print_r( $sync_directory, true ), array( 'source' => 'Myfreight Shipping' ) );
								}
							}
						} else {
							chmod( $sync_directory, 0755 );
						}
						$order_file = fopen( $order_file_path, 'w' );
						chmod( $order_file_path, 0644 );
					}
					if ( $order_file ) {
						if ( fwrite( $order_file, $order_json ) ) {

							$index_file = fopen( $index_file_path, 'a' );
							chmod( $index_file_path, 0644 );
							if ( !$index_file ) {
								$index_file = fopen( $index_file_path, 'a' );
								chmod( $index_file_path, 0644 );
							}
							if ( $index_file ) {
								if ( fwrite( $index_file, $index_json ) ) {
									fclose( $index_file );
									fclose( $order_file );

									/* Save order shipping details */
									$booking_link = MYPARCEL_URL . '/?orderdata=' . rawurlencode( $settings['myparcel_sync_url'] . $booking_number . '.txt' );
									$con_info = array(
										'account_type'  => 'Myparcel',
										'test_mode'     => $settings['test_mode'],
										'id'            => null,
										'status'        => 'unbooked',
										'con_link'      => $booking_link,
										'tracking_link' => null,
										'booking'       => $booking_number . ':' . $datime
									);

									if ( method_exists( $order, 'add_meta_data' ) ) {
										$order->add_meta_data( 'consignment', $con_info );
										$order->save_meta_data();
									}

									if ( $booking_part == 1 ) {
										$counters = get_option( 'myfreightshipping_counters', array() );
										$consign_count = strtolower( $settings['account_type'] ) . ( $settings['test_mode'] == 'no' ? '_live' : '_test' ) . '_consign_count';
										if ( empty($counters[$consign_count]) ) {
											$counters[$consign_count] = 1;
											$counters[$consign_count . '_start'] = date( 'Y-m-d' );
										} else {
											$counters[$consign_count]++;
										}
										update_option( 'myfreightshipping_counters', $counters, false );
									}

									$success = true;

									if ( !empty( self::$logger ) ) {
										self::$logger->debug( 'BOOKING INDEX ENTRY SAVED: ' . print_r( $index_json, true ), array( 'source' => 'Myfreight Shipping' ) );
									}

									continue;
								} else {
									fclose( $index_file );

									if ( !empty( self::$logger ) ) {
										self::$logger->debug( 'BOOKING INDEX ENTRY SAVE ERROR: ' . print_r( $index_json, true ), array( 'source' => 'Myfreight Shipping' ) );
									}
								}
							} else {
								if ( !empty( self::$logger ) ) {
									self::$logger->debug( 'ERROR OPENING BOOKING INDEX: ' . print_r( $index_file_path, true ), array( 'source' => 'Myfreight Shipping' ) );
								}
							}

							if ( !empty( self::$logger ) ) {
								self::$logger->debug( 'BOOKING DATA SAVED: ' . print_r( $order_json, true ), array( 'source' => 'Myfreight Shipping' ) );
							}
						} else {
							if ( !empty( self::$logger ) ) {
								self::$logger->debug( 'BOOKING DATA SAVE ERROR: ' . print_r( $order_json, true ), array( 'source' => 'Myfreight Shipping' ) );
							}
						}

						fclose( $order_file );
					} else {
						if ( !empty( self::$logger ) ) {
							self::$logger->debug( 'ERROR OPENING BOOKING FILE: ' . print_r( $order_file_path, true ), array( 'source' => 'Myfreight Shipping' ) );
						}
					}
				}

				if ( $success ) {
					return 0;
				} else {
					return false;
				}
			}


			/**
			 * Remove order from Myparcel shipments pending
			 */
			function remove_myparcel_order( $order ) {

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );

				/* Initialise debug */
				if ( $settings['debug'] == 'yes' ) {
					$timezone_string = get_option( 'timezone_string' );
					date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );
					self::$logger = wc_get_logger();
					self::$logger->debug( '==== Begin Removing Myparcel Consignment ====', array( 'source' => 'Myfreight Shipping' ) );
				} else {
					self::$logger = null;
				}

				$sync_directory = ABSPATH . ltrim( parse_url( $settings['myparcel_sync_url'], PHP_URL_PATH ), '/' );
				$index_file_name = preg_replace( '/[^a-z0-9]+/', '-', strtolower( $settings['myparcel_member_username'] ) ) . '.txt';
				$index_file_path = $sync_directory . $index_file_name;

				$order_number = $order->get_order_number();

				$index_json = file_get_contents( $index_file_path );
				$last_comma_pos = strrpos( $index_json, ',' );
				if ( $last_comma_pos !== false ) {
					$index_json[$last_comma_pos] = '}'; 
					$index_json = '{' . $index_json;
				}

				$index = json_decode( $index_json, true );
				if ( !empty($index) ) {
					if ( isset($index[ $order_number ]) ) {
						unset( $index[ $order_number ] );	// Remove single booking
					}
					for ( $part = 1; $part <= 10; $part++ ) {	// Remove part bookings
						if ( isset($index[ $order_number . '-' . $part ]) ) {
							unset( $index[ $order_number . '-' . $part ] );
						} else {
							break;
						}
					}

					if ( count( $index ) > 1100 ) {
						$index = array_slice( $index, -1000 );
					}

					if ( !empty($index) ) {
						$index_json = str_replace( "],", "],\n", trim( json_encode( $index ), '{}' ) . ',' );
					} else {
						$index_json = '';
					}

					if ( file_put_contents( $index_file_path, $index_json, LOCK_EX ) !== false ) {

						if ( !empty( self::$logger ) ) {
							self::$logger->debug( 'INDEX JSON SAVED AFTER REMOVAL: ' . print_r( $index_json, true ), array( 'source' => 'Myfreight Shipping' ) );
						}

						if ( $this->myparcel_remove_order_tracking_last_pending( $order ) ) {	// Remove last tracking line if still pending
							if ( !empty( self::$logger ) ) {
								self::$logger->debug( 'TRACKING LINE REMOVED', array( 'source' => 'Myfreight Shipping' ) );
							}
						} else {
							if ( !empty( self::$logger ) ) {
								self::$logger->debug( 'TRACKING LINE NOT REMOVED', array( 'source' => 'Myfreight Shipping' ) );
							}
						}

						return true;
					} else {
						if ( !empty( self::$logger ) ) {
							self::$logger->debug( 'INDEX JSON SAVE ERROR AFTER REMOVAL: ' . print_r( $index_json, true ), array( 'source' => 'Myfreight Shipping' ) );
						}
					}
				}
				return false;
			}


			/**
			 * Remove last pending tracking line from order
			 */
			private function myparcel_remove_order_tracking_last_pending( $order ) {

				/* Save order shipping status - DEPRECATED */
//				$shipping_methods = $order->get_shipping_methods();
//				$shipping_method = reset( $shipping_methods );
//				if ( !empty($shipping_method) and method_exists( $shipping_method, 'add_meta_data' ) ) {
//					$shipping_method->add_meta_data( 'Myparcel_Dashboard', 'Removed from pending' );
//					$shipping_method->save();
//				}

				$settings = get_option( 'woocommerce_myfreightshipping_settings' );
				$account_type = $settings['account_type'];

				if ( empty($order) || !method_exists( $order, 'get_meta' ) ) {
					return false;
				}

				$consignments = $order->get_meta( 'consignment', false );
				if ( empty($consignments) ) {
					return false;
				}

				$last_consignment = end( $consignments );
				if ( empty($last_consignment) ) {
					return false;
				}

				$con_meta = $last_consignment->get_data();
				if ( empty($con_meta) || empty($con_meta['value']) ) {
					return false;
				}

				$con_info = $con_meta['value'];
				if ( empty($con_info) || empty($con_info['account_type']) || $con_info['account_type'] != $account_type ) {
					return false;
				}

				if ( empty($con_info['id']) && !empty($con_meta['id']) ) {
					$last_con_meta_id = $con_meta['id'];
					if ( $last_con_meta_id && method_exists( $order, 'delete_meta_data_by_mid' ) ) {
						$order->delete_meta_data_by_mid( $last_con_meta_id );
						$order->save_meta_data();
						return true;
					}
				}

				return false;
			}


			/**
			 * Display tracking box in admin order page
			 */
			public function myfreightshipping_order_tracking_box( $order_id ) {

				$order = wc_get_order( $order_id );
				if ( !empty($order) && method_exists( $order, 'get_meta' ) ) {
					$consignments = $order->get_meta( 'consignment', false );
					if ( !empty($consignments) ) {

						$settings = get_option( 'woocommerce_myfreightshipping_settings' );

						echo '</tbody>';
						echo '<tbody id="order_tracking">';
						echo '<tr class="consignment_tracking_line">';
						echo '<td class="thumb">';
						echo '<img src="' . plugins_url( 'img/' . $settings['account_type'] . '-tracking-icon.png', __FILE__ ) . '" alt="' . $settings['account_type'] . '" width="24" />';
						echo '</td>';
						echo '<td class="name" colspan="6">Tracking';
						echo '<div class="view">';
						echo '<table class="display_meta" style="width:100%;">';
						echo '<tbody class="order_tracking_lines">';

						foreach ( $consignments as &$consignment ) {
							$con_meta = $consignment->get_data();
							$con_info = $con_meta['value'];
							if ( $con_info['account_type'] == 'Tracking' ) {
								echo '<tr>';
								echo '<th style="vertical-align:middle; width:56px;">Tracking:</th>';
								echo '<td style="vertical-align:middle; font-size:13px;">manual</td>';
								if ( !empty($con_info['service']) ) {
									echo '<th style="vertical-align:middle; text-align:right; width:50px;">Service:</th>';
									echo '<td style="vertical-align:middle; font-size:13px;">' . $con_info['service'] . '</td>';
								}
								if ( $settings['tracking_sync'] == 'yes' && !empty($con_info['status']) ) {
									echo '<th style="vertical-align:middle; text-align:right; width:46px;">Status:</th>';
									echo '<td style="vertical-align:middle; font-size:13px;">' . self::get_myfreight_state_short_description($con_info['status']) . '</td>';
								}
								echo '<td style="vertical-align:middle;">&nbsp;</td>';
								echo '<td style="vertical-align:middle;">&nbsp;</td>';
								echo '</tr>';
							} else if ( $con_info['account_type'] == 'Myfreight' ) {
								echo '<tr>';
								echo '<th style="vertical-align:middle; width:80px;">Myfreight Consignment:</th>';
								echo '<td style="vertical-align:middle; font-size:13px;">' . ( !empty($con_info['id']) ? $con_info['id'] : 'pending' ) . '</td>';
								if ( !empty($con_info['service']) ) {
									echo '<th style="vertical-align:middle; text-align:right; width:50px;">Service:</th>';
									echo '<td style="vertical-align:middle; font-size:13px;">' . $con_info['service'] . '</td>';
								}
								if ( $settings['tracking_sync'] == 'yes' && !empty($con_info['status']) ) {
									echo '<th style="vertical-align:middle; text-align:right; width:46px;">Status:</th>';
									echo '<td style="vertical-align:middle; font-size:13px;">' . self::get_myfreight_state_short_description($con_info['status']) . '</td>';
								}
								echo '<td style="vertical-align:middle; text-align:center;"><a class="button" href="' . $con_info['con_link'] . '" target="_blank" title="Click to open consignment in Myfreight"> Consignment → </a></td>';
								if ( $settings['tracking_sync'] == 'yes' && !empty($con_info['tracking_link']) ) {
									echo '<td style="vertical-align:middle; text-align:center;"><a class="button" href="' . $con_info['tracking_link'] . '" target="_blank" title="Click for Myfreight shipment tracking"> Tracking → </a></td>';
								} else {
									echo '<td style="vertical-align:middle;">&nbsp;</td>';
								}
								echo '</tr>';
							} else {
								echo '<tr>';
								echo '<th style="vertical-align:middle; width:56px;">Myparcel Booking:</th>';
								echo '<td style="vertical-align:middle; font-size:13px;">' . ( !empty($con_info['id']) ? $con_info['booking'] : 'pending' ) . '</td>';
								if ( !empty($con_info['service']) ) {
									echo '<th style="vertical-align:middle; text-align:right; width:50px;">Service:</th>';
									echo '<td style="vertical-align:middle; font-size:13px;">' . $con_info['service'] . '</td>';
								}
								if ( $settings['tracking_sync'] == 'yes' && !empty($con_info['status']) ) {
									echo '<th style="vertical-align:middle; text-align:right; width:46px;">Status:</th>';
									echo '<td style="vertical-align:middle; font-size:13px;">' . self::get_myfreight_state_short_description($con_info['status']) . '</td>';
								}
								if ( empty($con_info['id']) ) {
									echo '<td style="vertical-align:middle; text-align:center;"><a class="button" href="' . $con_info['con_link'] . '" target="_blank" title="Click to start order booking in Myparcel"> Book Shipment → </a></td>';
								} else {
									$booking_query = ( !empty($con_info['id']) && !empty($con_info['booking']) ) ? '?record=' . $con_info['booking'] : '';
									echo '<td style="vertical-align:middle; text-align:center;"><a class="button" href="' . MYPARCEL_URL . '/member-dashboard/' . $booking_query . '#_latestbookings" target="_blank" title="Click to view order booking in Myparcel"> View Shipment → </a></td>';
								}
								if ( $settings['tracking_sync'] == 'yes' && !empty($con_info['tracking_link']) ) {
									echo '<td style="vertical-align:middle; text-align:center;"><a class="button" href="' . $con_info['tracking_link'] . '" target="_blank" title="Click for shipment tracking"> Tracking → </a></td>';
								} else if ( empty($con_info['id']) ) {
									$order_query = '?order=' . $order_id;
									echo '<td style="vertical-align:middle; text-align:center;"><a class="button" href="' . MYPARCEL_URL . '/member-dashboard/' . $order_query . '#_shipmentspending" target="_blank" title="Click to view all order shipments pending in Myparcel"> All Pending → </a></td>';
								} else {
									echo '<td style="vertical-align:middle;">&nbsp;</td>';
								}
								echo '</tr>';
//								if ( empty($con_info['id']) ) {
//									echo '<tr>';
//									echo '<th>&nbsp;</th><td>&nbsp;</td><th>&nbsp;</th><td>&nbsp;</td><td style="vertical-align:middle; text-align:center;"><a href="#" title="Click to update tracking status manually">Manual tracking</a></td><td>&nbsp;</td>';
//									echo '</tr>';
//								}
							}
						}
						unset( $consignment );

						echo '</tbody>';
						echo '</table>';
						echo '</div>';
						echo '</td>';
						echo '</tr>';
					}
				}
			}


			/**
			 * Generate shipping information HTML for Completed Order email
			 */
			public function myfreightshipping_email_completed_shipping( $order, $sent_to_admin, $plain_text, $email ) {
				if ( !empty($email) && $email->id == 'customer_completed_order' ) {

					$shipping_methods = $order->get_shipping_methods();
					if ( !empty($shipping_methods) ) {
						$shipping_method = reset( $shipping_methods );
						if ( method_exists( $shipping_method, 'get_meta' ) ) {
							$service = $shipping_method->get_meta( 'Carrier_Service' );
						} else if ( method_exists( $shipping_method, 'get_method_title' ) ) {
							$service = $shipping_method->get_method_title();
						}
					}

					if ( !empty($order) && method_exists( $order, 'get_meta' ) ) {
						$consignments = $order->get_meta( 'consignment', false );
						if ( !empty($consignments) ) {
							$consignment = reset( $consignments );
							if ( !empty($consignment) && method_exists( $consignment, 'get_data' ) ) {
								$con_meta = $consignment->get_data();
								if ( !empty($con_meta) ) {
									$con_info = $con_meta['value'];

									if ( !empty($con_info['service']) && strlen( $con_info['service'] ) > 8 ) {
										$service = $con_info['service'];
									}
									if ( $con_info['account_type'] == 'Myfreight' ) {
										$reference = !empty($con_info['id']) ? $con_info['id'] : '-';
									} else {
										$reference = !empty($con_info['booking']) ? $con_info['booking'] : '-';
									}
									$tracking_link = !empty($con_info['tracking_link']) ? $con_info['tracking_link'] : null;

									$shipping_html = '<h2 class="email_shipping_data_heading">Shipping Information</h2>';
									$shipping_html .= '<div class="email_shipping_data_table" style="margin-bottom: 40px;">';
									$shipping_html .= '<table class="td" cellspacing="0" cellpadding="6" style="width: 100%; font-family: \'Helvetica Neue\', Helvetica, Roboto, Arial, sans-serif;" border="1">';
									$shipping_html .= '<thead>';
									$shipping_html .= '<tr>';
									$shipping_html .= '<th class="td" scope="col" style="text-align: left;">' . esc_html__( 'Shipping Service', 'woocommerce' ) . '</th>';
									$shipping_html .= '<th class="td" scope="col" style="text-align: left;">' . esc_html__( 'Reference', 'woocommerce' ) . '</th>';
									if ( $tracking_link ) {
										$shipping_html .= '<th class="td" scope="col" style="text-align: left;">' . esc_html__( 'Status', 'woocommerce' ) . '</th>';
									}
									$shipping_html .= '</tr>';
									$shipping_html .= '</thead>';
									$shipping_html .= '<tbody>';
									$shipping_html .= '<tr>';
									$shipping_html .= '<td class="td" scope="col" style="text-align: left;">' . esc_html__( $service, 'woocommerce' ) . '</td>';
									$shipping_html .= '<td class="td" scope="col" style="text-align: left;">' . esc_html__( $reference, 'woocommerce' ) . '</td>';
									if ( $tracking_link ) {
										$shipping_html .= '<td class="td" scope="col" style="text-align: left;"><a href="' . $tracking_link . '">Click to track</a></td>';
									}
									$shipping_html .= '</tr>';
									$shipping_html .= '</tbody>';
									$shipping_html .= '</table>';
									$shipping_html .= '</div>';

									/*
									 * Filter hook for customisation of customer order confirmation email shipping section.
									 */
									$shipping_html = apply_filters( 'my_email_shipping_info', $shipping_html, $order, $email->id );

									echo $shipping_html;
								}
							}
						}
					}
				}
			}


			/*
			 * Get Myfreight state short description.
			 */
			static function get_myfreight_state_short_description( $state ) {
				$mf_state_short_descriptions = array(
					"quote"                 => "Quoting",
					"unbooked"              => "Awaiting Booking",
					"open"                  => "Awaiting Despatch",
					"despatchable"          => "Despatch Pending",
					"despatching"           => "Despatching",
					"despatch_processing"   => "Despatching",
					"despatch_failure"      => "Despatch Fail/Retry",
					"despatched"            => "Despatched",
					"booked"                => "Pickup Pending",
					"failed_pickup"         => "Pickup Failed",
					"picked_up"             => "In Transit",
					"in_transit"            => "In Transit",
					"on_board_for_delivery" => "In Transit",
					"delivered"             => "Delivered",
					"delivered_with_pod"    => "Delivered & Signed",
					"partial_delivery"      => "Delivered Partial",
					"failed_delivery"       => "Delivery Failed",
					"failed_permanently"    => "Delivery Failed",
					"permanently_failed"    => "Delayed",
					"cancelled"             => "Cancelled"
				);

				$state_trimmed = trim($state);
				if ( array_key_exists($state_trimmed, $mf_state_short_descriptions) )
					return $mf_state_short_descriptions[$state_trimmed];
				else
					return $state_trimmed;
			}

		}
	}
	// Instantiate class
	new WC_Myfreight_Order();

}

