<?php
/*
 * Plugin Name: Myfreight Shipping
 * Plugin URI: https://www.myparcel.com.au/myfreight-woocommerce-shipping-plugin
 * Description: Australian shipping pricing and consigning for WooCommerce
 * Version: 1.15.1
 * Minimum PHP: 5.4
 * Author: Myfreight Australia
 * Author URI: https://www.myfreight.com.au
 * Developer: TaDah Data
 * Developer URI: https://www.tadahdata.com.au
 * Text Domain: myfreightshipping
 *
 * WC requires at least: 2.2
 * WC tested up to: 7.6.0
 *
 * License: GNU General Public License v3.0
 * License URI: https://www.gnu.org/licenses/gpl-3.0.html
 */


if ( ! defined( 'ABSPATH' ) ) {
	exit;	// Exit if accessed directly
}

/* Plugin definitions */
define( 'MYFREIGHT_SHIPPING_VERSION', '1.15.1' );
define( 'MY_DEFAULT_TIMEZONE', 'Australia/Sydney' );
define( 'MY_DESPATCH_DELAY_DAYS', 1 );
define( 'MYPARCEL_LINES_PER_BOOKING', 4 );
define( 'MYPARCEL_URL', 'https://www.myparcel.com.au' );
define( 'MYPARCEL_SURCHARGES_URL', 'https://www.myparcel.com.au/wp-content/uploads/myfreightshipping/myparcel_surcharges.json' );
define( 'MYFREIGHT_SURCHARGES_URL', 'https://www.myparcel.com.au/wp-content/uploads/myfreightshipping/myfreight_surcharges.json' );
define( 'EMPTY_SKID_WEIGHT',   10 );		// Nominal weight of empty skid (kg)
define( 'EMPTY_SKID_LENGTH',   60 );		// Nominal length of empty skid (cm)
define( 'EMPTY_SKID_WIDTH',    60 );		// Nominal width of empty skid (cm)
define( 'EMPTY_SKID_HEIGHT',   10 );		// Nominal height of empty skid (cm)
define( 'EMPTY_PALLET_WEIGHT', 30 );		// Nominal weight of empty pallet (kg)
define( 'EMPTY_PALLET_LENGTH', 116 );		// Nominal length of empty pallet (cm)
define( 'EMPTY_PALLET_WIDTH',  116 );		// Nominal width of empty pallet (cm)
define( 'EMPTY_PALLET_HEIGHT', 15 );		// Nominal height of empty pallet (cm)
define( 'MAX_PALLET_HEIGHT',   200 );		// Maximum height of auto packed pallet (cm)
define( 'PACKING_GAP_PERCENT', 10 );		// Default percentage increase in item shipping volume to account for packing padding/inefficiency
define( 'VOLUMETRIC_WEIGHT_FACTOR', 250 );	// Multiplier to derive volumetric (cubic) weight in kgs from volume in m3
define( 'GST_PERCENTAGE',      10.0 );		// Goods and Services Tax percentage

include_once( plugin_dir_path( __FILE__ ) . 'myfreight-localities.php' );
include_once( plugin_dir_path( __FILE__ ) . 'myfreight-ancillary.php' );
include_once( plugin_dir_path( __FILE__ ) . 'myfreight-order.php' );
include_once( plugin_dir_path( __FILE__ ) . 'myfreight-admin-order.php' );
include_once( plugin_dir_path( __FILE__ ) . 'myfreight-tracking.php' );


/**
 * Check if WooCommerce is active
 */
if ( in_array( 'woocommerce/woocommerce.php', apply_filters( 'active_plugins', get_option( 'active_plugins' ) ) ) ) {

	function myfreight_shipping_method_init() {
		if ( ! class_exists( 'WC_Myfreight_Shipping' ) ) {
			class WC_Myfreight_Shipping extends WC_Shipping_Method {

				/**
				 * Constructor for Myfreight Shipping class
				 *
				 * @access public
				 * @return void
				 */
				public function __construct() {
					$this->id                 = 'myfreightshipping';				// Id for shipping method
					$this->method_title       = __( 'Myfreight Shipping' );				// Title shown in admin
					$this->method_description = __( 'Myfreight Australia Shipping Pricing and Consigment' );	// Description shown in admin
					$this->enabled            = 'yes';						// Plugin enabled
					$this->title              = 'Myfreight Shipping';				// Class title

					$this->init();
				}


				/**
				 * Init Myfreight Shipping settings
				 *
				 * @access public
				 * @return void
				 */
				function init() {
					// Load the settings API
					$this->init_form_fields();	// Settings API. Override the method to add Myfreight settings
					$this->init_settings();		// Settings API. Load settings previously set

					if ( is_admin() ) {
						// Save settings in admin if defined
						add_action( 'woocommerce_update_options_shipping_' . $this->id, array( $this, 'process_admin_options' ), 10, 0 );
						add_filter( 'woocommerce_settings_api_sanitized_fields_' . $this->id, array( $this, 'process_myfreight_options' ), 10, 1 );
					}
				}


				/**
				 * Initialise Shipping Settings Form Fields
				 *
				 * @access public
				 * @return void
				 */
				function init_form_fields() {

					if ( is_admin() ) {
						/* Prepare default sync directory url */
						$sync_url = wp_get_upload_dir()['baseurl'] . '/myparcel-' . uniqid() . '/';

						/* Prepare link to logfile */
						if ( defined( 'WC_LOG_DIR' ) ) {
							$timezone_string = get_option( 'timezone_string' );
							date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );
							$log_url = add_query_arg( 'tab', 'logs', add_query_arg( 'page', 'wc-status', admin_url( 'admin.php' ) ) );
							$log_key = 'Myfreight-Shipping-' . date( 'Y-m-d' ) . '-' . sanitize_file_name( wp_hash( 'Myfreight Shipping' ) ) . '.log';
							$log_url = add_query_arg( 'log_file', $log_key, $log_url );
							$log_link = sprintf( __( '%1$sView Log%2$s', 'myfreightshipping' ), '<a href="' . esc_url( $log_url ) . '">', '</a>' );
						}
					} else {
						$sync_url = null;
						$log_link = null;
					}

					$this->form_fields = array(
						'shipping_settings_title' => array(
							'description' => __( 'Myfreight and Myparcel Shipping Settings<br /><a href="' . MYPARCEL_URL . '/myfreight-woocommerce-shipping-plugin/installation/" target="_blank">View online documentation</a><br />&nbsp;', 'myfreightshipping' ),
							'type' => 'title'
						),
						'highlight_basic_settings' => array(
							'title' => __( 'Configuration Assistance', 'myfreightshipping' ),
							'description' => __( 'Highlight initial fields to be completed. Turn this off later.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'class' => 'basic-setting',
							'label' => 'Highlight Important Settings',
							'default' => 'yes'
						),
						'feature_settings_title' => array(
							'title' => __( 'Feature Activation', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'shipping_pricing' => array(
							'title' => __( 'Shipping Pricing', 'myfreightshipping' ),
							'description' => __( 'Display shipping pricing on Cart and Checkout pages.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'class' => 'basic-setting',
							'label' => 'Enable Pricing',
							'default' => 'yes'
						),
						'auto_consigning' => array(
							'title' => __( 'Shipping Ordering', 'myfreightshipping' ),
							'description' => __( 'Activate shipping consignment auto generation upon successful order payment. Required for Order Tracking and Auto Completion below.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'class' => 'basic-setting',
							'label' => 'Enable Auto Consigning',
							'default' => 'yes'
						),
						'inhibit_empty_receiver' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Inhibit consignment generation if receiver address is empty. Tick for physical Point-of-Sale stores.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Consign Only If Receiver Has Address',
							'default' => 'no'
						),
						'consign_receiver_email' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'When auto consigning, choose which email address to use to pre-fill the shipping <strong>Receiver Email</strong>. Default is Purchaser email.<br />Myfreight and Myparcel send an Advanced Shipping Notice (ASN) to the receiver email, so change this if you wish only store branded emails to go to purchasers.', 'myfreightshipping' ),
							'type' => 'select',
							'css' => 'width: 400px;',
							'options' => array(
								'Purchaser' => 'Use Purchaser (Billing) email address for ASN receiver',
								'Sender'    => 'Use Sender (Store) email address for ASN receiver',
								'Empty'     => 'No email address for ASN receiver'
							),
							'default' => 'Purchaser'
						),
						'tracking_sync' => array(
							'title' => __( 'Shipment Tracking', 'myfreightshipping' ),
							'description' => __( 'Display Myfreight/Myparcel shipment status and tracking link on admin order pages.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'class' => 'basic-setting',
							'label' => 'Enable Admin Order Tracking',
							'default' => 'no'
						),
						'despatch_sync' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Set WooCommerce order status to <strong>Completed</strong> following despatch of associated consignment in Myfreight/Myparcel.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Auto Order Completion upon Despatch',
							'default' => 'no'
						),
						'email_completed_shipping' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Display shipping information near top of <strong>Completed order</strong> email sent to purchaser.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Shipping Info in Completed Email',
							'default' => 'no'
						),
						'address_helper_cart' => array(
							'title' => __( 'Address Entry', 'myfreightshipping' ),
							'description' => __( 'Activate locality lookup to assist with the correct entry of shipping address fields (Suburb, State and Postcode).', 'myfreightshipping' ),
							'type' => 'checkbox',
							'class' => 'basic-setting',
							'label' => 'Enable Address Helper on Cart Page',
							'default' => 'yes'
						),
						'address_helper_placeholder_cart' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Cart Address Helper placeholder text (default: Enter suburb or postcode ...).', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => 'Enter suburb or postcode ...'
						),
						'address_helper_checkout' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Activate locality lookup to assist with the correct entry of billing and shipping address fields (Suburb, State and Postcode).', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Address Helper on Checkout Page',
							'default' => 'yes'
						),
						'address_helper_placeholder_checkout' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Checkout Suburb placeholder text (default: Start entering suburb or postcode ...).', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => 'Start entering suburb or postcode ...'
						),
						'address_helper_product' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Activate locality lookup to assist with the correct entry of shipping address fields (Suburb, State and Postcode).<br />Requires plugin "WooCommerce Shipping Calculator on Product Page".', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Address Helper on Single Product Page',
							'default' => 'no'
						),
						'address_helper_placeholder_product' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Single Product Address Helper placeholder text (default: Enter suburb or postcode ...).', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => 'Enter suburb or postcode ...'
						),
						'phone_placeholder_checkout' => array(
							'title' => __( 'Phone Entry', 'myfreightshipping' ),
							'description' => __( 'Checkout Phone Number placeholder text (default: 10 digits with area code).<br />For shipping it is important that the full Australian 10 digit phone number for the receiver contact be entered.', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => '10 digits with area code'
						),
						'address_match_percentage' => array(
							'title' => __( 'Address Accuracy Threshold %', 'myfreightshipping' ),
							'description' => __( 'Percentage match with suburb/town/city and state required to generate shipping prices (default: 80). Postcode must be an exact match.<br />If perfect match required enter 100.', 'myfreightshipping' ),
							'placeholder' => __( '%', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '100'
							),
							'css' => 'width: 200px;',
							'default' => '80'
						),
						'service_options_settings_title' => array(
							'title' => __( 'Service Options', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'max_prices' => array(
							'title' => __( 'Maximum Price Selections', 'myfreightshipping' ),
							'description' => __( 'Maximum number of standard shipping price options to display to purchaser. Lowest price first.', 'myfreightshipping' ),
							'type' => 'select',
							'css' => 'width: 200px;',
							'options' => array(
								'1' => '1',
								'2' => '2',
								'3' => '3',
								'4' => '4',
								'5' => '5',
								'6' => '6'
							),
							'default' => '1'
						),
						'show_carrier_logo' => array(
							'title' => __( 'Price Selection Display', 'myfreightshipping' ),
							'description' => __( 'Show transport carrier logo to purchaser.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Display Carrier Logo',
							'default' => 'no'
						),
						'show_carrier' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Show transport carrier name to purchaser.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Display Carrier',
							'default' => 'yes'
						),
						'show_service' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Show transport carrier service description to purchaser.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Display Service',
							'default' => 'yes'
						),
						'show_eta' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Show delivery Estimated Time of Arrival (date) to purchaser. Note that ETA accuracy cannot be guaranteed.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Display ETA',
							'default' => 'no'
						),
						'date_format' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Date display format for ETA. Must be PHP date format string. Examples:  "- j/n/y",  "- j M \D\e\l\i\v\e\r\y",  "- \E\s\t D j F, Y" (default: "- j M").', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => '- j M'
						),
						'default_selection_text' => array(
							'title' => __( 'Standard Service', 'myfreightshipping' ),
							'description' => __( 'Default standard service text to show purchaser. Only used if Display Carrier and Display Service above not ticked.', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => 'Standard Shipping'
						),
						'standard_fulfilment_days' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Business days to fulfil and despatch standard shipments.', 'myfreightshipping' ),
							'placeholder' => __( 'days', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '20'
							),
							'css' => 'width: 200px;',
							'default' => '1'
						),
						'show_priority_service' => array(
							'title' => __( 'Priority Service', 'myfreightshipping' ),
							'description' => __( 'Show premium service option (with earlier ETA than lowest cost service) to purchaser.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Display Priority Service Option',
							'default' => 'no'
						),
						'priority_selection_text' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Default priority service option text to show purchaser. Only used if Display Carrier and Display Service above not ticked.', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => 'Priority Shipping'
						),
						'priority_fulfilment_days' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Business days to fulfil and despatch priority shipments.', 'myfreightshipping' ),
							'placeholder' => __( 'days', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '20'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'show_customer_collect' => array(
							'title' => __( 'Customer Collect', 'myfreightshipping' ),
							'description' => __( 'Show Customer Collect free shipping option to purchaser. Will be applied to generated consignment if selected during ordering.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Display Customer Collect Option',
							'default' => 'no'
						),
						'customer_collect_text' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Customer Collect option text to show purchaser.', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => 'Customer Collect'
						),
						'customer_collect_states' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Australian state(s) where purchaser is shown the Customer Collect option. Use Ctrl key to select/unselect multiple. Default all.', 'myfreightshipping' ),
							'type' => 'multiselect',
							'css' => 'width: 200px;',
							'options' => array(
								'NSW' => 'NSW',
								'ACT' => 'ACT',
								'QLD' => 'QLD',
								'VIC' => 'VIC',
								'TAS' => 'TAS',
								'SA'  => 'SA',
								'WA'  => 'WA',
								'NT'  => 'NT'
							),
//							'default' => array( WC()->countries->get_base_state() )
							'default' => null
						),
//						'customer_collect_clear_shipping' => array(
//							'title' => __( ' ', 'myfreightshipping' ),
//							'description' => __( 'Clear order shipping delivery address when customer collect option selected. Default not ticked.', 'myfreightshipping' ),
//							'type' => 'checkbox',
//							'label' => 'Clear Shipping when Customer Collect',
//							'default' => 'no'
//						),
						'customer_collect_fulfilment_days' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Business days to fulfil and despatch customer collect shipments.', 'myfreightshipping' ),
							'placeholder' => __( 'days', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '20'
							),
							'css' => 'width: 200px;',
							'default' => '1'
						),
						'free_shipping' => array(
							'title' => __( 'Free/Discounted Shipping', 'myfreightshipping' ),
							'description' => __( 'Activate Free or Discounted Shipping for purchaser when conditions specified below are met.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Conditional Free/Discounted Shipping',
							'default' => 'no'
						),
						'free_shipping_text' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Free/Discounted Shipping text to show purchaser. Start with colon to also display carrier/service. Clear to retain standard text.', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 300px;',
							'default' => 'Free Shipping'
						),
						'free_shipping_extent' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Indicate which shipping price options are eligible for free/discounted shipping. (default: First Shipping Option)', 'myfreightshipping' ),
							'type' => 'select',
							'css' => 'width: 300px;',
							'options' => array(
								'First' => 'First Shipping Option',
								'All' => 'All Shipping Options'
							),
							'default' => 'First'
						),
						'free_shipping_amount' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Discounted Shipping amount ($ ex GST). Zero for free shipping.', 'myfreightshipping' ),
							'placeholder' => __( '$', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.01'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'free_shipping_percentage' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Discounted Shipping percentage of full shipping cost (%). Zero for free shipping.', 'myfreightshipping' ),
							'placeholder' => __( '%', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '100'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'free_shipping_fulfilment_days' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Business days to fulfil and despatch free/discounted shipments.', 'myfreightshipping' ),
							'placeholder' => __( 'days', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '20'
							),
							'css' => 'width: 200px;',
							'default' => '1'
						),
						'free_shipping_conditions' => array(
							'title' => __( 'Free/Discounted Shipping Conditions', 'myfreightshipping' ),
							'description' => __( 'Indicate whether <strong>any</strong> or <strong>all</strong> of the 3 conditions below (when specified) need to be met for free/discounted shipping to apply.', 'myfreightshipping' ),
							'type' => 'select',
							'css' => 'width: 300px;',
							'options' => array(
								'Any' => 'Any One Condition Met',
								'All' => 'All Conditions Met'
							),
							'default' => 'Any'
						),
						'free_shipping_condition_min_product_amount' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( '<strong>Free/Discounted Shipping Condition:</strong>&nbsp; Minimum product spend amount (Cart Subtotal $).', 'myfreightshipping' ),
							'placeholder' => __( '$', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.01'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'free_shipping_condition_max_shipping_amount' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( '<strong>Free/Discounted Shipping Condition:</strong>&nbsp; Maximum shipping cost ($ ex GST).', 'myfreightshipping' ),
							'placeholder' => __( '$', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.01'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'free_shipping_condition_max_shipping_percentage' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( '<strong>Free/Discounted Shipping Condition:</strong>&nbsp; Maximum shipping cost as percentage of product spend amount (%). Calculated ex GST.', 'myfreightshipping' ),
							'placeholder' => __( '%', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '100'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'price_calculation_settings_title' => array(
							'title' => __( 'Price Calculation', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'markup_amount' => array(
							'title' => __( 'Markup $', 'myfreightshipping' ),
							'description' => __( 'Increase shipping prices to purchaser by this amount ($ ex GST).', 'myfreightshipping' ),
							'placeholder' => __( '$', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.01'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'markup_percentage' => array(
							'title' => __( 'Markup %', 'myfreightshipping' ),
							'description' => __( 'Increase shipping prices to purchaser by this percentage. Applied after Markup $ added.', 'myfreightshipping' ),
							'placeholder' => __( '%', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'minimum_price' => array(
							'title' => __( 'Minimum Price $', 'myfreightshipping' ),
							'description' => __( 'Minimum shipping price to charge purchaser ($ ex GST). Applied after Markup and Surcharges. Default 0.', 'myfreightshipping' ),
							'placeholder' => __( '$', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.01',
								'min'  => '0'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'round_up' => array(
							'title' => __( 'Rounding', 'myfreightshipping' ),
							'description' => __( 'Round up ex GST shipping prices to the nearest $. Applied after Markup and Surcharges. Default not ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Round Up Prices - Pre Tax',
							'default' => 'no'
						),
						'round_up_post_tax' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Round up inc GST shipping prices to the nearest $. Applied after Markup and Surcharges. Default not ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Round Up Prices - Post Tax',
							'default' => 'no'
						),
						'default_price' => array(
							'title' => __( 'Default Price $', 'myfreightshipping' ),
							'description' => __( 'Fallback shipping price to use if API offline ($ ex GST). Enter 0 to disallow shipping price display when API offline.', 'myfreightshipping' ),
							'placeholder' => __( '$', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.01',
								'min'  => '0'
							),
							'css' => 'width: 200px;',
							'default' => '0'
						),
						'default_weight' => array(
							'title' => __( 'Default Item Weight', 'myfreightshipping' ),
							'description' => __( 'Fallback weight to use if missing from product shipping data (kg). <strong>WARNING</strong>: If used, may give inaccurate shipping pricing.', 'myfreightshipping' ),
							'placeholder' => __( 'kg', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.1',
								'min'  => '0'
							),
							'class' => 'basic-setting',
							'css' => 'width: 200px;',
							'default' => ''
						),
						'default_length' => array(
							'title' => __( 'Default Item Dimensions', 'myfreightshipping' ),
							'description' => __( 'Fallback length to use if missing from product shipping data (cm). <strong>WARNING</strong>: If used, may give inaccurate shipping pricing.', 'myfreightshipping' ),
							'placeholder' => __( 'Length', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0'
							),
							'class' => 'basic-setting',
							'css' => 'width: 200px;',
							'default' => ''
						),
						'default_width' => array(
//							'title' => __( 'Default Item Width', 'myfreightshipping' ),
							'description' => __( 'Fallback width to use if missing from product shipping data (cm). <strong>WARNING</strong>: If used, may give inaccurate shipping pricing.', 'myfreightshipping' ),
							'placeholder' => __( 'Width', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0'
							),
							'class' => 'basic-setting',
							'css' => 'width: 200px;',
							'default' => ''
						),
						'default_height' => array(
//							'title' => __( 'Default Item Height', 'myfreightshipping' ),
							'description' => __( 'Fallback height to use if missing from product shipping data (cm). <strong>WARNING</strong>: If used, may give inaccurate shipping pricing.', 'myfreightshipping' ),
							'placeholder' => __( 'Height', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0'
							),
							'class' => 'basic-setting',
							'css' => 'width: 200px;',
							'default' => ''
						),
						'surcharge_settings_title' => array(
							'title' => __( 'Surcharges', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'surcharges_auto_update' => array(
							'title' => __( 'Surcharge Updates', 'myfreightshipping' ),
							'description' => __( 'Automatically upload the latest surcharge data from Myfreight/Myparcel daily. Default ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Surcharge Auto Update',
							'default' => 'yes'
						),
						'surcharges' => array(
							'title' => __( 'Surcharge Data', 'myfreightshipping' ),
							'description' => __( 'List of surcharge amounts (in JSON format) by carrier and service code. This data is automatically populated if Surcharge Auto Update is enabled.', 'myfreightshipping' ),
							'placeholder' => __( '- automatic upload -', 'myfreightshipping' ),
							'type' => 'textarea',
							'default' => ''
						),
						'delivery_address_type' => array(
							'title' => __( 'Delivery Address Type', 'myfreightshipping' ),
							'description' => __( 'Are delivery addresses Residential or Commercial/Business?<br />Can the presence/absence of a Company name be used to decide, or can the purchaser be asked (via tickbox in Cart/Checkout)?<br />Default: "Check for Company - Default Residential" (if purchaser logged in or in Checkout use the Shipping/Billing Company to decide).<br />This setting affects whether a Residential Delivery Surcharge is applied to the shipping price for applicable transport carriers.', 'myfreightshipping' ),
							'type' => 'select',
							'class' => 'basic-setting',
							'css' => 'width: 300px;',
							'options' => array(
								'Residential'         => 'Residential',
								'Commercial'          => 'Commercial',
								'Company Residential' => 'Check for Company - Default Residential',
								'Company Commercial'  => 'Check for Company - Default Commercial',
								'Enquire Residential' => 'Ask Purchaser - Default Residential',
								'Enquire Commercial'  => 'Ask Purchaser - Default Commercial'
							),
							'default' => 'Company Residential'
						),
						'unloading_threshold' => array(
							'title' => __( 'Unloading Weight Threshold', 'myfreightshipping' ),
							'description' => __( 'Single item weight (kg) above which a truck with tailgate lift is requested and Unloading Assist Surcharges may be applied.<br />Default is 30kg as drivers are generally not permitted to lift items exceeding this weight unaided.<br />Leave empty if heavy deliveries are to self unloading facilities (e.g. forklift or unloading facility available).<br />This setting affects whether an Unloading Assist Surcharge is applied to the shipping price for applicable transport carriers.', 'myfreightshipping' ),
							'placeholder' => __( 'kg', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0'
							),
							'css' => 'width: 200px;',
							'default' => '30'
						),
						'unloading_capability_check' => array(
							'title' => __( 'Receiver Unloading Capability', 'myfreightshipping' ),
							'description' => __( 'Ask purchaser (via tickbox in Cart and Checkout) whether unloading equipment/dock facilities are available for heavy items at dropoff (e.g. forklift).<br />This allows non-tailgate lift services to be included and any Unloading Assist Surcharges to be avoided. Default not ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Display Tickbox in Cart/Checkout to Query Receiver Unloading Facilities',
							'default' => 'no'
						),
						'unloading_capability_text' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Receiver Unloading Capability tickbox query text (default: Forklift or unloading facility at delivery address?).', 'myfreightshipping' ),
							'type' => 'text',
							'css' => 'width: 400px;',
							'default' => 'Forklift or unloading facility at delivery address?'
						),
						'packing_containers_title' => array(
							'title' => __( 'Packing Containers', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'packing_containers_note' => array(
							'description' => __( 'Auto Packing is recommended initially as it is simpler than Packing Rules.', 'myfreightshipping' ),
							'type' => 'title'
						),
						'packing_rules_active' => array(
							'title' => __( 'Packing Rule Activation', 'myfreightshipping' ),
							'description' => __( 'Activate rules to consolidate multiple products into larger containers for shipping. Specify Packing Rules below.<br />Myfreight custom item types can be used in rules. Packing Rules are applied prior to Auto Packing below. Default not ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Custom Packing Rules for Pricing and Ordering',
							'default' => 'no'
						),
						'packing_rules' => array(
							'title' => __( 'Packing Rules', 'myfreightshipping' ),
							'description' => __( 'Enter packing rules in CSV or JSON format. Include packing container dimensions and the number of packing units per container.<br />Rules entered as CSV (comma or tab delimited) will be automatically converted to JSON format. <a href="' . MYPARCEL_URL . '/wp-content/uploads/myfreightshipping/Myfreight_WooCommerce_Packing_Rules_Template.csv" target="_blank">Download CSV template.</a> | <a href="' . MYPARCEL_URL . '/wp-content/uploads/myfreightshipping/Myfreight_WooCommerce_Packing_Rules_Template.json" target="_blank">Download JSON template.</a><br />Note that rules are only applied to products with a <strong>Packing Unit</strong> type specified in the product&#39;s Shipping tab.', 'myfreightshipping' ),
							'placeholder' => __( '- optional -', 'myfreightshipping' ),
							'type' => 'textarea',
							'default' => ''
						),
						'packing_auto_active' => array(
							'title' => __( 'Auto Packing Activation', 'myfreightshipping' ),
							'description' => __( 'Consolidate products into a minimum number of containers for shipping.<br />The final item types (Satchel, Carton, Skid or Pallet) are estimated (unless disabled below) based on net weight and dimensions.<br />Auto packing occurs after Packing Rules above have been applied. Default not ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Automatic Packing Estimation for Pricing and Ordering',
							'default' => 'no'
						),
						'packing_auto_include_rule_packed' => array(
							'title' => __( 'Auto Pack If Rule Packed', 'myfreightshipping' ),
							'description' => __( 'Apply auto packing to items packed by packing rules above (i.e. packing within packing) as well as unpacked items.<br />If not ticked, items for which packing rules have been applied will be bypassed by auto packing. Default not ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Automatic Packing of Rule Packed Items',
							'default' => 'no'
						),
						'packing_auto_min_qty' => array(
							'title' => __( 'Auto Packing Minimum Qty', 'myfreightshipping' ),
							'description' => __( 'Minimum number of items required in order for auto packing to be applied. Default 2.', 'myfreightshipping' ),
							'placeholder' => __( '1 - 999', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '1'
							),
							'css' => 'width: 200px;',
							'default' => '2'
						),
						'packing_auto_max_container_weight' => array(
							'title' => __( 'Auto Packing Maximum Container Weight', 'myfreightshipping' ),
							'description' => __( 'Maximum weight for each auto packed container (kg). Default 20.<br />Setting this below the Unloading Weight Threshold may reduce transport carrier surcharges. Leave empty if no upper weight limit required.', 'myfreightshipping' ),
							'placeholder' => __( 'kg', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '0.1',
								'min'  => '0'
							),
							'css' => 'width: 200px;',
							'default' => '20'
						),
						'packing_auto_gap_percentage' => array(
							'title' => __( 'Auto Packing Gap %', 'myfreightshipping' ),
							'description' => __( 'Percentage to increase shipping volume by to account for gaps/padding when packing (%). Default 10.<br />Consider packing materials/padding and the space between pallet items and edge of pallet.', 'myfreightshipping' ),
							'placeholder' => __( '%', 'myfreightshipping' ),
							'type' => 'number',
							'custom_attributes' => array(
								'step' => '1',
								'min'  => '0',
								'max'  => '200'
							),
							'css' => 'width: 200px;',
							'default' => '10'
						),
						'item_type_auto_active' => array(
							'title' => __( 'Auto Item Type', 'myfreightshipping' ),
							'description' => __( 'Enable item type estimating based on product/container weight and dimensions.<br />If not ticked, products with no Shipping Class item type specified and auto packed containers will be classified as <strong>Carton</strong> for shipping. Default ticked.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Automatic Item Type Estimation',
							'default' => 'yes'
						),
						'store_settings_title' => array(
							'title' => __( 'Store/Warehouse Shipping Details', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'store_address_note' => array(
							'description' => __( 'The Primary <strong>Store Address</strong> used for pickup is entered at <a href="./admin.php?page=wc-settings&tab=general">WooCommerce->Settings->General</a>.', 'myfreightshipping' ),
							'type' => 'title'
						),
						'store_company' => array(
							'title' => __( 'Primary Store Company', 'myfreightshipping' ),
							'description' => __( 'Company name for pickup.', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'store_name' => array(
							'title' => __( 'Primary Store Contact', 'myfreightshipping' ),
							'description' => __( 'Name of shipping contact.', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'store_phone' => array(
							'title' => __( 'Primary Store Phone', 'myfreightshipping' ),
							'description' => __( 'Australian 10 digit phone number for shipping contact.', 'myfreightshipping' ),
							'type' => 'tel',
							'class' => 'basic-setting',
							'default' => ''
						),
						'store_email' => array(
							'title' => __( 'Primary Store Email', 'myfreightshipping' ),
							'description' => __( 'Email address for shipping admin.', 'myfreightshipping' ),
							'type' => 'email',
							'class' => 'basic-setting',
							'default' => ''
						),
						'multiple_senders' => array(
							'title' => __( 'Multi Store', 'myfreightshipping' ),
							'description' => __( 'Indicate whether products are shipped from more than one location. Provide store codes and locations below.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Ship from Multiple Stores/Warehouses',
							'default' => 'no'
						),
						'account_settings_title' => array(
							'title' => __( 'Shipping Account', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'account_type' => array(
							'title' => __( 'Account Type', 'myfreightshipping' ),
							'description' => __( 'Indicate whether you have Myparcel <a href="' . MYPARCEL_URL . '/sign-up/" target="_blank" title="Myparcel Sign Up">membership</a> or a Myfreight account.', 'myfreightshipping' ),
							'type' => 'select',
							'options' => array(
								'Myparcel'  => 'Myparcel',
								'Myfreight' => 'Myfreight'
							),
							'class' => 'basic-setting',
							'css' => 'width: 200px;',
							'default' => 'Myparcel'
						),
						'myparcel_account_settings_title' => array(
							'title' => __( 'Myparcel Member', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title',
							'class' => 'myparcel'
						),
						'myparcel_member_username' => array(
							'title' => __( 'Myparcel Member ID', 'myfreightshipping' ),
							'description' => __( 'Your member username (email address) for Myparcel website <a href="' . MYPARCEL_URL . '/" target="_blank" title="Myparcel">www.myparcel.com.au</a>.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myparcel member use only -', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myparcel_username' => array(
							'title' => __( 'Myparcel API Username', 'myfreightshipping' ),
							'description' => __( 'Your API username obtained from your Myparcel member profile.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myparcel member use only -', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myparcel_password' => array(
							'title' => __( 'Myparcel API Password', 'myfreightshipping' ),
							'description' => __( 'Your API password obtained from your Myparcel member profile.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myparcel member use only -', 'myfreightshipping' ),
							'type' => 'password',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myparcel_site_code' => array(
							'title' => __( 'Myparcel Customer Code', 'myfreightshipping' ),
							'description' => __( 'Your customer code obtained from your Myparcel member profile.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myparcel member use only -', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myparcel_sync_url' => array(
							'title' => __( 'Myparcel Sync URL', 'myfreightshipping' ),
							'description' => __( 'URL for Myparcel order shipping synchronisation. Copy and paste to your Myparcel member profile.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myparcel member use only -', 'myfreightshipping' ),
							'type' => 'url',
							'class' => 'basic-setting',
							'css' => 'width: 600px;',
							'default' => $sync_url
						),
						'myparcel_sender_codes' => array(
							'title' => __( 'Default Store Code(s)', 'myfreightshipping' ),
							'description' => __( 'Specify <strong>default</strong> shipping sender address code(s). Separate multiple store codes with commas. Leave blank to ship from Primary Store only.<br />Must be a subset of stores/warehouses specified below. Maximum of 5 recommended.', 'myfreightshipping' ),
							'placeholder' => __( '- for Multi Store only -', 'myfreightshipping' ),
							'type' => 'text',
							'default' => ''
						),
						'myparcel_sender_addresses' => array(
							'title' => __( 'Store Locations', 'myfreightshipping' ),
							'description' => __( 'Enter a complete list of store codes and addresses in CSV or JSON format. Include the Primary Store in the list. Active only if Multi Store above ticked.<br />A list entered as CSV (comma or tab delimited) will be automatically converted to JSON format. <a href="' . MYPARCEL_URL . '/wp-content/uploads/myfreightshipping/Myfreight_WooCommerce_Multi_Store_Locations_Template.csv" target="_blank">Download CSV template.</a> | <a href="' . MYPARCEL_URL . '/wp-content/uploads/myfreightshipping/Myfreight_WooCommerce_Multi_Store_Locations_Template.json" target="_blank">Download JSON template.</a>', 'myfreightshipping' ),
							'placeholder' => __( '- for Multi Store only -', 'myfreightshipping' ),
							'type' => 'textarea',
							'default' => ''
						),
						'myparcel_remove_completed' => array(
							'title' => __( 'Order Completed', 'myfreightshipping' ),
							'description' => __( 'Automatically remove orders with <strong>Status = Completed</strong> from the Myparcel dashboard <strong>Order Shipments Pending</strong> list (recommended).', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Remove Pending Shipment upon Order Completion',
							'default' => 'yes'
						),
						'myparcel_Satchel' => array(
							'title' => __( 'Supported Item Types', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of generic satchel for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Satchel',
							'default' => 'yes'
						),
						'myparcel_Carton' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of carton for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Carton',
							'default' => 'yes'
						),
						'myparcel_Skid' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of skid for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Skid',
							'default' => 'yes'
						),
						'myparcel_Pallet' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of pallet for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Pallet',
							'default' => 'yes'
						),
						'myfreight_account_settings_title' => array(
							'title' => __( 'Myfreight Account', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title',
							'class' => 'myfreight'
						),
						'myfreight_username' => array(
							'title' => __( 'Myfreight API Username', 'myfreightshipping' ),
							'description' => __( 'Your API username provided by Myfreight.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myfreight account use only -', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myfreight_password' => array(
							'title' => __( 'Myfreight API Password', 'myfreightshipping' ),
							'description' => __( 'Your API password provided by Myfreight.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myfreight account use only -', 'myfreightshipping' ),
							'type' => 'password',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myfreight_site_code' => array(
							'title' => __( 'Myfreight Customer Code', 'myfreightshipping' ),
							'description' => __( 'Your customer code provided by Myfreight.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myfreight account use only -', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myfreight_sender_codes' => array(
							'title' => __( 'Default Store Code(s)', 'myfreightshipping' ),
							'description' => __( 'Specify <strong>default</strong> shipping sender address code(s) provided by Myfreight. Separate multiple store codes with commas.<br />Must be a single store code or a subset of stores specified below. Maximum of 5 recommended.', 'myfreightshipping' ),
							'placeholder' => __( '- for Myfreight account use only -', 'myfreightshipping' ),
							'type' => 'text',
							'class' => 'basic-setting',
							'default' => ''
						),
						'myfreight_sender_addresses' => array(
							'title' => __( 'Store Locations', 'myfreightshipping' ),
							'description' => __( 'Enter a complete list of store codes (provided by Myfreight) and addresses in CSV or JSON format. Include the Primary Store in the list.<br />A list entered as CSV will be automatically converted to JSON format. <a href="' . MYPARCEL_URL . '/wp-content/uploads/myfreightshipping/Myfreight_WooCommerce_Multi_Store_Locations_Template.csv" target="_blank">Download CSV template.</a> | <a href="' . MYPARCEL_URL . '/wp-content/uploads/myfreightshipping/Myfreight_WooCommerce_Multi_Store_Locations_Template.json" target="_blank">Download JSON template.</a>', 'myfreightshipping' ),
							'placeholder' => __( '- for Multi Store only -', 'myfreightshipping' ),
							'type' => 'textarea',
							'default' => ''
						),
						'myfreight_500gSatchel' => array(
							'title' => __( 'Supported Item Types', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of 500g satchel for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => '500g Satchel',
							'default' => 'no'
						),
						'myfreight_1kgSatchel' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of 1kg satchel for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => '1kg Satchel',
							'default' => 'no'
						),
						'myfreight_3kgSatchel' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of 3kg satchel for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => '3kg Satchel',
							'default' => 'no'
						),
						'myfreight_5kgSatchel' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of 5kg satchel for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => '5kg Satchel',
							'default' => 'no'
						),
						'myfreight_Satchel' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of standard satchel for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Satchel',
							'default' => 'yes'
						),
						'myfreight_Carton' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of carton for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Carton',
							'default' => 'yes'
						),
						'myfreight_Skid' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of skid for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Skid',
							'default' => 'yes'
						),
						'myfreight_Pallet' => array(
							'title' => __( ' ', 'myfreightshipping' ),
							'description' => __( 'Enable estimation of pallet for shipping.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Pallet',
							'default' => 'yes'
						),
						'diagnostic_settings_title' => array(
							'title' => __( 'Diagnostics', 'myfreightshipping' ),
							'description' => '<hr />',
							'type' => 'title'
						),
						'test_mode' => array(
							'title' => __( 'Test Mode', 'myfreightshipping' ),
							'description' => __( 'Connect to test API instead of production API. Disable during live operation.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'class' => 'basic-setting',
							'label' => 'Enable Test Mode of Operation',
							'default' => 'no'
						),
						'debug' => array(
							'title' => __( 'Debug', 'myfreightshipping' ),
							'description' => __( 'Save debug messages to WooCommerce->Status->Logs. Disable during live operation.', 'myfreightshipping' ),
							'type' => 'checkbox',
							'label' => 'Enable Debug Logging | ' . $log_link,
							'default' => 'no'
						 )
					);
				}


				/**
				 * Initialise Shipping Settings Page
				 *
				 * @access public
				 * @return void
				 */
				function admin_options() {

					wp_enqueue_style( 'myfreight-admin-styles', plugins_url( 'css/myfreightshipping-admin.css', __FILE__ ), null, false, 'screen' );
					wp_register_script( 'myfreight-admin-settings', plugins_url( 'js/myfreightAdminSettings.js', __FILE__ ), null, false, true );
					wp_enqueue_script( 'myfreight-admin-settings' );

?>
<h2><?php _e('Myfreight Shipping','myfreightshipping'); ?></h2>
<table class="form-table">
<?php $this->generate_settings_html(); ?>
</table>
<?php

				}


				/**
				 * process_myfreight_options function.
				 *
				 * @access public
				 * @param array $myfreight_settings
				 * @return array
				 */
				public function process_myfreight_options( $myfreight_settings ) {

					if ( $myfreight_settings['surcharges_auto_update'] == 'yes' ) {
						$surcharges_json = (new WC_Myfreight_Tracking)->mf_surcharge_update_exec( false );	// Check for updated surcharges
						if ( !empty($surcharges_json) ) {
							$myfreight_settings['surcharges'] = $surcharges_json;
						}
					}

					if ( $myfreight_settings['packing_rules_active'] == 'yes' ) {	// If packing rules active

						// Convert packing rules CSV to JSON if entered
						$packing_rules = ltrim( $myfreight_settings['packing_rules'] );
						if ( !empty($packing_rules) && substr( $packing_rules, 0, 14 ) == 'container_code' ) {
							if ( $packing_rules[14] == "\t" ) {		// If tab delimited
								$packing_rules_array = array_map( function($v){return str_getcsv($v, "\t");}, explode( "\n", $packing_rules ) );
							} else {					// If comma delimited
								$packing_rules_array = array_map( 'str_getcsv', explode( "\n", $packing_rules ) );
							}
							$packing_rules_object = array();
							for ( $i = 1; $i < count($packing_rules_array); $i++ ) {
								$packing_rules_object[ $packing_rules_array[$i][0] ] = array_combine( $packing_rules_array[0], $packing_rules_array[$i] );
							}
							$myfreight_settings['packing_rules'] = json_encode( $packing_rules_object, JSON_PRETTY_PRINT );
						}
					}

					if ( $myfreight_settings['multiple_senders'] == 'yes' ) {		// If Multi Store enabled

						// Convert store address list CSV to JSON if entered
						$sender_addresses = ltrim( $myfreight_settings['myparcel_sender_addresses'] );
						if ( !empty($sender_addresses) && substr( $sender_addresses, 0, 10 ) == 'store_code' ) {
							if ( $sender_addresses[10] == "\t" ) {		// If tab delimited
								$sender_address_array = array_map( function($v){return str_getcsv($v, "\t");}, explode( "\n", $sender_addresses ) );
							} else {					// If comma delimited
								$sender_address_array = array_map( 'str_getcsv', explode( "\n", $sender_addresses ) );
							}
							$sender_address_object = array();
							for ( $i = 1; $i < count($sender_address_array); $i++ ) {
								$sender_address_object[ $sender_address_array[$i][0] ] = array_combine( $sender_address_array[0], $sender_address_array[$i] );
							}
							$myfreight_settings['myparcel_sender_addresses'] = json_encode( $sender_address_object, JSON_PRETTY_PRINT );
						}
						$sender_addresses = ltrim( $myfreight_settings['myfreight_sender_addresses'] );
						if ( !empty($sender_addresses) && substr( $sender_addresses, 0, 10 ) == 'store_code' ) {
							if ( $sender_addresses[10] == "\t" ) {		// If tab delimited
								$sender_address_array = array_map( function($v){return str_getcsv($v, "\t");}, explode( "\n", $sender_addresses ) );
							} else {					// If comma delimited
								$sender_address_array = array_map( 'str_getcsv', explode( "\n", $sender_addresses ) );
							}
							$sender_address_object = array();
							for ( $i = 1; $i < count($sender_address_array); $i++ ) {
								$sender_address_object[ $sender_address_array[$i][0] ] = array_combine( $sender_address_array[0], $sender_address_array[$i] );
							}
							$myfreight_settings['myfreight_sender_addresses'] = json_encode( $sender_address_object, JSON_PRETTY_PRINT );
						}
					}

					return $myfreight_settings;
				}


				private $logger;		// Handle for debug logging


				/**
				 * calculate_shipping function.
				 *
				 * @access public
				 * @param mixed $package
				 * @return array
				 */
				public function calculate_shipping( $package = array() ) {
error_log("PACKAGE: " . print_r($package, true));

					/* Initialise debug */
					if ( $this->settings['debug'] == 'yes' ) {
						$timezone_string = get_option( 'timezone_string' );
						date_default_timezone_set( !empty($timezone_string) ? $timezone_string : MY_DEFAULT_TIMEZONE );
						$this->logger = wc_get_logger();
						if ( $this->settings['shipping_pricing'] == 'yes' ) {
							$this->logger->debug( '==== Begin Calculating ' . $this->settings['account_type'] . ' Shipping Price ====', array( 'source' => $this->title ) );
						}
					} else {
						$this->logger = null;
					}

					/* Check whether shipping pricing enabled */
					if ( $this->settings['shipping_pricing'] == 'no' ) {
						return null;
					}

					/* Get unloading assist settings */
					$unloading_threshold = (int)$this->settings['unloading_threshold'];

					/* Get default store codes */
					if ( $this->settings['account_type'] == 'Myfreight' ) {
						$default_sender_codes = explode( ',', str_replace( ' ', '', $this->settings['myfreight_sender_codes'] ) );
					} else {
						$default_sender_codes = explode( ',', str_replace( ' ', '', $this->settings['myparcel_sender_codes'] ) );
					}
					$common_sender_codes = array( 'init' );

					$shipping_items = array();

					$dangerous_item = false;
					$fragile_item = false;
					$heavy_item = false;
					$packing_item = false;

					foreach ( $package['contents'] as &$item ) {

						$variation_id = isset($item['variation_id']) ? $item['variation_id'] : 0;
						if ( $variation_id > 0 ) {						// If variation
							$variable_product_id = $item['product_id'];
							$variable_product = wc_get_product( $variable_product_id );	// Get parent product
						}

						$item_quantity = (int)($item['quantity']);
						if ( empty($item_quantity) ) {
							continue;
						}
						$product = $item['data'];
						if ( !is_object( $product ) ) {
							continue;
						}

						/* Retrieve product weight and dimensions, or defaults if missing */
						$item_weight = method_exists( $product, 'get_weight' ) ? (float)($product->get_weight()) : 0.0;
						if ( empty($item_weight) && $variation_id > 0 ) {			// If variation
							$item_weight = method_exists( $variable_product, 'get_weight' ) ? (float)($variable_product->get_weight()) : 0.0;
						}
						if ( empty($item_weight) ) {
							$item_weight = (float)($this->settings['default_weight']);
						}

						$item_length = method_exists( $product, 'get_length' ) ? (int)($product->get_length()) : 0;
						if ( empty($item_length) && $variation_id > 0 ) {			// If variation
							$item_length = method_exists( $variable_product, 'get_length' ) ? (int)($variable_product->get_length()) : 0;
						}
						if ( empty($item_length) ) {
							$item_length = (int)($this->settings['default_length']);
						}

						$item_width = method_exists( $product, 'get_width' ) ? (int)($product->get_width()) : 0;
						if ( empty($item_width) && $variation_id > 0 ) {			// If variation
							$item_width = method_exists( $variable_product, 'get_width' ) ? (int)($variable_product->get_width()) : 0;
						}
						if ( empty($item_width) ) {
							$item_width = (int)($this->settings['default_width']);
						}

						$height_string = method_exists( $product, 'get_height' ) ? $product->get_height() : null;
						if ( empty($height_string) && $variation_id > 0 ) {			// If variation
							$height_string = method_exists( $variable_product, 'get_height' ) ? $variable_product->get_height() : null;
						}
						if ( empty($height_string) ) {
							$height_string = $this->settings['default_height'];
						}
						$item_height = (int)$height_string;
						$item_height_decimal = (float)$height_string;
						if ( $item_height == 0 && $item_height_decimal > 0.0 ) {
							$item_height = 1;
						}

						$shipping_class = method_exists( $product, 'get_shipping_class' ) ? $product->get_shipping_class() : null;
						if ( empty($shipping_class) && $variation_id > 0 ) {			// If variation
							$shipping_class = method_exists( $variable_product, 'get_shipping_class' ) ? $variable_product->get_shipping_class() : null;
						}
						$shipping_class_prefix = strtolower( substr( $shipping_class, 0, 6 ) );

						/* Check whether any dangerous, fragile, heavy or long items */
						$item_dangerous = false;
						$item_fragile = false;
						if ( $shipping_class_prefix == 'danger' || $shipping_class_prefix == 'hazard' ) {	// If product shipping class is dangerous goods
							$item_dangerous = true;
							$dangerous_item = true;
						} else if ( $shipping_class_prefix == 'fragil' || $shipping_class_prefix == 'handle' || $shipping_class_prefix == 'manual' ) {	// If product shipping class is fragile
							$item_fragile = true;
							$fragile_item = true;
						}

						if ( !$dangerous_item && $this->settings['account_type'] == 'Myfreight' ) {
							if ( $variation_id > 0 ) {					// If variation
								$dg_data = method_exists( $variable_product, 'get_meta' ) ? $variable_product->get_meta('product_dangerous_goods_data') : null;
							} else {
								$dg_data = method_exists( $product, 'get_meta' ) ? $product->get_meta('product_dangerous_goods_data') : null;
							}
							if ( !empty($dg_data) ) {
								if ( $variation_id > 0 ) {				// If variation
									$dg_qty = (float)($product->get_meta('dangerous_goods_qty'));
									if ( empty($dg_qty) ) {
										$dg_qty = (float)($variable_product->get_meta('product_dangerous_goods_qty'));
									}
								} else {
									$dg_qty = (float)($product->get_meta('product_dangerous_goods_qty'));
								}
								if ( !empty($dg_qty) && $dg_qty > 0.0 ) {		// If product dangerous goods fields completed
									$dangerous_item = true;
								}
							}
						}
						if ( !empty($unloading_threshold) && $item_weight >= (float)$unloading_threshold ) {
							$heavy_item = true;
						}
						if ( $variation_id > 0 ) {				// If variation
							$packing_unit = method_exists( $product, 'get_meta' ) ? $product->get_meta('packing_unit') : null;
							if ( empty($packing_unit) ) {
								$packing_unit = method_exists( $variable_product, 'get_meta' ) ? $variable_product->get_meta('product_packing_unit') : null;
							}
						} else {
							$packing_unit = method_exists( $product, 'get_meta' ) ? $product->get_meta('product_packing_unit') : null;
						}
						if ( !empty($packing_unit) ) {				// If packing unit field completed
							$packing_item = true;
						} else {
							$packing_unit = null;
						}

						/* Load items array */
						$shipping_items[] = array(
							'item_type_description' => apply_filters( 'my_est_item_type', 'Carton', $item_quantity, $item_weight, $item_length, $item_width, $item_height, $shipping_class, $this->settings['account_type'] ),
							'quantity'              => $item_quantity,
							'weight_in_kilograms'   => $item_weight * $item_quantity,
							'length'                => $item_length,
							'width'                 => $item_width,
							'height'                => $item_height,
							'height_decimal'        => $item_height_decimal,
							'fragile'               => $item_fragile,
							'packing_unit'          => $packing_unit
						);

						// If plugin "Add More Shipping Fields (For Multi Part Product)" (by PluginHive) active to support multiple shipping items per product
						if ( is_plugin_active( 'Addon_Plugin_dividing_single_product/wf-add-more-shipping-fileds-addon.php' ) && method_exists( $product, 'get_meta' ) ) {
							for ( $i = 2; $i < 10 && !empty($product->get_meta( '_weight' . $i )); $i++ ) {		// For each additional shipping item
								$item_weight = (float)($product->get_meta( '_weight' . $i ));
								$item_length = !empty($product->get_meta( '_length' . $i )) ? (int)($product->get_meta( '_length' . $i )) : (int)($this->settings['default_length']);
								$item_width  = !empty($product->get_meta( '_width' . $i ))  ? (int)($product->get_meta( '_width' . $i ))  : (int)($this->settings['default_width']);
								$height_string = !empty($product->get_meta( '_height' . $i )) ? $product->get_meta( '_height' . $i ) : $this->settings['default_height'];
								$item_height = (int)$height_string;
								$item_height_decimal = (float)$height_string;
								if ( $item_height == 0 && $item_height_decimal > 0.0 ) {
									$item_height = 1;
								}

								/* Load additional items into items array */
								$shipping_items[] = array(
									'item_type_description' => apply_filters( 'my_est_item_type', 'Carton', $item_quantity, $item_weight, $item_length, $item_width, $item_height, $shipping_class, $this->settings['account_type'] ),
									'quantity'              => $item_quantity,
									'weight_in_kilograms'   => $item_weight * $item_quantity,
									'length'                => $item_length,
									'width'                 => $item_width,
									'height'                => $item_height,
									'height_decimal'        => $item_height_decimal,
									'fragile'               => $item_fragile,
									'packing_unit'          => $packing_unit
								);
							}
						}

						/* Determine common stores for all products */
						if ( $this->settings['multiple_senders'] == 'yes' ) {		// If multi store enabled
							/* Get store codes for product */
							$product_sender_codes = explode( ',', str_replace( ' ', '', $product->get_meta('product_sender_codes') ) );
							if ( !empty($product_sender_codes[0]) ) {
								$sender_codes = $product_sender_codes;
							} else {
								$sender_codes = $default_sender_codes;
							}
							if ( !empty($common_sender_codes[0]) && $common_sender_codes[0] == 'init' ) {		// If initial item
								$common_sender_codes = $sender_codes;						// Use product stores
							} else {
								$common_sender_codes = array_intersect( $common_sender_codes, $sender_codes );	// Intersection of all product stores so far
							}
						}
					}
					unset( $item );

					$heavy_container = $heavy_item;

					/*
					 * Filter hook to allow custom packing of items into containers.
					 * Packing rules are applied first, then auto packing.
					 * Custom packing rules can also be hooked.
					 */
					$shipping_items = apply_filters( 'my_pack_items', $shipping_items, true, $packing_item );	// Pack shipping items together as defined by settings

					/* Check for container over weight threshold */
					if ( !empty($unloading_threshold) && !$heavy_container ) {
						$max_item_weight = 0.0;
						foreach ( $shipping_items as &$shipping_item ) {
							$item_weight = (float)round( $shipping_item['weight_in_kilograms'] / $shipping_item['quantity'], 1 );
							if ( $item_weight > $max_item_weight ) {
								$max_item_weight = $item_weight;
							}
						}
						unset( $shipping_item );

						if ( $max_item_weight >= (float)$unloading_threshold ) {
							$heavy_container = true;
						}
					}

					/* Calculate shipment total weight and volume for estimation of surcharges */
					$dead_weight = 0.0;
					$qty_volume = 0;
					foreach ( $shipping_items as &$shipping_item ) {
						$item_weight = (float)$shipping_item['weight_in_kilograms'];
						$item_qty    = (int)$shipping_item['quantity'];
						$item_length = (int)$shipping_item['length'];
						$item_width  = (int)$shipping_item['width'];
						$item_height = (int)$shipping_item['height'];
						$item_height_decimal = (float)$shipping_item['height_decimal'];
						switch ( $shipping_item['item_type_description'] ) {
							case 'Skid':
								$item_weight += (float)EMPTY_SKID_WEIGHT * $item_qty;		// Add nominal skid weight
								$item_height += (int)EMPTY_SKID_HEIGHT;				// Add nominal skid height
								break;
							case 'Pallet':
								$item_weight += (float)EMPTY_PALLET_WEIGHT * $item_qty;		// Add nominal pallet weight
								$item_height += (int)EMPTY_PALLET_HEIGHT;			// Add nominal pallet height
								break;
						}
						$dead_weight += $item_weight;
						$qty_volume += (int)($item_qty * $item_length * $item_width * $item_height_decimal);
					}
					unset( $shipping_item );

					/* Load receiver array */
					$receiver = WC_Myfreight_Ancillary::clean_myfreight_locality( $package['destination']['city'],
												      $package['destination']['state'],
												      $package['destination']['postcode'],
												      (int)($this->settings['address_match_percentage']) );

					/* Load senders array */
					$senders = array();		// To hold each store code, sender location and final prices

					if ( $this->settings['multiple_senders'] == 'no' ) {		// If not multi store
						/* Load primary store location */
						$senders['Primary_Store'] = array(
							'locality' => WC()->countries->get_base_city(),
							'region'   => WC()->countries->get_base_state(),
							'postcode' => WC()->countries->get_base_postcode()
						);
					} else {							// If multi store
						/* Get store locations JSON array */
						if ( $this->settings['account_type'] == 'Myfreight' ) {
							$sender_locations = json_decode( $this->settings['myfreight_sender_addresses'], true );
						} else {
							$sender_locations = json_decode( $this->settings['myparcel_sender_addresses'], true );
						}

						if ( !empty($common_sender_codes) && $common_sender_codes[0] != 'init' ) {	// If common stores found for products
							if ( !empty($sender_locations) ) {
								foreach ( $sender_locations as $sender_code => &$sender_location ) {	// Get common localities
									if ( in_array( (string)$sender_code, $common_sender_codes ) ) {
										$senders[$sender_code] = array(
											'locality' => $sender_location['sender_city'],
											'region'   => $sender_location['sender_state'],
											'postcode' => $sender_location['sender_postcode']
										);
									}
								}
								unset( $sender_location );
							}
						}

						if ( empty($senders) ) {							// If no common stores found for products
							if ( !empty($sender_locations) ) {
								foreach ( $sender_locations as $sender_code => &$sender_location ) {	// Get default localities
									if ( in_array( (string)$sender_code, $default_sender_codes ) ) {
										$senders[$sender_code] = array(
											'locality' => $sender_location['sender_city'],
											'region'   => $sender_location['sender_state'],
											'postcode' => $sender_location['sender_postcode']
										);
									}
								}
								unset( $sender_location );
							}
						}
					}

					$price_response = false;	// Flag whether price response received from API

					$senders_prices = $this->get_myfreight_prices_multi( $senders, $receiver, $shipping_items, $dangerous_item );	// Get shipping prices via API

					if ( $senders_prices !== false ) {

						/* Retrieve receiver capability data to assist with service selection and surcharges */
						$receiver_company = false;
						if ( isset($package['destination']['company']) ) {			// If manual order/quote creation
							$receiver_company = $package['destination']['company'];
						} else if ( isset($_POST) && isset($_POST['post_data']) ) {
							if ( strpos( $_POST['post_data'], 'ship_to_different_address=1' ) !== false ) {
								if ( preg_match( '/shipping_company=(.*?)&/', $_POST['post_data'], $match ) == 1 /* && strlen( $match[1] ) > 1 */ ) {
									$receiver_company = $match[1];
								}
							} else {
								if ( preg_match( '/billing_company=(.*?)&/', $_POST['post_data'], $match ) == 1 /* && strlen( $match[1] ) > 1 */ ) {
									$receiver_company = $match[1];
								}
							}
						} else if ( !empty( $customer_data = WC()->session->get('customer') ) ) {
							if ( !empty($customer_data['shipping_company']) ) {
								$receiver_company = $customer_data['shipping_company'];
							}
						}

						$receiver_address_type = false;						// Assume no address type indicated
						if ( isset($package['destination']['address_type']) ) {			// If manual order/quote creation
							$receiver_address_type = $package['destination']['address_type'];
						} else if ( substr( $this->settings['delivery_address_type'], 0, 7 ) == 'Enquire' ) {
							if ( WC()->session->get( 'address_type', 0 ) == 1 ) {		// If address type ticked
								$receiver_address_type = ( substr( $this->settings['delivery_address_type'], -10 ) == 'Commercial' ) ? 'Residential' : 'Commercial';
							} else {
								$receiver_address_type = ( substr( $this->settings['delivery_address_type'], -10 ) == 'Commercial' ) ? 'Commercial' : 'Residential';
							}
						}

						$receiver_unloading_capability = false;					// Assume no unloading capability for heavy items at dropoff
						if ( isset($package['destination']['unloading_capability']) ) {		// If manual order/quote creation
							$receiver_unloading_capability = $package['destination']['unloading_capability'];
						} else if ( $this->settings['unloading_capability_check'] == 'yes' ) {
							if ( WC()->session->get( 'unloading_capability', 0 ) == 1 ) {	// If unloading capability ticked, no tailgate required
								$receiver_unloading_capability = true;
							}
							WC()->session->set( 'unloading_heavy_item', $heavy_container );	// If heavy item/container, save to inhibit tickbox display
						}

						/*
						 * Filter hook to determine whether delivery address is residential (as opposed to commercial).
						 * The "Delivery Address Type" setting specifies how this is determined.
						 * If the setting is "Check for Company" and the purchaser is logged in, the absence of a Shipping (or Billing) company implies residential.
						 * Web developers can use this hook to determine whether a delivery address is residential via other means such as a cart input or customer database.
						 * The result determines whether residential delivery surcharges are applied to certain services.
						 */
						$residential = apply_filters( 'my_residential', true, $this->settings['delivery_address_type'], $receiver_company, $receiver_address_type );	// Is delivery address residential?

						$service_options = array(
							'constraint'       => true,
							'residential'      => $residential,
							'manual_handling'  => $fragile_item,	// Manual handling is applied by some carriers even if not fragile
							'tailgate_lift'    => $heavy_container && !$receiver_unloading_capability,
							'dangerous_goods'  => $dangerous_item,
							'minimum_shipping' => true,
							'minimum_total'    => true
						);

						$all_surcharges_json = htmlspecialchars_decode( $this->settings['surcharges'] );
						$all_surcharges = json_decode( $all_surcharges_json, true );

						foreach ( $senders_prices as $sender_code => $prices ) {

							if ( $prices !== false )		// If API responding
								$price_response = true;

							if ( empty($prices) ) {			// If API not responding or prices empty
								continue;
							}

							// Add anticipated surcharges and calculate final shipping prices for sender
							$surcharge_prices = $this->estimate_surcharges( $prices, $shipping_items, $dead_weight, $qty_volume, $service_options, $all_surcharges );

							if ( !empty($surcharge_prices) ) {

								/*
								 * Filter hook to sort quotes into order to be displayed in cart and checkout.
								 */
								$sorted_prices = apply_filters( 'my_sort_quotes', $surcharge_prices, $sender_code );	// Sort quotes by cost to customer then delivery date

								$senders[$sender_code]['prices'] = $sorted_prices;
							}
						}
					}

					$rates = array();						// To store rates

					if ( $price_response ) {					// If rates available

						/* Sort senders by lowest cost to customer (then delivery date then second lowest cost) */
						uasort( $senders, function ($a, $b) {
							if ( !isset($a['prices'][0]['cost']) || !isset($b['prices'][0]['cost']) ) {
								return 0;
							}
							if ( $a['prices'][0]['cost'] == $b['prices'][0]['cost'] ) {				// Least cost sorted lowest to highest
								if ( !isset($a['prices'][0]['estimated_delivery_date']) || !isset($b['prices'][0]['estimated_delivery_date']) ) {
									return 0;
								}
								if ( $a['prices'][0]['estimated_delivery_date'] == $b['prices'][0]['estimated_delivery_date'] ) {	// ETA sorted earliest to lastest
									if ( !isset($a['prices'][1]) || !isset($b['prices'][1]) || empty($a['prices'][1]['cost']) || empty($b['prices'][1]['cost']) ) {
										return 0;
									}
									if ( $a['prices'][1]['cost'] == $b['prices'][1]['cost'] ) {		// Next least cost sorted lowest to highest
										return 0;
									}
									return ( (float)$a['prices'][1]['cost'] < (float)$b['prices'][1]['cost'] ) ? -1 : 1;
								}
								return ( $a['prices'][0]['estimated_delivery_date'] < $b['prices'][0]['estimated_delivery_date'] ) ? -1 : 1;
							}
							return ( (float)$a['prices'][0]['cost'] < (float)$b['prices'][0]['cost'] ) ? -1 : 1;
						});

						$sender_chosen = reset( $senders );		// Get lowest cost sender
						$sender_code = key( $senders );

						$earliest_edd = '2100-01-01';
						$earliest_edd_index = 0;

						if ( !empty($sender_chosen['prices']) ) {
							/* Zero any negative costs and find earliest ETA */
							foreach ( $sender_chosen['prices'] as $index => &$price_chosen ) {
								if ( isset($price_chosen) ) {
									if ( isset($price_chosen['cost']) && $price_chosen['cost'] < 0 ) {
										$price_chosen['cost'] = 0.00;
									}
									if ( isset($price_chosen['estimated_delivery_date']) && strcmp( $price_chosen['estimated_delivery_date'], $earliest_edd ) < 0 ) {
										$earliest_edd_index = $index;
										$earliest_edd = $price_chosen['estimated_delivery_date'];
									}
								}
							}
							unset( $price_chosen );
						}

						if ( !empty( $this->logger ) ) {
							$this->logger->debug( 'STORE CODE SELECTED: ' . print_r( $sender_code, true ), array( 'source' => $this->title ) );
							$this->logger->debug( 'PRICING RESPONSE SORTED: ' . print_r( $sender_chosen, true ), array( 'source' => $this->title ) );
						}

						/* Provide rates to WooCommerce */
						$max_prices      = (int)($this->settings['max_prices']);
						$show_carrier    = $this->settings['show_carrier'] == 'yes';
						$show_service    = $this->settings['show_service'] == 'yes';
						$show_eta        = $this->settings['show_eta'] == 'yes';
						$free_shipping   = $this->settings['free_shipping'] == 'yes';
						$free_extent_all = $this->settings['free_shipping_extent'] == 'All';
						$free_shipping_applied = false;

						for ( $i = 0; $i < $max_prices; $i++ ) {					// Standard service(s)
							if ( isset($sender_chosen['prices'][$i]) ) {				// If not empty rate

								// Check whether free shipping conditions are met
								if ( $free_shipping && !$free_shipping_applied ) {
									$free_shipping_applicable = $this->free_shipping_condition_check( (float)$package['contents_cost'], (float)$package['cart_subtotal'], (float)$sender_chosen['prices'][$i]['cost'], $receiver );
									$free_shipping_applied = ( $free_shipping_applicable && !$free_extent_all );
								} else {
									$free_shipping_applicable = false;
								}

								// Format standard shipping description
								$free_shipping_text = $this->settings['free_shipping_text'];
								if ( $free_shipping_applicable && !empty($free_shipping_text) && $free_shipping_text[0] != ':' ) {
									$label = $free_shipping_text;
								} else {
									if ( !$show_carrier && !$show_service ) {
										$label = $this->settings['default_selection_text'] . ( $i == 0 ? '' : ' ' . (string)($i + 1) );
									} else {
										$label = $show_carrier ? $sender_chosen['prices'][$i]['carrier'] : '';
										$label .= ( $show_carrier && $show_service ) ? ' ' : '';
										$label .= $show_service ? $sender_chosen['prices'][$i]['service'] : '';
									}
									if ( $free_shipping_applicable && !empty($free_shipping_text) && $free_shipping_text[0] == ':' ) {
										$label .= $free_shipping_text;
									}
								}
								if ( $show_eta ) {
									$service_fulfilment_days = $free_shipping_applicable ? $this->settings['free_shipping_fulfilment_days'] : $this->settings['standard_fulfilment_days'];
									$fulfilment_days = !empty($service_fulfilment_days) ? (int)$service_fulfilment_days : 0;
									$eta = WC_Myfreight_Ancillary::next_working_date( $sender_chosen['prices'][$i]['estimated_delivery_date'],
															  $fulfilment_days - MY_DESPATCH_DELAY_DAYS );
									$label .= ' ' . $this->eta_date_format( $eta );
								}

								$standard_cost = (float)$sender_chosen['prices'][$i]['cost'];
								$cost = $free_shipping_applicable ? (float)min( $standard_cost, $this->free_shipping_cost( $standard_cost ) ) : $standard_cost;
								$id = $this->id . (string)$i . ( ( $free_shipping_applicable && $cost < $standard_cost ) ? '_free' : '' );

								// Save the standard rate
								$rates[] = array(
									'id'        => $id,
									'label'     => $label,
									'cost'      => $cost,
									'calc_tax'  => 'per_order',
									'meta_data' => array(
											'Carrier_Service'   => $sender_chosen['prices'][$i]['carrier'] . ' ' . $sender_chosen['prices'][$i]['service'],
											'Service_Code'      => $sender_chosen['prices'][$i]['service_code'],
											'Store_Code'        => $sender_code,
											'Sender_Postcode'   => $sender_chosen['postcode'],
											'Receiver_Postcode' => $receiver['postcode']
									)
								);

								$free_shipping_applicable = false;

							} else if ( $max_prices < 6 ) {		// If empty rate, continue looking
								$max_prices++;
							}
						}

						$show_priority = $this->settings['show_priority_service'] == 'yes';

						if ( $show_priority && ( $earliest_edd_index >= $max_prices || ( $free_shipping_applied && $max_prices == 1 ) ) ) {	// If priority service
							if ( isset($sender_chosen['prices'][$earliest_edd_index]) ) {				// If not empty rate

								// Format priority description
								if ( !$show_carrier && !$show_service ) {
									$label = $this->settings['priority_selection_text'];
								} else {
									$label = $show_carrier ? $sender_chosen['prices'][$earliest_edd_index]['carrier'] : '';
									$label .= ( $show_carrier && $show_service ) ? ' ' : '';
									$label .= $show_service ? $sender_chosen['prices'][$earliest_edd_index]['service'] : '';
								}
								if ( $show_eta ) {
									$priority_fulfilment_days = $this->settings['priority_fulfilment_days'];
									$fulfilment_days = !empty($priority_fulfilment_days) ? (int)$priority_fulfilment_days : 0;
									$eta = WC_Myfreight_Ancillary::next_working_date( $sender_chosen['prices'][$earliest_edd_index]['estimated_delivery_date'],
															  $fulfilment_days - MY_DESPATCH_DELAY_DAYS );
									$label .= ' ' . $this->eta_date_format( $eta );
								}

								// Save the priority rate
								$rates[] = array(
									'id'        => $this->id . (string)$earliest_edd_index . '_priority',
									'label'     => $label,
									'cost'      => $sender_chosen['prices'][$earliest_edd_index]['cost'],
									'calc_tax'  => 'per_order',
									'meta_data' => array(
											'Carrier_Service'   => $sender_chosen['prices'][$earliest_edd_index]['carrier'] . ' ' . $sender_chosen['prices'][$earliest_edd_index]['service'],
											'Service_Code'      => $sender_chosen['prices'][$earliest_edd_index]['service_code'],
											'Store_Code'        => $sender_code,
											'Sender_Postcode'   => $sender_chosen['postcode'],
											'Receiver_Postcode' => $receiver['postcode']
									)
								);
							}
						}

						$quote_count = strtolower( $this->settings['account_type'] ) . ( $this->settings['test_mode'] == 'no' ? '_live' : '_test' ) . '_quote_count';
						$counters = get_option( 'myfreightshipping_counters', array() );
						if ( empty($counters[$quote_count]) ) {
							$counters[$quote_count] = 1;
							$counters[$quote_count . '_start'] = date( 'Y-m-d' );
						} else {
							$counters[$quote_count]++;
						}
						update_option( 'myfreightshipping_counters', $counters, false );

					} else if ( isset($this->settings['default_price']) && $this->settings['default_price'] > 0 ) {		// If rates unavailable and default rate specified
						// Save the default rate
						$rates[] = array(
							'id'        => $this->id . '_default',
							'label'     => 'Default Flat Rate',
							'cost'      => $this->settings['default_price'],
							'calc_tax'  => 'per_order',
							'meta_data' => array(	'Carrier_Service'   => 'Default',
										'Service_Code'      => 'DEFAULT',
										'Store_Code'        => 'Not_Specified',
										'Sender_Postcode'   => WC()->countries->get_base_postcode(),
										'Receiver_Postcode' => $receiver['postcode']
									)
						);
					}

					$show_customer_collect = $this->settings['show_customer_collect'] == 'yes';
					$customer_collect_states = $this->settings['customer_collect_states'];

					if ( $show_customer_collect && ( empty($customer_collect_states) || in_array( $receiver['region'], $customer_collect_states ) ) ) {
																		// If customer collect option specified
						// Format collect description
						$customer_collect_text = !empty($this->settings['customer_collect_text']) ? $this->settings['customer_collect_text'] : 'Customer Collect';

						$show_eta = $this->settings['show_eta'] == 'yes';
						if ( $show_eta ) {
							$customer_collect_fulfilment_days = $this->settings['customer_collect_fulfilment_days'];
							$fulfilment_days = !empty($customer_collect_fulfilment_days) ? (int)$customer_collect_fulfilment_days : 0;
							$eta = WC_Myfreight_Ancillary::next_working_date( 'today', $fulfilment_days );
							$customer_collect_text .= ' ' . $this->eta_date_format( $eta );
						}

						if ( !empty($sender_chosen['prices'][0]['cost']) && !empty($sender_chosen['postcode']) ) {
							$sender_postcode = $sender_chosen['postcode'];
						} else {
							$sender_postcode = WC()->countries->get_base_postcode();
						}

						// Save the collect rate
						$rates[] = array(
							'id'        => $this->id . '_collect',
							'label'     => $customer_collect_text,
							'cost'      => 0,
							'calc_tax'  => 'per_order',
							'meta_data' => array(	'Carrier_Service'   => 'Customer Collect',
										'Service_Code'      => 'COLLECT',
										'Store_Code'        => ( isset($sender_code) && $sender_code !== null ) ? $sender_code : 'Not_Specified',
										'Sender_Postcode'   => $sender_postcode,
										'Receiver_Postcode' => $receiver['postcode']
									)
						);
					}

					if ( !empty($rates) ) {

						// Register the rate(s)
						foreach ( $rates as $rate ) {
							if ( !empty($rate) ) {
								$this->add_rate( $rate );	// Register a rate
							}
						}

						return $rates;
					} else {
						return null;
					}
				}


				/**
				 * get_myfreight_prices_multi function.
				 *
				 * @access class
				 * @param array $senders
				 * @param array $receiver
				 * @param mixed $shipping_items
				 * @param boolean $dangerous_item
				 * @return mixed
				 */
				function get_myfreight_prices_multi( $senders, $receiver, $shipping_items, $dangerous_item ) {

					if ( empty($senders) || empty($receiver) || empty($shipping_items) ) {
						return false;
					}

					foreach ( $shipping_items as &$shipping_item ) {
						if ( isset($shipping_item['height_decimal']) ) {
							unset( $shipping_item['height_decimal'] );	// Remove decimal place (float) variants of item height
						}
					}
					unset( $shipping_item );

					if ( $this->settings['test_mode'] == 'yes' ) {
						$url = 'https://uat.teamwilberforce.com/public/api/v2/tender';
					} else {
						$url = 'https://go.myfreight.com.au/public/api/v2/tender';
					}

					if ( $this->settings['account_type'] == 'Myfreight' ) {
						$credentials = $this->settings['myfreight_username'] . ':' . $this->settings['myfreight_password'];
						$site_code = $this->settings['myfreight_site_code'];
						$return_type = null;
						$account_payable = 'sender';
					} else {
						$credentials = $this->settings['myparcel_username'] . ':' . $this->settings['myparcel_password'];
						$site_code = $this->settings['myparcel_site_code'];
						$return_type = 'third_party';
						$account_payable = 'receiver';
					}

					$header = array(
						'Content-type: application/json',
						'Accept: application/json',
						'Authorization: Basic ' . base64_encode( $credentials )
					);

					$quote_request = array(
						'site_code'       => $site_code,
						'return_type'     => $return_type,
						'account_payable' => $account_payable,
						'despatch_date'   => WC_Myfreight_Ancillary::next_working_date( 'today', MY_DESPATCH_DELAY_DAYS ),
						'dangerous_goods' => $dangerous_item,
						'sender'          => null,
						'receiver'        => $receiver,
						'items'           => $shipping_items
					);

					$chs = array();
					$senders_quotes = array();

					$mh = curl_multi_init();

					foreach ( $senders as $sender_code => &$sender ) {

						$quote_request['sender'] = $sender;

						if ( !empty( $this->logger ) ) {
							$this->logger->debug( 'PRICING REQUEST: ' . print_r( $quote_request, true ), array( 'source' => $this->title ) );
						}

						$body = json_encode( $quote_request );

						if ( !empty( $this->logger ) ) {
							$this->logger->debug( 'PRICING JSON SENT: ' . print_r( $body, true ), array( 'source' => $this->title ) );
						}

						$chs[$sender_code] = curl_init();
						curl_setopt( $chs[$sender_code], CURLOPT_URL, $url );
						curl_setopt( $chs[$sender_code], CURLOPT_SSL_VERIFYPEER, false );
						curl_setopt( $chs[$sender_code], CURLOPT_HTTPAUTH, CURLAUTH_BASIC );
						curl_setopt( $chs[$sender_code], CURLOPT_HTTPHEADER, $header );
						curl_setopt( $chs[$sender_code], CURLOPT_POST, true );
						curl_setopt( $chs[$sender_code], CURLOPT_POSTFIELDS, $body );
						curl_setopt( $chs[$sender_code], CURLOPT_RETURNTRANSFER, true );
						curl_setopt( $chs[$sender_code], CURLOPT_ENCODING, '' );
						curl_setopt( $chs[$sender_code], CURLOPT_CONNECTTIMEOUT, 10 );
						curl_setopt( $chs[$sender_code], CURLOPT_TIMEOUT, 20 );
						curl_setopt( $chs[$sender_code], CURLOPT_VERBOSE, false );

						curl_multi_add_handle( $mh, $chs[$sender_code] );
					}
					unset( $sender );

					$session_status = session_status();
					if ( $session_status == PHP_SESSION_ACTIVE ) {
						session_write_close();		// Close session to prevent AJAX blocking
					}

					$running = null;
					do {
						// Execute curl while still active
						$status = curl_multi_exec( $mh, $running );
					} while ( $status == CURLM_CALL_MULTI_PERFORM );

					$max_loop = 100;
					$partial_replies = 1;
					while ( ( $running || $partial_replies ) && $status == CURLM_OK && $max_loop-- > 0 ) {

						// Wait for activity on any curl connection
						$select = curl_multi_select( $mh, 1.0 );
						if ( $select == -1 ) {
							usleep(100000);
						} else if ( $select == 0 ) {
							usleep(100000);
						} else {
							usleep(20000);
						}

						// Continue to exec until curl gets more data
						do {
							$status = curl_multi_exec( $mh, $running );
						} while ( $status == CURLM_CALL_MULTI_PERFORM );

						if ( !empty( $this->logger ) ) {
							$this->logger->debug( "PRICING COUNTDOWN: " . $max_loop . " - PENDING REPLIES: " . $running . " - PARTIAL: " . $partial_replies . " - WAIT: " . $select . " - STATUS: " . $status, array( 'source' => $this->title ) );
						}

						$partial_replies = 0;

						foreach ( $chs as $sender_code => $ch ) {

							$data = curl_multi_getcontent( $ch );

							if ( !curl_errno( $ch ) && !empty($data) ) {

								if ( !empty( $this->logger ) ) {
									$this->logger->debug( 'PRICING JSON RECEIVED FROM STORE ' . $sender_code . ': ' . print_r( $data, true ), array( 'source' => $this->title ) );
								}

								$content_type = curl_getinfo( $ch, CURLINFO_CONTENT_TYPE );

								if ( substr( $content_type, 0, 16 ) == 'application/json' ) {
									if ( substr( $data, -1 ) == '}' ) {
										if ( $quotes = (array)json_decode( $data, true ) ) {
											if ( !empty( $quotes['carriers'] ) && empty( $quotes['errors'] ) ) {
												$senders_quotes[$sender_code] = $quotes['carriers'];
											} else {
												$senders_quotes[$sender_code] = null;

												if ( !empty( $quotes['errors'] ) ) {
													if ( !empty( $this->logger ) ) {
														$this->logger->debug( 'QUOTE ERROR: ' . print_r( $quotes['errors'][0], true ), array( 'source' => $this->title ) );
													}
												}
											}
											curl_multi_remove_handle( $mh, $ch );
											unset( $chs[$sender_code] );
										} else {
											$senders_quotes[$sender_code] = false;

											if ( !empty( $this->logger ) ) {
												$this->logger->debug( 'QUOTE DECODE ERROR: ' . print_r( $data, true ), array( 'source' => $this->title ) );
											}
										}
									} else {
										$partial_replies++;
									}
								} else {
									$senders_quotes[$sender_code] = false;
								}
							} else {
								$senders_quotes[$sender_code] = false;

								if ( !empty( $this->logger ) ) {
									if ( curl_errno($ch) ) {
										$this->logger->debug( 'API ERROR: ' . print_r( curl_error($ch), true ), array( 'source' => $this->title ) );
									}
								}
							}
						}
					}
					foreach ( $chs as $ch ) {
						curl_multi_remove_handle( $mh, $ch );
					}
					curl_multi_close( $mh );

					if ( $session_status == PHP_SESSION_ACTIVE ) {
						session_start();		// Reopen session
					}

					return $senders_quotes;
				}


				/**
				 * estimate_surcharges function.
				 *
				 * @access class private
				 * @param array $prices
				 * @param array $shipping_items
				 * @param float $dead_weight (kgs)
				 * @param int $qty_volume (cubic cms)
				 * @param array $service_options ( string => bool )
				 * @param array $all_surcharges
				 * @return array
				 *
				 * For all rates for sender:
				 *    - Anticipate surcharges;
				 *    - Calculate shipping, surcharge and total prices to purchaser after markup and tax;
				 *    - Exclude any services that do not support required service options.
				 */
				private function estimate_surcharges( $prices, $shipping_items, $dead_weight, $qty_volume, $service_options, $all_surcharges ) {

					/* Prepare markup, minimum and roundup */
					$markup_amount     = (float)$this->settings['markup_amount'];
					$markup_multiplier = ( 100.0 + (float)$this->settings['markup_percentage'] ) / 100.0;
					$minimum_price     = (float)$this->settings['minimum_price'];
					$round_up_pre_tax  = $this->settings['round_up'] == 'yes';
					$round_up_post_tax = $this->settings['round_up_post_tax'] == 'yes';

					/* Prepare surcharge measures */
					$cubic_weight     = (float)$qty_volume * VOLUMETRIC_WEIGHT_FACTOR / 1000000;
					$surcharge_weight = (float)max( $dead_weight, $cubic_weight );			// Greater of dead (actual) weight and volumetric (cubic) weight

					foreach ( $prices as $key => &$price ) {

						if ( !isset($price['total_exclusive']) || (float)$price['total_exclusive'] <= 0.00 ) {
							unset( $prices[$key] );
							continue;
						}

						/*
						 * Filter hook for custom estimation of surcharges to be applied to each quote's service.
						 * Checks service option availability and calculates surcharges applicable to specified transport carrier.
						 */
						$carrier_surcharges = apply_filters( 'my_surcharges', array(), $price['carrier'], $price['service_code'], $all_surcharges );

						$surcharges = 0.00;
						foreach ( $service_options as $surcharge_type => $surcharge_applicable ) {

							if ( !isset( $carrier_surcharges[$surcharge_type] ) ) {
								continue;
							}

							if ( !$surcharge_applicable && !( $surcharge_type == 'manual_handling' && is_array( $carrier_surcharges[$surcharge_type] ) ) ) {
								continue;
							}

							if ( $surcharge_type == 'constraint' ) {	// Check any requirements for this service to be used
								if ( is_array( $carrier_surcharges[$surcharge_type] ) ) {
									foreach ( $carrier_surcharges[$surcharge_type] as $constrained_option => $constraints ) {
										if ( $service_options[$constrained_option] && is_array( $constraints ) ) {
											foreach ( $constraints as $constraint => $permitted_values ) {
												if ( $constraint == 'option' ) {
													$option_active = false;
													/* Search options active */
													foreach ( $permitted_values as $permitted_value ) {
														if ( !empty($service_options[$permitted_value]) ) {
															$option_active = true;
															break;
														}
													}
													if ( !$option_active ) {		// If no option
														unset( $prices[$key] );		// Remove price for this service
														continue 4;
													}
												} else if ( $constraint == 'item_type' ) {
													$not_permitted_item_type = false;
													/* Search shipping items */
													foreach ( $shipping_items as $shipping_item ) {
														$item_type = $shipping_item['item_type_description'];
														if ( !in_array( $item_type, $permitted_values ) ) {
															$not_permitted_item_type = true; 
															break;
														}
													}
													if ( $not_permitted_item_type ) {	// If item type not allowed
														unset( $prices[$key] );		// Remove price for this service
														continue 4;
													}
												}
											}
										}
									}
								}
								continue;
							} else if ( $carrier_surcharges[$surcharge_type] === false ) {		// If service option not supported
								unset( $prices[$key] );						// Remove price for this service
								continue 2;
							} else if ( !is_array( $carrier_surcharges[$surcharge_type] ) ) {	// If simple surcharge
								if ( $surcharge_type == 'minimum_shipping' ) {			// Enforce minimum base shipping amount ex GST
									if ( !empty($carrier_surcharges[$surcharge_type]) ) {
										if ( (float)$price['total_exclusive'] < (float)$carrier_surcharges[$surcharge_type] ) {
											$surcharges += (float)$carrier_surcharges[$surcharge_type] - (float)$price['total_exclusive'];
										}
									}
								} else if ( $surcharge_type == 'minimum_total' ) {		// Enforce minimum shipping+surcharges ex GST
									if ( !empty($carrier_surcharges[$surcharge_type]) ) {
										if ( ( (float)$price['total_exclusive'] + $surcharges ) < (float)$carrier_surcharges[$surcharge_type] ) {
											$surcharges += (float)$carrier_surcharges[$surcharge_type] - ( (float)$price['total_exclusive'] + $surcharges );
										}
									}
								} else {
									$surcharges += (float)$carrier_surcharges[$surcharge_type];	// Add if applicable
								}
							} else {								// If complex surcharge
								$check_items = false;

								/* Calculate surcharge applicable to whole shipment */
								foreach ( array_reverse( $carrier_surcharges[$surcharge_type] ) as $threshold => $amount ) {

									/* Extract any measure-based surcharge thresholds ( i.e. measure = value unit : surcharge$ ) */
									if ( !preg_match( '/^([a-z_]*)([<>=]*)([0-9\.]+)([a-z]+)$/i', $threshold, $threshold_parts ) ) {
										continue;
									}

									if ( empty($threshold_parts[1]) || substr( $threshold_parts[1], 0, 5 ) != 'item_' ) {	// If surcharge applied to whole shipment
										if ( $threshold_parts[4] == 'kg' ) {
											$threshold_value = (float)$threshold_parts[3];
											if ( $threshold_parts[1] == 'dead_weight' || $threshold_parts[1] == 'weight' ) {
												if ( $dead_weight >= $threshold_value ) {
													$surcharges += (float)$amount;
													break;
												}
											} else if ( $threshold_parts[1] == 'cubic_weight' ) {
												if ( $cubic_weight >= $threshold_value ) {
													$surcharges += (float)$amount;
													break;
												}
											} else {	// If surcharge_weight or empty
												if ( $surcharge_weight >= $threshold_value ) {
													$surcharges += (float)$amount;
													break;
												}
											}
										} else if ( $threshold_parts[4] == 'cm' ) {
											$threshold_value = (int)$threshold_parts[3];
											if ( $threshold_parts[1] == 'volume' ) {
												if ( $qty_volume >= $threshold_value ) {
													$surcharges += (float)$amount;
													break;
												}
											}
										}
									} else {		// If surcharge applied per individual item
										$check_items = true;
									}
								}

								if ( !$check_items ) {		// If no per item surcharges apply
									continue;
								}

								/* Calculate surcharge applicable per individual item */
								foreach ( $shipping_items as &$shipping_item ) {
									$item_qty = (int)$shipping_item['quantity'];
									$compare_item = Array(
										'weight'   => (float)$shipping_item['weight_in_kilograms'] / $item_qty,
										'length'   => (int)$shipping_item['length'],
										'width'    => (int)$shipping_item['width'],
										'height'   => (int)$shipping_item['height'],
										'diagonal' => hypot( (float)$shipping_item['length'], (float)$shipping_item['width'] ),
										'volume'   => (int)((int)$shipping_item['length'] * (int)$shipping_item['width'] * ( isset($shipping_item['height_decimal']) ? (float)$shipping_item['height_decimal'] : (int)$shipping_item['height'] )),
										'fragile'  => (bool)$shipping_item['fragile']
									);

									foreach ( array_reverse( $carrier_surcharges[$surcharge_type] ) as $threshold => $amount ) {

										/* Extract any measure-based surcharge thresholds */
										if ( !preg_match( '/^([a-z_]*)([<>=]*)([0-9\.]+)([a-z]+)$/i', $threshold, $threshold_parts ) ) {
											continue;
										}
										if ( empty($threshold_parts[1]) || substr( $threshold_parts[1], 0, 5 ) != 'item_' ) {	// If surcharge applied to whole shipment
											continue;
										}

										$threshold_measure = substr( $threshold_parts[1], 5 );

										if ( !array_key_exists( $threshold_measure, $compare_item ) ) {		// If no matching measure
											continue;
										}

										$compare_measure = $compare_item[$threshold_measure];

										if ( $threshold_parts[4] == 'kg' ) {
											$threshold_value = (float)$threshold_parts[3];
										} else if ( $threshold_parts[4] == 'cm' ) {
											$threshold_value = (int)$threshold_parts[3];
										} else if ( $threshold_parts[4] == 'm' ) {
											if ( $threshold_measure == 'volume' ) {
												$threshold_value = (int)( 1000000 * (float)$threshold_parts[3] );
											} else {
												$threshold_value = (int)( 100 * (float)$threshold_parts[3] );
											}
										} else if ( $threshold_parts[4] == 'yes' || $threshold_parts[4] == 'no' ) {
											$threshold_value = (bool)$threshold_parts[3];
										} else {
											continue;
										}

										switch ( $threshold_parts[2] ) {
										case '>':
											$apply = ( $compare_measure > $threshold_value );
											break;
										case '>=':
											$apply = ( $compare_measure >= $threshold_value );
											break;
										case '<':
											$apply = ( $compare_measure < $threshold_value );
											break;
										case '<=':
											$apply = ( $compare_measure <= $threshold_value );
											break;
										case '=':
											$apply = ( $compare_measure == $threshold_value );
											break;
										default:
											$apply = ( $compare_measure >= $threshold_value );
											break;
										}

										if ( $apply ) {
											$surcharges += $item_qty * (float)$amount;
											break;
										}
									}
								}
								unset( $shipping_item );
							}
						}

						/* Calculate shipping price */
						$shipping = ( (float)$price['total_exclusive'] + $surcharges + $markup_amount ) * $markup_multiplier;

						/* Apply minimum price */
						if ( !empty($minimum_price) && $minimum_price > $shipping ) {
							$shipping = $minimum_price;
						}

						/* Round up to dollar or round to cent */
						$precision = 2;
						if ( $round_up_post_tax ) {
							$shipping = ceil( $shipping * ( 100.0 + GST_PERCENTAGE ) / 100.0 ) * 100.0 / ( 100.0 + GST_PERCENTAGE );
							$precision = 4;							// Increase precision to prevent rounding errors when tax added
						} else if ( $round_up_pre_tax ) {
							$shipping = ceil( $shipping );
						} else {
							$shipping = round( $shipping, 2 );
						}

						$price['surcharges'] = number_format( $surcharges, 2, '.', '' );	// Save estimated surcharges
						$price['cost'] = number_format( $shipping, $precision, '.', '' );	// Save customer cost

					}
					unset( $price );

					return $prices;
				}


				/**
				 * free_shipping_condition_check function.
				 *
				 * @access class private
				 * @param float $cart_subtotal_ex
				 * @param float $cart_subtotal_inc
				 * @param float $shipping_ex
				 * @param array $receiver
				 * @return bool
				 */
				private function free_shipping_condition_check( $cart_subtotal_ex, $cart_subtotal_inc, $shipping_ex, $receiver ) {
					$conditions = $this->settings['free_shipping_conditions'];
					$condition_min_product_amount      = (float)$this->settings['free_shipping_condition_min_product_amount'];
					$condition_max_shipping_amount     = (float)$this->settings['free_shipping_condition_max_shipping_amount'];
					$condition_max_shipping_percentage = (float)$this->settings['free_shipping_condition_max_shipping_percentage'];

					$meet_all = ( !empty($conditions) && $conditions == 'All' );

					if ( $meet_all &&
					    ( empty($condition_min_product_amount) || $cart_subtotal_inc >= $condition_min_product_amount ) &&
					    ( empty($condition_max_shipping_amount) || $shipping_ex <= $condition_max_shipping_amount ) &&
					    ( empty($condition_max_shipping_percentage) || ( 100.0 * $shipping_ex / $cart_subtotal_ex ) <= $condition_max_shipping_percentage ) ) {
						$eligible = true;
					} else if ( !$meet_all &&
					    ( ( !empty($condition_min_product_amount) && $cart_subtotal_inc >= $condition_min_product_amount ) ||
					    ( !empty($condition_max_shipping_amount) && $shipping_ex <= $condition_max_shipping_amount ) ||
					    ( !empty($condition_max_shipping_percentage) && ( 100.0 * $shipping_ex / $cart_subtotal_ex ) <= $condition_max_shipping_percentage ) ) ) {
						$eligible = true;
					} else {
						$eligible = false;
					}

					/*
					 * Filter hook to modify free shipping eligibility.
					 * This is executed following eligibility checks against settings "Free/Discounted Shipping Conditions". 
					 */
					return apply_filters( 'my_free_shipping_eligible', $eligible, $cart_subtotal_ex, $cart_subtotal_inc, $shipping_ex, $receiver );
				}


				/**
				 * free_shipping_cost function.
				 *
				 * @access class private
				 * @param float $standard_cost
				 * @return float
				 */
				private function free_shipping_cost( $standard_cost ) {
					$free_shipping_amount     = (float)$this->settings['free_shipping_amount'];
					$free_shipping_percentage = (float)$this->settings['free_shipping_percentage'];

					$discount_cost = !empty($free_shipping_amount) ? $free_shipping_amount : 0.00;
					$discount_cost += !empty($free_shipping_percentage) ? ( $standard_cost * $free_shipping_percentage / 100.0 ) : 0.00;

					return $discount_cost;
				}


				/**
				 * eta_date_format function.
				 *
				 * @access class private
				 * @param string $date
				 * @return string
				 */
				private function eta_date_format( $date ) {
					$date_format = $this->settings['date_format'];
					if ( empty($date_format) ) {
						$date_format = '- j M';
					}
					return date( $date_format, strtotime( $date ) );
				}

			}
		}
	}
	// Hook shipping method into WooCommerce

	add_action( 'woocommerce_shipping_init', 'myfreight_shipping_method_init' );

	function myfreight_shipping_method( $methods ) {
		$methods['myfreight_shipping_method'] = 'WC_Myfreight_Shipping';
		return $methods;
	}

	add_filter( 'woocommerce_shipping_methods', 'myfreight_shipping_method' );

}

